import os, argparse, tempfile, colors, textwrap, warnings, contextlib, tee, re, errno, shutil, psutil
from argparse_color_formatter import ColorHelpFormatter

# Fix a bug in the tee module #########
class FixTee(object):
    def flush(self):
        if (self.stream is not None):
            self.stream.flush()
        if (self.fp is not None):
            self.fp.flush()
            os.fsync(self.fp.fileno())

class StdoutTee(FixTee, tee.StdoutTee):
    pass
class StderrTee(FixTee, tee.StderrTee):
    pass
########################################

class HysopArgumentWarning(UserWarning):
    pass

class SplitAppendAction(argparse._AppendAction):
    def __init__(self, *args, **kwds):
        separator = kwds.pop('sep', ',')
        container = kwds.pop('container', tuple)
        convert   = kwds.pop('convert', lambda x: x)
        append    = kwds.pop('append', False)
        super(SplitAppendAction, self).__init__(*args, **kwds)
        self._separator = separator
        self._container = container
        self._convert   = convert
        self._append    = append
    
    def __call__(self, parser, namespace, values, option_string=None):
        assert isinstance(values, str), type(values)
        for c in ('(','{','[',']','}',')'):
            values = values.replace(c, '')
        try:
            values = tuple(self._convert(v) for v in values.split(self._separator))
        except:
            msg='Failed to convert \'{}\' to {} of {}s for parameter {}.'
            msg=msg.format(values, self._container.__name__, self._convert.__name__,
                    self.dest)
            parser.error(msg)
        assert len(values)>0
        if self._append:
            items = argparse._ensure_value(namespace, self.dest, self._container())
        else:
            items = self._container()
        if (self._container is list):
            items.append(values)
        elif (self._container is tuple):
            items += values
        elif (self._container is set):
            items.update(values)
        else:
            msg='Unknown container type {}.'.format(self._container)
            raise TypeError(msg)
        setattr(namespace, self.dest, items)


class HysopArgParser(argparse.ArgumentParser):
    
    @classmethod
    def set_tmp_dir(cls, tempdir):
        # default tmp dir is in order: 
        # TMPDIR, TEMP, TMP, default system tmp, current working directory
        tempfile.tempdir = tempdir

    @classmethod
    def tmp_dir(cls):
        return tempfile.gettempdir()
    
    @classmethod
    def get_fs_type(cls, path):
        import subprocess
        cmd = ['stat', '-f', '-c', '%T', path]
        fs_type = subprocess.check_output(cmd)
        return fs_type.replace('\n','')

    @classmethod
    def is_shared_fs(cls, path):
        return (cls.get_fs_type(path) in ('nfs',))

    @classmethod
    def set_env(cls, target, value, hysop=True):
        if hysop:
            target = 'HYSOP_{}'.format(target)
        if (value is None):
            pass
        elif (value is True):
            os.environ[target] = '1'
        elif (value is False):
            os.environ[target] = '0'
        elif isinstance(value, str):
            os.environ[target] = value
        elif isinstance(value, int):
            os.environ[target] = str(value)
        else:
            msg='Invalid value of type {}.'.format(type(value))
            raise TypeError(msg)
    
    # custom actions
    split = SplitAppendAction
   
    def _get_formatter(self):
        # try to find out terminal width (max=120 characters)
        try:
            TERMINAL_WIDTH = min(int(os.popen('stty size', 'r').read().split()[1]), 120)
        except:
            TERMINAL_WIDTH = min(os.environ.get('COLUMNS', 80), 120)
        formatter_opts = {'prog':self.prog, 'width':TERMINAL_WIDTH}
        return self.formatter_class(**formatter_opts)

    def __init__(self, prog_name, description, 
            domain=None, default_dump_dir=None, 
            **kwds):
        
        prog = prog_name
        epilog  = colors.color('[ADDITIONAL NOTES]', fg='yellow', style='bold')
        epilog += '\nOpenCL and Autotuner parameters only have an effect on the OpenCL backend.'
        epilog += '\nWhen default arguments are not specified, hysop defaults are used.'

        super(HysopArgParser, self).__init__(prog=prog, description=description, 
                formatter_class=HysopHelpFormatter, epilog=epilog,
                add_help=False, 
                **kwds)
        
        if (domain is None):
            domain = 'box'
        if (default_dump_dir is None):
            default_dump_dir = '{}/hysop/{}'.format(self.tmp_dir(), prog_name)

        self._domain = domain
        
        self._add_positional_args()
        self._add_main_args()
        self._add_domain_args()
        self._add_simu_args()
        self._add_problem_args()
        self._add_method_args()
        self._add_threading_args()
        self._add_opencl_args()
        self._add_autotuner_args()
        self._add_graphical_io_args()
        self._add_file_io_args(default_dump_dir)
        self._add_term_io_args()
        self._add_misc_args()
    
    def parse(self): 
        args = self.parse_args()
        args.__class__ = HysopNamespace
        
        self._check_threading_args(args)

        self._setup_hysop_env(args)
        
        # /!\ only import hysop from there on /!\
        # (env. variables have been correctly set)
        import hysop

        self._check_positional_args(args)
        self._check_main_args(args)
        self._check_domain_args(args)
        self._check_problem_args(args)
        self._check_simu_args(args)
        self._check_method_args(args)
        self._check_opencl_args(args)
        self._check_autotuner_args(args)
        self._check_graphical_io_args(args)
        self._check_file_io_args(args)
        self._check_term_io_args(args)
        self._check_misc_args(args)
        
        self._setup_parameters(args)
        self._setup_implementation(args)
        
        return args

    def run(self, program, **kwds):
        from mpi4py import MPI
        size = MPI.COMM_WORLD.Get_size()
        rank = MPI.COMM_WORLD.Get_rank()
        
        args = self.parse()
        args.stdout = self._fmt_filename(args.stdout, rank, size, args)
        args.stderr = self._fmt_filename(args.stderr, rank, size, args)

        self._rmfile(args.stdout)
        self._rmfile(args.stderr)
        if args.clean:
            dump_dir = '{}/p{}'.format(args.dump_dir, size)
            self._rmfiles(dump_dir, 'h5')
            self._rmfiles(dump_dir, 'xmf')
            self._rmfiles(dump_dir, 'out')
            self._rmfiles(dump_dir, 'txt')
            self._rmfiles(dump_dir, 'png')
            self._rmdir(dump_dir, 'generated_kernels')

        with self.redirect_stdout(rank, args), self.redirect_stderr(rank, args):
                from hysop.tools.contexts import printoptions
                with printoptions(threshold=10000, linewidth=240, 
                                  nanstr='nan', infstr='inf', 
                                  formatter={'float': lambda x: '{:>6.2f}'.format(x)}):
                    program(args, **kwds)

    @staticmethod
    def _fmt_filename(filename, rank, size, args):
        return filename.format(rank=rank, size=size, dpath=args.dump_dir)

    @staticmethod
    def _color_filter(msg):
        return re.sub(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]', "", msg)

    @staticmethod
    def _null_filter(msg):
        return None

    @staticmethod
    def _mkdir(path):
        path = os.path.dirname(os.path.realpath(path))
        try:
            os.makedirs(path)
        except OSError as e:
            if (e.errno != errno.EEXIST):
                raise

    @staticmethod
    def _rmfile(path):
        try:
            os.remove(path)
        except OSError as e:
            if (e.errno != errno.ENOENT):
                raise
    
    @classmethod
    def _rmfiles(cls, path, ext):
        ext='.{}'.format(ext)
        if os.path.exists(path):
            for fname in os.listdir(path):
                if fname.endswith(ext):
                    fname='{}/{}'.format(path, fname)
                    cls._rmfile(fname)
    
    @classmethod
    def _rmdir(cls, path, subdir):
        path += '/' + subdir
        assert path not in ('/', '/tmp')
        if os.path.exists(path):
            msg='Are you sure you want to delete {} y/n ?'
            msg=msg.format(path)
            remove='u'
            while(remove not in ('y','n')):
                remove = raw_input(msg)
            if (remove=='y'):
                shutil.rmtree(path)

    @contextlib.contextmanager
    def redirect_stdout(self, rank, args):
        redirect_to_terminal = (rank in args.tee_ranks)
        
        file_filters = [self._color_filter]
        if redirect_to_terminal:
            stream_filters = []
        else:
            stream_filters = [self._null_filter]
        
        self._mkdir(args.stdout)
        with StdoutTee(args.stdout, mode='a', 
                file_filters=file_filters,
                stream_filters=stream_filters):
            yield
    
    @contextlib.contextmanager
    def redirect_stderr(self, rank, args):
        redirect_to_terminal = (rank in args.tee_ranks)
        
        file_filters = [self._color_filter]
        if redirect_to_terminal:
            stream_filters = []
        else:
            stream_filters = [self._null_filter]
        
        self._mkdir(args.stderr)
        with StderrTee(args.stderr, mode='a', 
                file_filters=file_filters,
                stream_filters=stream_filters):
            yield

    def _add_positional_args(self):
        pass
    
    def _check_positional_args(self, args):
        pass

    def _add_main_args(self):
        args = self.add_argument_group('Main parameters')
        args.add_argument('-impl', '--implementation', type=str, default='python',
                dest='impl',
                help='Backend implementation (either python, fortran or opencl).')
        args.add_argument('-cp', '--compute-precision', type=str, default='fp32',
                dest='compute_precision', 
                help='Floating-point precision used to discretize the parameters and fields.')
        args.add_argument('-ei', '--enforce-implementation', type=str, default='true',
                dest='enforce_implementation',
                help='If set to false, the library may use another implementation than user specified one for some operators.')
        return args

    def _check_main_args(self, args):
        self._check_default(args, ('impl','compute_precision'), str, allow_none=False)
        args.impl = self._convert_implementation('impl', args.impl)
        args.compute_precision = self._convert_precision('compute_precision', 
                                                            args.compute_precision)
        args.dtype = self._precision_to_dtype('compute_precision', args.compute_precision)
        args.enforce_implementation = self._convert_bool('enforce_implementation', args.enforce_implementation)

    def _add_domain_args(self):
        discretization = self.add_argument_group('Discretization parameters')
        discretization.add_argument('-dim', '--ndim', type=int, default=2, 
                dest='ndim', 
                help='Number of dimensions.')
        if (self._domain == 'box'):
            discretization.add_argument('-d', '--discretization', type=str, default=(64,), 
                    action=self.split, container=tuple, convert=int, append=False,
                    dest='npts', 
                    help='Cartesian discretization, number of points in each direction.')
            discretization.add_argument('-bo', '--box-origin', type=str, default=(0.0,), 
                    action=self.split, container=tuple, convert=float, append=False,
                    dest='box_origin',
                    help='Physical coordinates of left-most corner of the domain.')
            discretization.add_argument('-bl', '--box-length', type=str, default=(1.0,), 
                    action=self.split, container=tuple, convert=float, append=False,
                    dest='box_length',
                    help='Physical length of the box.')
            return discretization
        else:
            msg='Unknown domain value {}.'.format(self._domain)
            raise ValueError(msg)
    
    def _check_domain_args(self, args):
        dim = args.ndim
        if (dim<=0):
            msg='Negative or zero dimension, got {}.'.format(dim)
            self.error(msg)
        if (self._domain == 'box'):
            npts       = args.npts
            box_origin  = args.box_origin
            box_length = args.box_length
            self._check_default(args, ('npts', 'box_origin', 'box_length'), tuple)
            if len(npts)==1:
                npts *= dim
            if len(npts)!=dim:
                msg='Discretization should be of the same size as the dimension.'
                msg+='\nGot {} but ndim={}.'.format(npts, dim)
                self.error(msg)
            if any(x<=0 for x in npts):
                msg='Negative discretization encountered.'
                msg+='\nGot {}.'.format(npts)
                self.error(msg)
            if len(box_origin)==1:
                box_origin*=dim
            if len(box_origin)!=dim:
                msg='Box start should be of the same size as the dimension.'
                msg+='\nGot {} but ndim={}.'.format(box_origin, dim)
                self.error(msg)
            if len(box_length)==1:
                box_length*=dim
            if len(box_length)!=dim:
                msg='Box length should be of the same size as the dimension.'
                msg+='\nGot {} but ndim={}.'.format(box_length, dim)
                self.error(msg)
            if any(x<=0 for x in box_length):
                msg='Negative box length encountered.'
                msg+='\nGot {}.'.format(box_length)
                self.error(msg)
            args.npts       = npts
            args.box_origin = box_origin
            args.box_length = box_length
        else:
            msg='Unknown domain value {}.'.format(self._domain)
            raise ValueError(msg)


    def _add_simu_args(self):
        simu = self.add_argument_group('Simulation parameters')
        simu.add_argument('-ts', '--tstart', type=float, default=0.0, 
                dest='tstart',
                help='Set simulation initial time.')
        simu.add_argument('-te', '--tend', type=float, default=10.0, 
                dest='tend',
                help='Set simulation end time.')
        simu.add_argument('-niter', '--nb-iterations', type=int, default=None, 
                dest='nb_iter',
                help='Number of iterations.')
        simu.add_argument('-maxiter', '--max-iterations', type=int, default=None, 
                dest='max_iter',
                help=('Maximal number of iterations at witch the simulation will stop '+
                        '[default=no limit].'))
        simu.add_argument('-dt', '--timestep', type=float, default=None, 
                dest='dt',
                help=('Specify timestep instead of a number of iterations '+
                        '(has priority over number of iterations).'+
                        ' This will be the initial timestep when using adaptive timestep.'))
        simu.add_argument('-cfl', '--cfl', type=float, default=None, 
                dest='cfl',
                help='Specify CFL for adaptive time stepping.')
        simu.add_argument('-lcfl', '--lagrangian-cfl', type=float, default=None, 
                dest='lcfl',
                help='Specify LCFL for adaptive time stepping.')
        simu.add_argument('-dr', '--dry-run', default=False, action='store_true',
                dest='dry_run',
                help='Stop execution before the first simulation iteration.')
        
    def _add_problem_args(self):
        problem = self.add_argument_group('Problem parameters')
        problem.add_argument('-stopi', '--stop-at-initialization', default=False, action='store_true',
                dest='stop_at_initialization',
                help='Stop execution before problem initialization.')
        problem.add_argument('-stopd', '--stop-at-discretization', default=False, action='store_true',
                dest='stop_at_discretization',
                help='Stop execution before problem discretization.')
        problem.add_argument('-stopwp', '--stop-at-work-properties', default=False, action='store_true',
                dest='stop_at_work_properties',
                help='Stop execution before problem work properties retrieval.')
        problem.add_argument('-stopwa', '--stop-at-work-allocation', default=False, action='store_true',
                dest='stop_at_work_allocation',
                help='Stop execution before problem work properties allocation.')
        problem.add_argument('-stops', '--stop-at-setup', default=False, action='store_true',
                dest='stop_at_setup',
                help='Stop execution before problem setup.')
        problem.add_argument('-stopb', '--stop-at-build', default=False, action='store_true',
                dest='stop_at_build',
                help='Stop execution once the problem has been built.')
        return problem

    def _check_problem_args(self, args):
        self._check_default(args, ('stop_at_initialization', 'stop_at_discretization', 'stop_at_setup', 
                                    'stop_at_work_properties', 'stop_at_work_allocation', 'stop_at_build'), bool, allow_none=False)

    def _check_simu_args(self, args):
        self._check_default(args, ('tstart', 'tend'), float)
        self._check_default(args, 'dt', float, allow_none=True)
        self._check_default(args, 'nb_iter', int, allow_none=True)
        self._check_default(args, 'dry_run', bool, allow_none=False)
        self._check_positive(args, 'dt', strict=True, allow_none=True)
        self._check_positive(args, 'nb_iter', strict=True, allow_none=True)
        self._check_positive(args, 'max_iter', strict=True, allow_none=True)
        self._check_positive(args, 'cfl', strict=True, allow_none=True)
        self._check_positive(args, 'lcfl', strict=True, allow_none=True)
        if ((args.nb_iter is not None) and (args.dt is not None)):
            msg=('\nTimestep and number of iterations specified in the same time, '
                 + 'using timestep.')
            warnings.warn(msg, HysopArgumentWarning)
    
    def _add_method_args(self):
        method = self.add_argument_group('Method parameters')
        method.add_argument('-cg', '--compute-granularity', type=int, 
                default=0, 
                dest='compute_granularity',
                help='Set default operator compute granularity.')
        method.add_argument('-fdo', '--finite-differences-order', type=int, 
                default=4, 
                dest='fd_order',
                help='Set default order of finite differences stencils.')
        method.add_argument('-ti', '--time-integrator', type=str, 
                default='RK2', 
                dest='time_integrator',
                help='Set the default time integrator.')
        method.add_argument('-rk', '--remesh-kernel', type=str, 
                default='L4_2', 
                dest='remesh_kernel',
                help='Set the default remeshing formula for advection-remeshing.')
        method.add_argument('-interp', '--interpolation', type=str, 
                default='linear', 
                dest='interpolation',
                help='Set the default interpolation formula to compute subgrid field values.')
        method.add_argument('-sf', '--stretching-formulation', type=str, 
                default='conservative', 
                dest='stretching_formulation',
                help='Set the default stretching formulation.')
        method.add_argument('-rf', '--reprojection-frequency', type=int, 
                default=0, 
                dest='reprojection_frequency',
                help='Set the solenoidal reprojection frequency of the vorticity.')
        method.add_argument('-dso', '--directional-splitting-order', type=int, 
                default=2, 
                dest='strang_order',
                help='Set the default directional splitting order.')
        return method

    def _check_method_args(self, args):
        self._check_default(args, ('compute_granularity', 'fd_order', 'strang_order',
                                    'reprojection_frequency'), int, allow_none=False)
        self._check_default(args, ('time_integrator', 'remesh_kernel', 'interpolation', 
                                'stretching_formulation'), str, allow_none=False)
        self._check_positive(args, 'compute_granularity', strict=False, allow_none=False)
        self._check_positive(args, 'fd_order', strict=True, allow_none=False)
        self._check_positive(args, 'strang_order', strict=True, allow_none=False)
        self._check_positive(args, 'reprojection_frequency', strict=False, allow_none=False)
        
        args.strang_order = self._convert_strang_order('strang_order', 
                args.strang_order)
        args.time_integrator = self._convert_time_integrator('time_integrator', 
                args.time_integrator)
        args.remesh_kernel = self._convert_remesh_kernel('remesh_kernel', 
                args.remesh_kernel)
        args.stretching_formulation = self._convert_stretching_formulation(
                'stretching_formulation', args.stretching_formulation)
        args.interpolation = self._convert_interpolation('interpolation', 
                args.interpolation)
        
    def _add_threading_args(self):
        threading = self.add_argument_group('threading parameters')
        msg =  "Enable threads for backends that supports it (Numba and FFTW) by setting HYSOP_ENABLE_THREADS. "
        msg += "Disabling threading will limit all threading backends to one thread and set numba default backend to 'cpu' instead of 'parallel'."
        threading.add_argument('--enable-threading', type=str, default='1', 
                dest='enable_threading',
                help=msg)
        msg='Set the default maximum usable threads for threading backends (OpenMP, MKL) and operator backends using threads (Numba, FFTW). '
        msg+='This parameter will set HYSOP_MAX_THREADS and does not affect the OpenCL backend.'
        msg+="If this parameter is set to 'physical', the maximum number of threads will be set to the number of physical cores available to the process (taking into account the cpuset). "
        msg+="If set to 'logical', logical cores will be chosen instead. Else this parameter expects a positive integer."
        msg+='If --enable-threads is set to False, this parameter is ignored and HYSOP_MAX_THREADS will be set to 1.'
        threading.add_argument('--max-threads', type=str, default='physical', 
                dest='max_threads',
                help=msg)
        threading.add_argument('--openmp-threads', type=str, default=None, 
                dest='openmp_threads',
                help='This parameter will set OMP_NUM_THREADS to a custom value (overrides --max-threads).')
        threading.add_argument('--mkl-threads', type=str, default=None, 
                dest='mkl_threads',
                help='This parameter will set MKL_NUM_THREADS to a custom value (overrides --max-threads).')
        threading.add_argument('--numba-threads', type=str, default=None, 
                dest='numba_threads',
                help='This parameter will set NUMBA_NUM_THREADS to a custom value (overrides --max-threads).')
        threading.add_argument('--numba-threading-layer', type=str, default='workqueue', 
                dest='numba_threading_layer',
                help="This parameter will set NUMBA_THREADING_LAYER to a custom value ('workqueue' is available on all platforms, but not 'omp' and 'tbb'). Use 'numba -s' to list available numba threading layers.")
        threading.add_argument('--fftw-threads', type=str, default=None, 
              dest='fftw_threads',
              help='This parameter will set HYSOP_FFTW_NUM_THREADS to a custom value (overrides --max-threads).')
        threading.add_argument('--fftw-planner-effort', type=str, default='estimate', 
              dest='fftw_planner_effort',
              help='Set default planning effort for FFTW plans. The actual number of threads used by FFTW may depend on the planning step. This parameter will set HYSOP_FFTW_PLANNER_EFFORT.')
        threading.add_argument('--fftw-planner-timelimit', type=str, default='-1', 
              dest='fftw_planner_timelimit',
              help='Set an approximate upper bound in seconds for FFTW planning. This parameter will set HYSOP_FFTW_PLANNER_TIMELIMIT.')
        return threading


    def _add_opencl_args(self):
        opencl = self.add_argument_group('OpenCL parameters')
        opencl.add_argument('--opencl-platform-id', type=int, default=None, 
                dest='cl_platform_id',
                help=('R|OpenCL platform id for the OpenCL implementation.'
                    +'\nDefaults to hysop.__DEFAULT_PLATFORM_ID__.'))
        opencl.add_argument('--opencl-device-id', type=int, default=None, 
                dest='cl_device_id',
                help=('R|OpenCL device id for the OpenCL implementation.'
                    +'\nDefaults to hysop.__DEFAULT_DEVICE_ID__.'))
        opencl.add_argument('--opencl-device-type', type=str, default=None, 
                dest='cl_device_type',
                help='OpenCL device type for the OpenCL implementation.')
        opencl.add_argument('--opencl-build-options', type=str, default=[], 
                action=self.split, container=list, append=True,
                dest='cl_build_opts', 
                help='Extra OpenCL build options.')
        opencl.add_argument('--opencl-defines', type=int, default=[], 
                action=self.split, container=list, append=True,
                dest='cl_defines',
                help='Extra OpenCL defines.')
        opencl.add_argument('--opencl-fp-precision', type=str, default=None, 
                dest='cl_fp_precision',
                help=('Force OpenCL kernels floating-point precision.'
                  + '\nEither fp16, fp32 or fp64.'))
        opencl.add_argument('--opencl-fp-dump-mode', type=str, default=None, 
                dest='cl_fp_dump_mode',
                help=('OpenCL floating-point dumping format. '+ 
                        'Either dec (decimal) or hex (hexadecimal).'))
        opencl.add_argument('--opencl-enable-short-circuit', type=str, default=None, 
                dest='cl_enable_short_circuit',
                help='Enable short circuits for code-generated OpenCL kernels.')
        opencl.add_argument('--opencl-enable-loop-unrolling', type=str, default=None, 
                dest='cl_enable_loop_unrolling',
                help='Enable loop unrolling for code-generated OpenCL kernels.')
        return opencl
    
    def _check_threading_args(self, args):
        self._check_default(args, ('enable_threading', 'max_threads', 'numba_threading_layer',
                                   'fftw_planner_effort', 'fftw_planner_timelimit'), str, allow_none=False)
        self._check_default(args, ('openmp_threads', 'mkl_threads', 'numba_threads', 'fftw_threads'), 
                            str, allow_none=True)

        args.enable_threading = self._convert_bool('enable_threading', args.enable_threading)
        if args.enable_threading:
            args.max_threads = self._convert_threads('max_threads', args.max_threads, default=None)
        else:
            args.max_threads = 1
        for argname in ('openmp_threads', 'mkl_threads', 'numba_threads', 'fftw_threads'):
            setattr(args, argname, self._convert_threads(argname, getattr(args, argname), 
                default=args.max_threads))
        args.numba_threading_layer = self._convert_numba_threading_layer('numba_threading_layer', 
                args.numba_threading_layer)
        args.fftw_planner_effort = self._convert_fftw_planner_effort('fftw_planner_effort',
                args.fftw_planner_effort)

    def _check_opencl_args(self, args):
        self._check_default(args, ('cl_platform_id', 'cl_device_id'), int, allow_none=True)
        self._check_default(args, ('cl_build_opts','cl_defines'), list, allow_none=False)
        self._check_default(args, ('cl_device_type', 'cl_fp_precision', 'cl_fp_dump_mode'), 
            str, allow_none=True)
        self._check_default(args, ('cl_enable_short_circuit', 'cl_enable_loop_unrolling'), 
            str, allow_none=True)
        
        self._check_positive(args, ('cl_platform_id', 'cl_device_id'), allow_none=True)
        
        args.cl_device_type  = self._convert_device_type('cl_device_type', args.cl_device_type)
        args.cl_fp_precision = self._convert_precision('cl_fp_precision', args.cl_fp_precision)
        args.cl_fp_dump_mode = self._convert_fp_dump_mode('cl_fp_dump_mode', 
                args.cl_fp_dump_mode)
        args.cl_enable_short_circuit = self._convert_bool('cl_enable_short_circuit', 
                                                          args.cl_enable_short_circuit)
        args.cl_enable_loop_unrolling = self._convert_bool('cl_enable_loop_unrolling', 
                                                          args.cl_enable_loop_unrolling)
    
    def _build_opencl_kernel_config(self, args, autotuner_config):
        from hysop.methods import OpenClKernelConfig
        kernel_config = OpenClKernelConfig(autotuner_config=autotuner_config,
                user_build_options=args.cl_build_opts,
                user_size_constants=args.cl_defines,
                precision=args.cl_fp_precision,
                float_dump_mode=args.cl_fp_dump_mode,
                use_short_circuit_ops=args.cl_enable_short_circuit,
                unroll_loops=args.cl_enable_loop_unrolling)
        return kernel_config

    def _add_autotuner_args(self):
        autotuner = self.add_argument_group('Kernel autotuner parameters')
        autotuner.add_argument('--autotuner-flag', type=str, default=None, 
                dest='autotuner_flag', 
                help=('Configure kernel autotuner rigor flag'
                        +' (either estimate,measure,patient or exhaustive).'))
        autotuner.add_argument('--autotuner-nruns', type=int, default=None,
                dest='autotuner_nruns', 
                help='Configure kernel autotuner initial kernel profiling runs.')
        autotuner.add_argument('--autotuner-prune-threshold', type=float, default=None, 
                dest='autotuner_prune_threshold',
                help='Configure kernel autotuner kernel pruning threshold.')
        autotuner.add_argument('--autotuner-max-candidates', type=int, default=None,
                dest='autotuner_max_candidates',
                help='Configure kernel autotuner maximal number of candidate kernels.')
        autotuner.add_argument('--autotuner-verbose', type=int, default=None,
                dest='autotuner_verbose',
                help='Configure kernel autotuner kernel verbosity (0 to 5).')
        autotuner.add_argument('--autotuner-debug', type=bool, default=None, 
                dest='autotuner_debug',
                help='Configure kernel autotuner kernel debug flag.')
        autotuner.add_argument('--autotuner-dump-kernels', type=bool, default=None, 
                dest='autotuner_dump_kernels',
                help='Configure kernel autotuner kernel source dumping.')
        autotuner.add_argument('--autotuner-dump-isolation', type=bool, default=None, 
                dest='autotuner_dump_isolation',
                help='Configure kernel autotuner kernel isolation file generation.')
        autotuner.add_argument('--autotuner-cache-override', type=bool, default=None, 
                dest='autotuner_cache_override',
                help='Override kernel autotuner cached data.')
        autotuner.add_argument('--autotuner-dump-dir', type=str, default=None, 
                dest='autotuner_dump_dir',
                help='Configure kernel autotuner dump directory.')
        return autotuner
    
    def _check_autotuner_args(self, args):
        self._check_default(args, ('autotuner_flag', 'autotuner_dump_dir'), 
                                        str, allow_none=True)
        self._check_default(args, ('autotuner_nruns', 'autotuner_max_candidates', 
                                   'autotuner_verbose'), int, allow_none=True)
        self._check_default(args, ('autotuner_dump_kernels', 'autotuner_dump_isolation'), 
                                        bool, allow_none=True)
        self._check_default(args, 'autotuner_prune_threshold', float, allow_none=True)
        
        self._check_positive(args, ('autotuner_nruns', 'autotuner_max_candidates'),
                                strict=True, allow_none=True)
        self._check_positive(args, 'autotuner_verbose', strict=False, allow_none=True)
        self._check_range(args, 'autotuner_prune_threshold', 1.0, 5.0, allow_none=True)
        self._check_dir(args, 'autotuner_dump_dir', allow_shared=True, allow_none=True)
        
        args.autotuner_flag = self._convert_autotuner_flag('autotuner_flag', 
                args.autotuner_flag)
    
    def _build_autotuner_config(self, args):
        from hysop.methods import OpenClKernelAutotunerConfig
        if (args.autotuner_cache_override is None):
            override_cache = args.override_cache
        else:
            override_cache = args.autotuner_cache_override
        autotuner_config = OpenClKernelAutotunerConfig(
                autotuner_flag=args.autotuner_flag,
                nruns=args.autotuner_nruns,
                prune_threshold=args.autotuner_prune_threshold,
                max_candidates=args.autotuner_max_candidates,
                verbose=args.autotuner_verbose,
                debug=args.autotuner_debug,
                dump_kernels=args.autotuner_dump_kernels,
                generate_isolation_file=args.autotuner_dump_isolation,
                override_cache=override_cache,
                dump_folder=args.autotuner_dump_dir)
        return autotuner_config

    def _add_file_io_args(self, default_dump_dir):
        file_io = self.add_argument_group('File I/O')
        file_io.add_argument('--dump-dir', type=str, default=default_dump_dir, 
                dest='dump_dir',
                help=('HySoP output directory.'
                    +' Overrides HYSOP_DUMP_DIR.'.format(default_dump_dir)))
        file_io.add_argument('--dump-freq', type=int, default=10, 
                dest='dump_freq',
                help=('HDF5 output frequency in terms of iterations.' 
                     +' Use 0 to disable frequency based dumping.'))
        file_io.add_argument('--dump-times', type=str, default=None, convert=float,
                action=self.split, container=tuple, append=False,
                dest='dump_times',
                help='Comma delimited list of additional HDF5 output times of interest.')
        file_io.add_argument('--cache-dir', type=str, default=None, 
                dest='cache_dir',
                help=('Specify an alternative HySoP caching directory.'
                   + ' Overrides HYSOP_CACHE_DIR.'))
        file_io.add_argument('--override-cache', default=None, action='store_true',
                dest='override_cache',
                help='Ignore cached data.')
        file_io.add_argument('-C', '--clean', action='store_true', default=False,
                dest='clean',
                help=('Clean the dump_folder prior to launch '+
                        '(remove all *.txt, *.png, *.xmf and *.h5 files)'))
        file_io.add_argument('--debug-dump-dir', type=str, 
                default='{}/hysop/debug'.format(self.tmp_dir()), 
                dest='debug_dump_dir',
                help=('Target root directory for debug dumps. Debug dumps will appear into <dump dir>/<target>.'))
        file_io.add_argument('--debug-dump-target', type=str, default=None, 
                dest='debug_dump_target',
                help=('Tag for field debug dumps. Debug dumps will appear into <dump dir>/<target>.'))
        return file_io
    
    def _check_file_io_args(self, args):
        self._check_default(args, ('dump_dir', 'cache_dir', 
            'debug_dump_dir', 'debug_dump_target'), str, allow_none=True)
        self._check_default(args, ('dump_freq'), int, allow_none=True)
        self._check_default(args, ('dump_times'), tuple, allow_none=True)
        self._check_default(args, ('override_cache'), bool, allow_none=True)
        self._check_positive(args, 'dump_freq', strict=False, allow_none=False)
        self._check_default(args, 'clean', bool, allow_none=True)
        self._check_dir(args, 'dump_dir', allow_shared=True, allow_none=True)
        self._check_dir(args, 'cache_dir', allow_shared=False, allow_none=True)

        if (args.tstart >= args.tend):
            msg='Invalid time range for the simulation: tstart={}, tend={}.'
            msg=msg.format(args.tstart, args.tend)
            self.error(msg)
        
        if (args.dump_times is not None):
            args.dump_times = set(args.dump_times)
            msg='args.dump_times = {}\n'.format(args.dump_times)
            for dp in args.dump_times:
                if not isinstance(dp, float):
                    msg='Dump time {} is not a float but a {}.'.format(dp, type(dp))
                    self.error(msg)
                elif (dp<args.tstart):
                    msg+='Dump time of interest t={} happens before tstart={}.'
                    msg=msg.format(dp, args.tstart)
                    self.error(msg)
                elif (dp==args.tend):
                    msg+='Dump time of interest t={} happens exactly at tend={}.'
                    msg=msg.format(dp, args.tend)
                    self.error(msg)
                elif (dp>=args.tend):
                    msg+='Dump time of interest t={} happens after tend={}.'
                    msg=msg.format(dp, args.tend)
                    self.error(msg)
    
    def _add_graphical_io_args(self):
        graphical_io = self.add_argument_group('Graphical I/O')
        if 'DISPLAY' in os.environ:
            default_visu_rank = 0
        else:
            default_visu_rank = None
        graphical_io.add_argument('--visu-rank', type=int, default=default_visu_rank, 
                dest='visu_rank',
                help='Specify a MPI rank where the graphical outputs will occur.')
        graphical_io.add_argument('-dg', '--display-graph', action='store_true',
                dest='display_graph',
                help='Stop execution before first simulation iteration.')
        return graphical_io

    def _check_graphical_io_args(self, args):
        if (args.visu_rank<0):
            args.visu_rank = None
        self._check_default(args, 'visu_rank', int, allow_none=True)
        self._check_positive(args, 'visu_rank', strict=False, allow_none=True)
        self._check_default(args, 'display_graph', bool, allow_none=False)

    def _add_term_io_args(self):
        term_io = self.add_argument_group('Terminal I/O')
        msg0=('All {rank} and {size} occurences are replaced by the MPI rank and '
             +'MPI communicator size. {dpath} is replaced by default output path.')
             #+'{program} and {hostname} are replaced by actual program and host name.')
        term_io.add_argument('-stdout', '--std-out', 
                type=str, default='{dpath}/p{size}/{rank}.out', 
                dest='stdout', 
                help='Redirect stdout to this file. ' + msg0)
        term_io.add_argument('-stderr', '--std-err', 
                type=str, default='{dpath}/p{size}/{rank}.out', 
                dest='stderr',
                help='Redirect stderr to this file. '+msg0)
        term_io.add_argument('-tee', '--tee-ranks', type=str, default=(0,), 
                action=self.split, container=tuple, append=False, convert=int,
                dest='tee_ranks',
                help='Tee stdout and stderr of specified MPI ranks to terminal.')
        term_io.add_argument('-V', '--verbose', action='store_true', default=None,
                dest='verbose',
                help='Enable verbosity. Overrides HYSOP_VERBOSE.')
        term_io.add_argument('-D', '--debug', action='store_true', default=None, 
                dest='debug',
                help='Enable debugging informations. Overrides HYSOP_DEBUG.')
        term_io.add_argument('-K', '--kernel-debug', action='store_true', default=None, 
                dest='kernel_debug',
                help='Enable kernel debugging informations. Overrides HYSOP_KERNEL_DEBUG.')
        term_io.add_argument('-P', '--profile', action='store_true', default=None, 
            dest='profile',
            help='Enable profiling informations. Overrides HYSOP_PROFILE.')
        term_io.add_argument('-T', '--trace', type=str, default=None, 
                action=self.split, container=set, append=True,
                dest='trace',
            help=textwrap.dedent('R|'
                   +  'Enable trace informations.'
                   +  '\nComma delimited list of modules to trace:'
                   +  '\n  [calls]      all python calls'
                   +  '\n  [warnings]   warnings traceback' 
                   +  '\n  [memallocs]  memory allocations'
                   +  '\n  [bigallocs]  fat memory allocations'
                   +  '\n  [kernels]    opencl and cuda kernel calls'
                   +  '\n  [all]        enable [warnings,memallocs,kernels]'
                   +  '\nOverrides HYSOP_TRACE_{CALLS,WARNINGS,MEMALLOCS,KERNELS}.'))
        return term_io

    def _check_term_io_args(self, args):
        self._check_default(args, ('stdout','stderr'), str, allow_none=False)
        self._check_default(args, ('tee_ranks'), tuple, allow_none=False)
        self._check_default(args, ('verbose', 'debug', 'profile', 'kernel_debug'),
                            bool, allow_none=True)
        self._check_default(args, 'trace', set, allow_none=True)

    def _add_misc_args(self):
        class _HysopVersion(argparse._VersionAction):
            def __call__(self, parser, namespace, values, option_string=None):
                HysopArgParser.set_env('VERBOSE', False)
                from hysop import version
                version = 'HySoP {}'.format(version)
                formatter = parser._get_formatter()
                formatter.add_text(version)
                parser.exit(message=formatter.format_help())
        class _HwInfo(argparse._StoreAction):
            def __call__(self, parser, namespace, values, option_strings=None):
                HysopArgParser.set_env('VERBOSE', False)
                from hysop.backend.hardware.hwinfo import Topology, PCIIds
                pciids = PCIIds(values)
                topo = Topology.parse(pciids)
                parser.exit(message=str(topo))
        class _HwSummary(argparse._StoreAction):
            def __call__(self, parser, namespace, values, option_strings=None):
                HysopArgParser.set_env('VERBOSE', False)
                from hysop.tools.hysop_ls import run
                args = []
                if (values is not None):
                    args = ['--pci-ids', values]
                run(args)
                parser.exit(message=None)

        misc = self.add_argument_group('Miscellaneous')
        misc.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
                        help='Show this help message and exit')
        misc.add_argument('-v', '--version', action=_HysopVersion,
                        help='Show hysop version and exit.')
        misc.add_argument('-hwinfo', '--hardware-info', action=_HwInfo, 
                        nargs='?', default=None, metavar='pci.ids',
                        help=('R|Show hardware information and exit.'
                           +'\nBench all OpenCL devices on first run.'))
        misc.add_argument('-hwstats', '--hardware-statistics', action=_HwSummary, 
                        nargs='?', default=None, metavar='pci.ids',
                        help=('R|Show localhost hardware statistics and exit.'
                           +'\nBench all OpenCL devices on first run.'))
        return misc

    def _check_misc_args(self, args):
        pass

    def _check_default(self, args, argnames, types, allow_none=False):
        if not isinstance(argnames, tuple):
            argnames = (argnames,)
        assert all(isinstance(a, str) for a in argnames)
        if not isinstance(types, tuple):
            types = (types,)
        assert all(isinstance(a, type) for a in types)
        if allow_none:
            types += (type(None),)
        for argname in argnames:
            if not hasattr(args, argname):
                msg='Unknown argument name \'{}\'.'.format(argname)
                raise ValueError(msg)
            argvalue = getattr(args, argname)
            if not isinstance(argvalue, types):
                msg='\nBad initialization type for argument \'{}\''
                msg+='\n >Expected one of the following type:\n  *{}'
                msg+='\n >Got value {} of type {}.'
                msg=msg.format(argname, '\n  *'.join(x.__name__ for x in types), 
                                    argvalue, type(argvalue).__name__)
                raise TypeError(msg)

    def _check_positive(self, args, argnames, strict=False, allow_none=False):
        if not isinstance(argnames, tuple):
            argnames = (argnames,)
        for argname in argnames:
            if not hasattr(args, argname):
                msg='Unknown argument name \'{}\'.'.format(argname)
                raise ValueError(msg)
            argvalue = getattr(args, argname)
            if (argvalue is None):
                if allow_none:
                    continue
                msg='\nGot invalid None value for parameter \'{}\', positive value expected.'
                msg=msg.format(argname)
                self.error(msg)
            if (argvalue < 0) or (strict and (argvalue == 0)):
                if strict:
                    msg='Strictly positive'
                else:
                    msg='Positive'
                msg+=' value expected for parameter {} but got value {}.'
                msg=msg.format(argname, argvalue)
                self.error(msg)

    def _check_range(self, args, argnames, fmin, fmax, allow_none=False):
        if not isinstance(argnames, tuple):
            argnames = (argnames,)
        for argname in argnames:
            if not hasattr(args, argname):
                msg='Unknown argument name \'{}\'.'.format(argname)
                raise ValueError(msg)
            argvalue = getattr(args, argname)
            if (argvalue is None):
                if allow_none:
                    continue
                msg='Got invalid None value for parameter \'{}\'.'
                msg=msg.format(argname)
                self.error(msg)
            if (fmin is not None) and (fmax is not None) and not (fmin<=argvalue<=fmax):
                msg='Value {} is not in range [{}, {}] for parameter {}.'
                msg=msg.format(argvalue, fmin, fmax, argname)
                self.error(msg)
            elif (fmin is not None) and (argvalue < fmin):
                msg='Value {} is bellow minimum value {} for parameter {}. '
                msg=msg.format(argvalue, fmin, argname)
                self.error(msg)
            elif (fmin is not None) and (argvalue > fmax):
                msg='Value {} is beyond maximal value {} for parameter {}. '
                msg=msg.format(argvalue, fmax, argname)
                self.error(msg)

    def _check_dir(self, args, argnames, allow_shared=False, allow_none=False,
                                                assert_root_folder_exists=True):
        if not isinstance(argnames, tuple):
            argnames = (argnames,)
        for argname in argnames:
            if not hasattr(args, argname):
                msg='Unknown argument name \'{}\'.'.format(argname)
                raise ValueError(msg)
            argvalue = getattr(args, argname)
            if (argvalue is None):
                if allow_none:
                    continue
                msg='Got invalid None value for parameter \'{}\'.'
                msg=msg.format(argname)
                self.error(msg)
        
            path = argvalue
            while not os.path.isdir(path):
                path = os.path.realpath(argvalue+'/..')
                if path == '/':
                    break

            if (not allow_shared) and self.is_shared_fs(path):
                msg='{} directory \'{}\' cannot be stored on a network file system.'
                msg=msg.format(argname, argvalue)
                self.error(msg)
            setattr(args, argname, os.path.realpath(argvalue))

    def _setup_hysop_env(self, args):
        self.set_env('VERBOSE',   args.verbose)
        self.set_env('DEBUG',     args.debug)
        self.set_env('KERNEL_DEBUG', args.kernel_debug)
        self.set_env('PROFILE',   args.profile)
        self.set_env('DUMP_DIR',  args.dump_dir)
        self.set_env('CACHE_DIR', args.cache_dir)

        if (args.trace is not None):
            for module in args.trace: 
                module = module.lower()
                if module == 'calls':
                    self.set_env('TRACE_CALLS', '1')
                elif module == 'bigallocs':
                    self.set_env('BACKTRACE_BIG_MEMALLOCS', '1')
                elif module in ('all', 'warnings'):
                    self.set_env('TRACE_WARNINGS', '1')
                elif module in ('all', 'memallocs'):
                    self.set_env('TRACE_MEMALLOCS', '1')
                elif module in ('all', 'kernels'):
                    self.set_env('TRACE_KERNELS', '1')
                else:
                    msg='Uknown tracing module \'{}\'.'.format(module)
                    self.error(msg)

        self.set_env('ENABLE_THREADING',       args.enable_threading,          True)
        self.set_env('MAX_THREADS',            args.max_threads,               True)
        self.set_env('FFTW_NUM_THREADS',       args.fftw_threads,              True)
        self.set_env('FFTW_PLANNER_EFFORT',    args.fftw_planner_effort,       True)
        self.set_env('FFTW_PLANNER_TIMELIMIT', args.fftw_planner_timelimit,    True)
    
        # those environment variables are not part of HySoP
        self.set_env('OMP_NUM_THREADS',        args.openmp_threads,            False)
        self.set_env('MKL_NUM_THREADS',        args.mkl_threads,               False)
        self.set_env('NUMBA_NUM_THREADS',      args.numba_threads,             False)
        self.set_env('NUMBA_THREADING_LAYER',  args.numba_threading_layer,     False)

    def _setup_parameters(self, args):
        pass

    def _setup_implementation(self, args):
        from hysop.constants import Implementation
        impl = args.impl
        
        autotuner_config = None
        kernel_config    = None

        if impl in (Implementation.PYTHON, Implementation.FORTRAN):
            pass
        elif impl is Implementation.OPENCL:
            # check and configure OpenCL kernel generation and tuning
            from hysop.backend import __HAS_OPENCL_BACKEND__
            if not __HAS_OPENCL_BACKEND__:
                msg='No OpenCL backend found (pyopencl package not found).'
                self.error(msg)
            autotuner_config = self._build_autotuner_config(args)
            kernel_config    = self._build_opencl_kernel_config(args, autotuner_config)
        else:
            msg='Unknown implementation \'{}\'.'.format(impl)
            raise ValueError(msg)
        args.impl = impl
        args.autotuner_config = autotuner_config
        args.opencl_kernel_config = kernel_config
   
    def _check_convert(self, argname, argvalue, values):
        if (argvalue is None):
            return None
        if isinstance(argvalue, str):
            argvalue = argvalue.strip().lower()
        if (argvalue in values):
            return values[argvalue]
        msg='Failed to convert argument {}: {}'.format(argname, argvalue)
        msg+='\nPossible values are:\n  *'
        msg+='\n  *'.join('{}: {}'.format(k, v) for (k,v) in values.iteritems())
        self.error(msg)

    def _convert_bool(self, argname, val):
        values = {
             None:   None,
            'none':  None,
            'false': False,
            'true':  True, 
            '0':     False,
            '1':     True,
        }
        return self._check_convert(argname, val, values)
    
    def _convert_threads(self, argname, val, default):
        if (val == 'physical'):
            val = psutil.cpu_count(logical=False)
        elif (val == 'logical'):
            val = psutil.cpu_count(logical=True)
        elif (val is None):
            if (default is None):
                msg = "'Parameter '{}' has been set to None and no default value has been set."
                msg = msg.format(argname)
                self.error(msg)
            else:
                val = default
        val = int(val)
        if not (val > 0): 
           msg = "'Parameter '{}' has been set to an invalid number of threads {}."
           msg = msg.format(argname, val)
           self.error(msg) 
        return val

    def _convert_fftw_planner_effort(self, argname, val):
        values = {
            'estimate':   'FFTW_ESTIMATE',
            'measure':    'FFTW_MEASURE',
            'patient':    'FFTW_PATIENT',
            'exhaustive': 'FFTW_EXHAUSTIVE',
        }
        return self._check_convert(argname, val, values)
    
    def _convert_numba_threading_layer(self, argname, val):
        values = {
            'workqueue': 'workqueue',
            'omp':       'omp',
            'tbb':       'tbb'
        }
        return self._check_convert(argname, val, values)

    def _convert_implementation(self, argname, impl):
        from hysop.constants import Implementation
        implementations = {
            'python':  Implementation.PYTHON,
            'fortran': Implementation.FORTRAN,
            'cpp':     Implementation.CPP,
            'opencl':  Implementation.OPENCL,
            'cl':      Implementation.OPENCL
        }
        return self._check_convert(argname, impl, implementations)

    def _convert_autotuner_flag(self, argname, flag):
        from hysop.constants import AutotunerFlags
        flags = {
            'estimate':   AutotunerFlags.ESTIMATE,
            'measure':    AutotunerFlags.MEASURE,
            'patient':    AutotunerFlags.PATIENT,
            'exhaustive': AutotunerFlags.EXHAUSTIVE
        }
        return self._check_convert(argname, flag, flags)
    
    def _convert_precision(self, argname, precision):
        from hysop.constants import Precision
        if (precision is None):
            return None
        precisions = {
            'default':      Precision.DEFAULT,
            'same':         Precision.SAME,
            'fp16':         Precision.HALF,
            'fp32':         Precision.FLOAT,
            'fp64':         Precision.DOUBLE,
            'fp80':         Precision.LONG_DOUBLE,
            'fp128':        Precision.QUAD,
            'half':         Precision.HALF,
            'float':        Precision.FLOAT,
            'double':       Precision.DOUBLE,
            'long double':  Precision.LONG_DOUBLE,
            '__float128':   Precision.QUAD,
        }
        return self._check_convert(argname, precision, precisions)
    
    def _precision_to_dtype(self, argname, precision):
        import numpy as np
        from hysop.constants import Precision, HYSOP_REAL
        if (precision is None):
            return None
        precisions = {
            Precision.DEFAULT:     HYSOP_REAL,
            Precision.HALF:        np.float16,
            Precision.FLOAT:       np.float32,
            Precision.DOUBLE:      np.float64,
            Precision.LONG_DOUBLE: np.float128
        }
        return self._check_convert(argname, precision, precisions)

    def _convert_device_type(self, argname, device_type):
        from hysop.constants import DeviceType
        device_types = {
            'all':     DeviceType.ALL,
            'acc':     DeviceType.ACCELERATOR,
            'cpu':     DeviceType.CPU,
            'custom':  DeviceType.CUSTOM,
            'gpu':     DeviceType.GPU,
            'default': DeviceType.DEFAULT
        }
        return self._check_convert(argname, device_type, device_types)

    def _convert_fp_dump_mode(self, argname, fp_dump_mode):
        fp_dump_modes = {
            'decimal':     'dec',
            'hexadecimal': 'hex',
            'dec':         'dec',
            'hex':         'hex'
        }
        return self._check_convert(argname, fp_dump_mode, fp_dump_modes)
    
    def _convert_strang_order(self, argname, strang_order):
        from hysop.methods import StrangOrder
        strang_orders = {
            1: StrangOrder.STRANG_FIRST_ORDER,
            2: StrangOrder.STRANG_SECOND_ORDER,
        }
        return self._check_convert(argname, strang_order, strang_orders)
        
    def _convert_time_integrator(self, argname, time_integrator):
        from hysop.numerics.odesolvers.runge_kutta import Euler, RK2, RK3, RK4
        time_integrators = {
            'euler': Euler,
            'rk1':   Euler,
            'rk2':   RK2,
            'rk3':   RK3,
            'rk4':   RK4
        }
        return self._check_convert(argname, time_integrator, time_integrators)
    
    def _convert_remesh_kernel(self, argname, remesh_kernel):
        from hysop.methods import Remesh
        remesh_kernels = {
            'l2_1':  Remesh.L2_1,
            'l2_2':  Remesh.L2_2,
            'l4_2':  Remesh.L4_2,
            'l4_4':  Remesh.L4_4,
            'l6_4':  Remesh.L6_4,
            'l6_6':  Remesh.L6_6,
            'l8_4':  Remesh.L8_4,
            'l2_1s': Remesh.L2_1s,
            'l2_2s': Remesh.L2_2s,
            'l4_2s': Remesh.L4_2s,
            'l4_4s': Remesh.L4_4s,
            'l6_4s': Remesh.L6_4s,
            'l6_6s': Remesh.L6_6s,
            'l8_4s': Remesh.L8_4s,
        }
        return self._check_convert(argname, remesh_kernel, remesh_kernels)
    
    def _convert_stretching_formulation(self, argname, stretching_formulation):
        from hysop.constants import StretchingFormulation
        stretching_formulations = {
            'grad_uw':       StretchingFormulation.GRAD_UW,
            'grad_uw_t':     StretchingFormulation.GRAD_UW_T,
            'mixed_grad_uw': StretchingFormulation.MIXED_GRAD_UW,
            'conservative':  StretchingFormulation.CONSERVATIVE
        }
        return self._check_convert(argname, stretching_formulation, stretching_formulations)
    
    def _convert_interpolation(self, argname, interpolation):
        from hysop.methods import Interpolation
        interpolations = {
            'linear': Interpolation.LINEAR
        }
        return self._check_convert(argname, interpolation, interpolations)
        
class HysopHelpFormatter(ColorHelpFormatter):
    def _format_args(self, *args, **kwds):
        args = super(HysopHelpFormatter, self)._format_args(*args, **kwds)
        return colors.color(args, fg='green', style='bold')
    
    def _format_action(self, *args, **kwds):
        args = super(HysopHelpFormatter, self)._format_action(*args, **kwds)
        args = args.split(' ')
        args = (colors.color(arg, fg='red', style='bold') if arg and (arg[0]=='-') 
                else arg for arg in args)
        return ' '.join(args)

    def _fill_text(self, text, width, indent):
        return ''.join([indent + line for line in text.splitlines(True)])

    def _format_usage(self, usage, actions, groups, prefix):
        def predicate(action):
            blacklist = ('-tee', '-stdout', '-stderr', '-V', '-D', '-K', '-P', '-T', '-h', 
                    '--version', '--hardware-info', '--hardware-statistics')
            p  = not action.option_strings[0].startswith('--opencl')
            p &= not action.option_strings[0].startswith('--autotuner')
            p &= not action.option_strings[0].startswith('--fftw')
            p &= 'thread' not in action.option_strings[0]
            p &= 'stop' not in action.option_strings[0]
            p &= (action.option_strings[0] not in blacklist)
            p &= (len(action.option_strings)<2) or (action.option_strings[1] not in blacklist)
            return p
        actions = filter(predicate, actions)
        usage = super(HysopHelpFormatter, self)._format_usage(usage=usage, actions=actions,
                groups=groups, prefix=prefix)
        usage = usage.rstrip()
        s=' '*(8+len(self._prog))
        opencl_parameters    = colors.color('OPENCL_PARAMETERS', fg='green', style='bold')
        autotuner_parameters = colors.color('AUTOTUNER_PARAMETERS', fg='green', style='bold')
        trace                = colors.color('TRACE',  fg='green', style='bold')
        tee                  = colors.color('RANKS',  fg='green', style='bold')
        stdout               = colors.color('STDOUT', fg='green', style='bold')
        stderr               = colors.color('STDERR', fg='green', style='bold')

        usage +='\n{s}[--opencl-{{...}} {}]\n{s}[--autotuner-{{...}} {}]'.format(
                opencl_parameters, autotuner_parameters, s=s)
        usage +='\n{s}[-tee {}] [-stdout {}] [-stderr {}]'.format(tee, stdout, stderr, s=s)
        usage +='\n{s}[-V] [-D] [-K] [-P] [-T {}]'.format(trace, s=s)
        usage +='\n{s}[--help] [--version] [--hardware-info] [--hardware-summary]'.format(s=s)
        usage +='\n\n'
        return usage

    def start_section(self, heading):
        heading = colors.color('[{}]'.format(heading.upper()), fg='yellow', style='bold')
        return super(HysopHelpFormatter, self).start_section(heading)

    def _get_help_string(self, action):
        help = action.help
        if '%(default)' not in action.help:
            if action.default not in (argparse.SUPPRESS, None):
                defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                if action.option_strings or action.nargs in defaulting_nargs:
                    help += ' [default: %(default)s]'
        return help

    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()  
        return super(HysopHelpFormatter, self)._split_lines(text, width)


class HysopNamespace(argparse.Namespace):
    pass

