# coding: utf-8

import numpy as np
import sympy as sm


from hysop import Field, Box, Simulation, Problem, \
                  ScalarParameter, MPIParams, CartesianDiscretization, CartesianTopology
from hysop.constants import Implementation, Backend
from hysop.operators import DirectionalAdvection, StrangSplitting, Integrate, \
    AnalyticField, Advection
from hysop.methods import Remesh, TimeIntegrator, ComputeGranularity, \
                          Interpolation, StrangOrder
import hysop.numerics.odesolvers.runge_kutta as rk
from hysop.defaults import TimeParameters

pi = np.pi
dim  = 3

## Function to compute initial scalar values
def init_scalar(data, coords):
    if dim == 3:
        (x,y,z) = coords[0]
        rr = np.sqrt((x - 0.35) ** 2 + (y - 0.35) ** 2 + (z - 0.35) ** 2)
    else:
        (x,y) = coords[0]
        rr = np.sqrt((x - 0.5) ** 2 + (y - 0.75) ** 2)
    data[0][...] = 0.
    data[0][rr < 0.15] = 1.
    rr = np.sqrt((x - 0.75) ** 2 + (y - 0.75) ** 2 + (z - 0.75) ** 2)
    data[0][rr < 0.1] += 1.

# Define domain
npts   = (64,)*dim
npts_s = (64, )*dim
box  = Box(origin=(0.,)*dim, length=(1.,)*dim, dim=dim)
if dim == 3:
    dt0 = 0.35 / (4. * pi)
else:
    dt0 = 0.35 / (2. * pi)
cfl = 1. * dt0 * max(npts)

# Get default MPI Parameters from domain (even for serial jobs)
mpi_params = MPIParams(comm=box.task_comm,
                       task_id=box.current_task())

# Define parameters and field (time and analytic field)
t, dt = TimeParameters(dtype=np.float64)
velo   = Field(domain=box, name='V', is_vector=True, dtype=np.float64)
scalar = Field(domain=box, name='S0', nb_components=1, dtype=np.float64)
vol = ScalarParameter('volume', dtype=np.float64)

# Setup operator method dictionnary
# Advection-Remesh operator discretization parameters
method = {
           TimeIntegrator:      rk.RK2,
           Remesh:              Remesh.L4_2,
           Interpolation:       Interpolation.LINEAR
}

# Create a simulation and solve the problem
# (do not forget to specify the dt parameter here)
simu = Simulation(start=0., end=3.+dt0/2,
                  max_iter=100000,
                  times_of_interest=(0., 3.),
                  t=t, dt=dt, dt0=dt0)
simu.write_parameters(simu.t, vol, filename='volume.txt', precision=8)


# ghosts = (2,)*dim
# d3d    = CartesianDiscretization(npts, ghosts, default_boundaries=True)
# topo = CartesianTopology(domain=box, discretization=d3d, backend=Backend.OPENCL)

# Setup implementation specific variables
#impl = Implementation.PYTHON
impl = Implementation.FORTRAN
#impl = Implementation.OPENCL
extra_op_kwds = { 'mpi_params': mpi_params }
if (impl is Implementation.PYTHON) or (impl is Implementation.FORTRAN):
    sin, cos = np.sin, np.cos
    def compute_velocity(data, coords, t):
        (x,y,z) = coords[0]
        data[0][...] = 2.*sin(pi*x)**2*sin(2.*pi*y)*sin(2.*pi*z)*cos(t()*pi/3.)
        data[1][...] = -sin(2.*pi*x)*sin(pi*y)**2*sin(2.*pi*z)*cos(t()*pi/3.)
        data[2][...] = -sin(2.*pi*x)*sin(2.*pi*y)*sin(pi*z)**2*cos(t()*pi/3.)
    formula = compute_velocity
elif (impl is Implementation.OPENCL):
    sin, cos = sm.sin, sm.cos
    if dim == 3:
        x, y, z = box.frame.coords
        formula = (2.*sin(pi*x)**2*sin(2.*pi*y)*sin(2.*pi*z)*cos(t.s*pi/3.),
                   -sin(2.*pi*x)*sin(pi*y)**2*sin(2.*pi*z)*cos(t.s*pi/3.),
                   -sin(2.*pi*x)*sin(2.*pi*y)*sin(pi*z)**2*cos(t.s*pi/3.))
    else:
        x, y = box.frame.coords
        formula = (-sin(x*pi)**2*sin(y*pi*2)*cos(t.s*pi/3.),
                   sin(y*pi)**2*sin(x*pi*2)*cos(t.s*pi/3.))
else:
    msg='Unknown implementation \'{}\'.'.format(impl)
    raise ValueError(msg)


analytic = AnalyticField(field=velo, formula=formula,
                         variables={velo: npts},
                         extra_input_kwds={'t': simu.t},
                         implementation=Implementation.PYTHON if (impl is Implementation.FORTRAN) else impl,
                         **extra_op_kwds)
# Build the directional advection operator
# here the cfl determines the maximum number of ghosts
# advec = Advection(name='advec',
#                              velocity = velo,
#                              #velocity_cfl = cfl,
#                              advected_fields = (scalar,),
#                              dt = dt,
#                              variables = {velo: npts, scalar: npts_s},
#                              implementation=Implementation.FORTRAN,#relax_min_particles=True,
#                              **extra_op_kwds)
volume = Integrate(name='volume',
                   field=scalar,
                   variables={scalar: npts_s},
                   parameter=vol,
                   implementation=impl,
                   **extra_op_kwds)

if (impl is Implementation.FORTRAN):
    advec = Advection(name='advec',
                      velocity = velo,
                      advected_fields = (scalar,),
                      dt = dt,
                      variables = {velo: npts, scalar: npts_s},
                      implementation=impl,#relax_min_particles=True,
                      **extra_op_kwds)
else:
    advec = DirectionalAdvection(name='advec',
                                 velocity = velo,
                                 velocity_cfl = cfl,
                                 advected_fields = (scalar,),
                                 dt = dt,
                                 variables = {velo: npts, scalar: npts_s},
                                 implementation=impl,#relax_min_particles=True,
                                 **extra_op_kwds)
    # Build the directional splitting operator graph
    splitting = StrangSplitting(splitting_dim=dim,
                                order=StrangOrder.STRANG_SECOND_ORDER)
    splitting.push_operators(advec)

# Create the problem we want to solve and insert our
# directional splitting subgraph.
# Add a writer of input field at given frequency.
problem = Problem(method=method)
problem.insert(analytic)
if (impl is Implementation.FORTRAN):
    #advec.dump_inputs(fields=scalar, filename='S0', frequency=10)
    problem.insert(advec)
    problem.dump_inputs(fields=scalar, filename='S0', frequency=1)
else:
    problem.insert(splitting)
    problem.dump_inputs(fields=scalar, filename='S0', frequency=10)
#problem.insert(volume)
problem.build()

# If a visu_rank was provided, and show_graph was set,
# display the graph on the given process rank.
# if args.display_graph:
#     problem.display(args.visu_rank)

# Initialize discrete velocity and scalar field
dfields = problem.input_discrete_fields
#dfields[velo].initialize(formula=init_velocity)
dfields[scalar].initialize(formula=init_scalar)

# Finally solve the problem
problem.solve(simu)

# Finalize
problem.finalize()
