"""Discrete field defined on device (GPU)
"""
from hysop import __VERBOSE__
from hysop.constants import HYSOP_ORDER, HYSOP_REAL, DirectionLabels
from hysop.fields.discrete_field import DiscreteField
from hysop.backend.device.opencl import cl
from hysop.backend.device.opencl.opencl_kernel import OpenClKernelLauncher, OpenClKernelListLauncher

from hysop.tools.numpywrappers import npw
from hysop.backend.device.opencl.opencl_array_backend import OpenClArrayBackend


class OpenClDiscreteField(DiscreteField):
    """GPU Discrete vector field implementation.
    Allocates OpenCL device memory for the field.
    """
    def __init__(self, cl_env, name, is_vector=False, topology=None):
        """GPU Discrete vector field implementation.
        Allocates OpenCL device memory for the field.

        Parameters
        ----------

        queue : OpenCL queue
        topology : :class:`~hysop.topology.topology.CartesianTopology`, optional
            mpi topology and local meshes info
        is_vector: boolean, optional
            true if parent field is a vector field, default=False
        name : string, optional
            Field name
        """
        # init base class
        super(OpenClDiscreteField, self).__init__(topology, is_vector, name)
        
        # OpenCL environment
        self.cl_env = cl_env
        self.npcl = OpenClArrayBackend(cl_env=cl_env)
        
        # OpenCL arrays
        self.gpu_data = [None] * self.nb_components
        
        # True if device allocations have been done,
        self.gpu_allocated = False
        
        # OpenCL Events list modifying this field
        self.events = []

        # Get the ids of processes involved in the field discretisation.
        # Default = all, otherwise, get info from input topology if given.
        if topology is None:
            from hysop.core.mpi import main_rank
            self._rank = main_rank
        else:
            self._rank = topology.rank
    
    @classmethod
    def from_field(cls, cl_env, vfield):
        if not isinstance(vfield, OpenClDiscreteField):
            vfield.__class__ = cls
            OpenClDiscreteField.__init__(
                vfield, cl_env,
                vfield.topology, vfield.nb_components > 1, vfield.name)

    def allocate(self):
        """Device blocking memory allocations."""
        queue = self.cl_env.default_queue()
        if not self.gpu_allocated:
            self.gpu_data = [ self.npcl.asarray(array) for array in self.data ]
    
    def wait(self):
        """
        Waiting for all events completion in the field list.
        Resets the events list.
        """
        if __VERBOSE__:
            print "{" + str(self._rank) + "}", "Wait events :", self.name
        for e in self.events:
            e.wait()
        self.events = []


    def to_device(self):
        pass

    def to_host(self):
        pass

    def initialize(self):
        pass
    def finalize(self):
        pass
    def get_profiling_info(self):
        pass
