import hysop

try:
    from hysop.f2hysop import fftw2py
except ImportError:
    raise
    msg =  'HySoP fortran fftw bindings are not available for your hysop install.'
    msg += 'Try to recompile HySoP with WITH_FFTW=ON'
    raise ImportError(msg)

from hysop.constants import HYSOP_ORDER, HYSOP_REAL, TranspositionState, BoundaryCondition
from hysop.tools.numpywrappers import npw
from hysop.tools.decorators import debug
from hysop.tools.types import check_instance
from hysop.core.graph.computational_operator import ComputationalGraphOperator
from hysop.topology.cartesian_descriptor import CartesianTopologyDescriptors
from hysop.fields.continuous_field import Field
from hysop.backend.host.fortran.fortran_operator import FortranOperator

class FortranFFTWOperator(FortranOperator):

    @debug
    def __init__(self, input_fields, output_fields, **kwds):
        super(FortranFFTWOperator, self).__init__(input_fields=input_fields,
                output_fields=output_fields, **kwds)

        check_instance(input_fields,  dict, keys=Field, values=CartesianTopologyDescriptors)
        check_instance(output_fields, dict, keys=Field, values=CartesianTopologyDescriptors)

        for f in set(input_fields.keys()+output_fields.keys()):
            for fi in f.fields:
                if (fi.dtype != HYSOP_REAL):
                    msg='FortranFFTW operators only work with HYSOP_REAL precision specified during hysop '
                    msg+='build.'
                    msg+='\nHYSOP_REAL is {} but field {} has dtype {}.'
                    msg=msg.format(HYSOP_REAL.__name__, fi.name, fi.dtype)
                    raise RuntimeError(msg)
                if (any((bd!=BoundaryCondition.PERIODIC) for bd in fi.lboundaries) or
                    any((bd!=BoundaryCondition.PERIODIC) for bd in fi.rboundaries)):
                    msg='FortranFFTW operators only work with PERIODIC boundary conditions:'
                    msg+='\n  operator:    {}'.format(self.name)
                    msg+='\n  field:       {}'.format(fi.pretty_name)
                    msg+='\n  lboundaries: {}'.format(fi.lboundaries)
                    msg+='\n  rboundaries: {}'.format(fi.rboundaries)
                    raise RuntimeError(msg)



        domain = self.input_fields.keys()[0].domain
        self.dim      = domain.dim
        self.domain   = domain

    @debug
    def get_field_requirements(self):
        requirements = super(FortranFFTWOperator, self).get_field_requirements()
        dim = self.domain.dim
        
        # Set can_split to True in all directions except the contiguous one
        # for inputs and outputs.
        for is_input, (field, td, req) in requirements.iter_requirements():
            can_split = req.can_split
            can_split[:-1] = False
            can_split[-1]  = True
            req.can_split = can_split
        return requirements
    
    @debug
    def get_node_requirements(self):
        node_reqs = super(FortranFFTWOperator, self).get_node_requirements()
        node_reqs.enforce_unique_ghosts = True
        return node_reqs

    def handle_topologies(self, input_topology_states, output_topology_states):
        super(FortranFFTWOperator,self).handle_topologies(input_topology_states, output_topology_states)

        topology = self.input_fields.values()[0].topology
        for (field,topoview) in self.input_fields.iteritems():
            assert all(topoview.topology.cart_shape == topology.cart_shape), 'topology mismatch'
        for (field,topoview) in self.output_fields.iteritems():
            assert all(topoview.topology.cart_shape == topology.cart_shape), 'topology mismatch'
        self.topology = topology

    @debug
    def discretize(self):
        if self.discretized:
            return
        super(FortranFFTWOperator,self).discretize()
        topo_view = self.input_discrete_fields.values()[0].topology
        self._fftw_discretize(topo_view)

    @debug
    def get_work_properties(self):
        return super(FortranFFTWOperator,self).get_work_properties()

    @debug
    def setup(self, work=None):
        super(FortranFFTWOperator,self).setup(work=work)

    @debug
    def finalize(self, clean_fftw_solver=False, **kwds):
        super(FortranFFTWOperator,self).finalize(**kwds)
        if clean_fftw_solver:
            fftw2py.clean_fftw_solver(self.dim)

    def _fftw_discretize(self, topo_view):
        """
        fftw specific way to discretize variables for a given
        'reference' resolution.
        It is assumed that in fft case, only one topology shape must be used
        for all variables of all fortran fftw operators.
        """
        self._set_domain_and_tasks()
        topo = topo_view
        comm = self.mpi_params.comm
        size = self.mpi_params.size

        msg = 'input topology is not compliant with fftw.'
        assert topo.cart_dim == 1, msg
        assert topo.cart_shape[-1] == size, msg
        
        global_resolution = topo.global_resolution.astype(npw.int64)
        length = topo.domain.length.astype(HYSOP_REAL)

        local_resolution, global_start = fftw2py.init_fftw_solver(
                global_resolution, length, comm=comm.py2f())

        assert (topo.mesh.local_resolution == local_resolution).all(),'Local resolution mismatch.'
        assert (topo.mesh.global_start     == global_start).all(), 'Global start mismatch.'

    @classmethod
    def supports_mpi(cls):
        return True
