"""
OpenCl backend base interface for fast Fourier Transforms.

:class:`~hysop.numerics.host_fft.HostFFTI`
:class:`~hysop.numerics.host_fft.HostFFTPlanI`
:class:`~hysop.numerics.host_fft.HostFFTQueue`
"""

import psutil
import numpy as np

from hysop import __FFTW_NUM_THREADS__, __FFTW_PLANNER_EFFORT__, __FFTW_PLANNER_TIMELIMIT__
from hysop.tools.types import first_not_None, check_instance
from hysop.backend.host.host_array_backend import HostArrayBackend
from hysop.backend.host.host_array import HostArray
from hysop.numerics.fft.fft import FFTQueueI, FFTPlanI, FFTI


class HostFFTQueue(FFTQueueI):
    """An host fft queue is just a tuple of callable objects."""
    def __init__(self, name):
        self._name = name
        self._plans = ()

    def __iadd__(self, *plans):
        for plan in plans:
            assert callable(plan)
            self._plans += (plan,)
        return self

    def execute(self):
        for plan in self._plans:
            plan()


class HostFFTPlanI(FFTPlanI):
    """
    Tag for FFT plans executing on host backend arrays.
    """
    pass


class HostFFTI(FFTI):
    """
    Abstract base for FFT interfaces targetting Host backends.
    """
    def __init__(self, backend=None, allocator=None, **kwds):
        from hysop.backend.host.host_array_backend import HostArrayBackend
        if (backend is None):
            backend = HostArrayBackend.get_or_create(allocator=allocator)
        if (allocator is not None):
            mdg='Host allocator does not match the one of the backend.'
            assert (backend.allocator is allocator)
        check_instance(backend, HostArrayBackend)
        super(HostFFTI, self).__init__(backend=backend, **kwds)

    @classmethod
    def default_interface(cls, threads=None, 
                       backend=None, allocator=None,
                       planner_effort=None, 
                       planning_timelimit=None,
                       destroy_input=False,
                       warn_on_allocation=True,
                       warn_on_misalignment=True,
                       error_on_allocation=False,
                       **kwds):
        """
        Get the default host FFT interface which is a multithreaded FFTW interface with 
        ESTIMATE planning effort.
        """
        threads            = first_not_None(threads,            __FFTW_NUM_THREADS__)
        planner_effort     = first_not_None(planner_effort,     __FFTW_PLANNER_EFFORT__)
        planning_timelimit = first_not_None(planning_timelimit, __FFTW_PLANNER_TIMELIMIT__)
        from hysop.numerics.fft.fftw_fft import FftwFFT
        return FftwFFT(threads=threads,  
                       planner_effort=planner_effort, 
                       planning_timelimit=planning_timelimit,
                       backend=backend, allocator=allocator,
                       destroy_input=destroy_input,
                       warn_on_allocation=warn_on_allocation,
                       warn_on_misalignment=warn_on_misalignment,
                       error_on_allocation=error_on_allocation,
                       **kwds)
    
    def new_queue(self, tg, name):
        return HostFFTQueue(name=name)
                       
    def plan_copy(self, tg, src, dst):
        src = self.ensure_callable(src)
        dst = self.ensure_callable(dst)
        def exec_copy(src=src, dst=dst):
            dst()[...] = src()
        return exec_copy
    
    def plan_accumulate(self, tg, src, dst):
        src = self.ensure_callable(src)
        dst = self.ensure_callable(dst)
        def exec_copy(src=src, dst=dst):
            dst()[...] += src()
        return exec_copy

    def plan_transpose(self, tg, src, dst, axes):
        src = self.ensure_callable(src)
        dst = self.ensure_callable(dst)
        def exec_transpose(src=src, dst=dst, axes=axes):
            dst()[...] = np.transpose(a=src(), axes=axes)
        return exec_transpose
    
    def plan_fill_zeros(self, tg, a, slices):
        assert slices
        a = self.ensure_callable(a)
        def exec_fill_zeros(a=a, slices=slices):
            buf = a()
            for slc in slices:
                buf[slc] = 0
        return exec_fill_zeros


    @classmethod
    def ensure_callable(cls, get_buffer):
        if callable(get_buffer):
            return get_buffer
        else:
            def get_buf(buf=get_buffer):
                return buf
            return get_buf
