"""Discrete advection for GPU
"""
from hysop import __VERBOSE__, __DEBUG__, __PROFILE__
from hysop.constants import HYSOP_REAL, debug
from hysop.tools.profiler import profile
from hysop.tools.misc import WorkSpaceTools

from hysop.operator.discrete.particle_advection_dir import ParticleAdvectionDir
from hysop.numerics.update_ghosts import UpdateGhostsFull

from hysop.backend.device.opencl import cl
from hysop.backend.device.opencl.opencl_discrete import OpenClDiscreteField
from hysop.backend.device.opencl.gpu_operator import GPUOperator
from hysop.backend.device.opencl.opencl_kernel   import OpenClKernelLauncher
from hysop.backend.device.kernel_autotuner import AutotunerConfig

from hysop.backend.device.codegen.structs.mesh_info import MeshInfoStruct, TranspositionState, MeshDirection

from hysop.methods import TimeIntegrator, Remesh, ExtraArgs, \
        DeviceSupport, Splitting, MultiScale, Interpolation, Precision, \
        StretchingFormulation, DirectionalSplitting, Backend, ExplicitRungeKutta

from hysop.numerics.odesolvers    import Euler, RK2, RK3, RK4
from hysop.numerics.interpolation.interpolation import Linear

from hysop.numerics.remeshing import L2_1, L2_2, L2_3, L2_4
from hysop.numerics.remeshing import L4_2, L4_3, L4_4
from hysop.numerics.remeshing import L6_3, L6_4, L6_5, L6_6
from hysop.numerics.remeshing import L8_4

from hysop.constants import callback_profiler

from hysop.constants import np

class InstanceOf(object):
    def __init__(self, cls):
        self.cls = cls
    def match_instance(self, obj):
        return isinstance(obj,self.cls)


class GPUParticleAdvectionDir(ParticleAdvectionDir, GPUOperator):
    """
    Interface for directionally splitted particle advection solvers on GPU
    """
    
    from abc import ABCMeta, abstractmethod
    __metaclass__ = ABCMeta

    _default_method = {
            TimeIntegrator: RK2, 
            Interpolation:  Linear,
            Backend:        Backend.OPENCL,
            DeviceSupport:  DeviceSupport.DEVICE_GPU,
            Splitting:      DirectionalSplitting.STRANG_FIRST_ORDER,
            Remesh:         L2_1, 
            MultiScale:     L2_1,
            Precision:      HYSOP_REAL,
            ExtraArgs:  {
                'user_src':None,
                'build_options':'',
                'platform_id':None,
                'device_id':None,
                'device_type':None,
                'stretching':None,
                'diffusion':None,
                'split_kernels':True,
                'use_builtin_copy':True,
                'autotuner_config':AutotunerConfig()
            }
        }
    
    _valid_method_values = {
        TimeIntegrator: [Euler,RK2,RK3,RK4,InstanceOf(ExplicitRungeKutta)], 
        Interpolation:  [Linear],
        Remesh:         [L2_1,L2_2,L2_3,L2_4,L4_2,L4_3,L4_4,L6_3,L6_4,L6_5,L6_6,L8_4], 
        DeviceSupport:  [DeviceSupport.DEVICE_GPU], 
        Backend:        [Backend.OPENCL],
        Splitting:      [DirectionalSplitting.STRANG_FIRST_ORDER, 
                         DirectionalSplitting.STRANG_SECOND_ORDER],
        Precision:      [np.float32,np.float64]
    }
    
    @staticmethod
    def check_input_method(user_method={}):
        default_method      = GPUParticleAdvectionDir._default_method
        valid_method_values = GPUParticleAdvectionDir._valid_method_values

        for k in user_method.keys():
            if k not in default_method.keys():
                msg = "Unknown method key '{}', valid keys are:\n\t{}"
                raise ValueError(msg.format(k, default_method.keys()))
            if k==ExtraArgs:
                default_extra_args = default_method[ExtraArgs]
                user_extra_args    = user_method[ExtraArgs]
                vke = default_extra_args.keys()
                for ke in user_extra_args.keys():
                    if ke not in vke:
                        msg="WARNING: Unknown extra arg '{}', ' \
                            + 'valid values are:\n\t{}".format(ke,vke)
                        print msg
            elif (valid_method_values[k] is not None):
                valid_vals = [v for v in valid_method_values[k]     if not isinstance(v,InstanceOf)]
                valid_cls  = [v.cls for v in valid_method_values[k] if     isinstance(v,InstanceOf)]
                if (user_method[k] not in valid_vals \
                        and user_method[k].__class__ not in valid_cls):
                    pass
                    msg = "Unknown method value '{}', valid values for key '{}' are:\n\t{}"
                    raise ValueError(msg.format(user_method[k],k,valid_method_values[k]))
        
        method = default_method.copy()
        method.update(user_method)
        
        method[ExtraArgs] = default_method[ExtraArgs].copy()
        if ExtraArgs in user_method.keys():
            method[ExtraArgs].update(user_method[ExtraArgs])
        
        if __DEBUG__:
            print "Checked input method for ParticleAdvection:"
            for k,v in method.iteritems():
                print "\t{}: {}".format(k,v)
            print

        return method

    @debug
    def __init__(self, **kwds):
        """Particular advection of field(s) in a given direction,
        on GPU.

        OpenCL kernels are build once per dimension in order to handle
        directional splitting with resolution non uniform in directions.

        Parameters
        ----------
        kwds : 
            Base classes parameters

        Note
        ----
        * warning : this operator is derived from ParticleAdvectionDir AND
        GPUOperator. kwds must then handle argument of both classes.
        """

        # Set or overwrite default method by user specified one
        if 'method' in kwds:
            method = self.check_input_method(kwds['method'])
        else:
            method = self.check_input_method()
        kwds['method'] = method
        
        # Kernel configuration
        extra_args = method[ExtraArgs]
        self._split_kernels    = extra_args.pop('split_kernels')
        self._user_src         = extra_args.pop('user_src')
        self._build_options    = extra_args.pop('build_options')
        self._use_builtin_copy = extra_args.pop('use_builtin_copy')
        self._autotuner_config = extra_args.pop('autotuner_config')
    
        stretching = extra_args.pop('stretching')
        if (stretching is not None):
            self._has_stretching = True
            self._stretching = stretching
        else:
            self._has_stretching = False

        diffusion  = extra_args.pop('diffusion')
        if (diffusion is not None):
            self._has_diffusion = True
            self._diffusion = diffusion
        else:
            self._has_diffusion = False

        super(GPUParticleAdvectionDir, self).__init__(**kwds)
        self.fields_on_grid = self.advected_fields
        
        self._synchronize = None
        if self._is_multi_scale:
            self._synchronize = UpdateGhostsFull(
                self.velocity_topo, self.velocity.nb_components)
        
        # Initialize base classes
        GPUOperator.__init__(self,
            platform_id = extra_args.pop('platform_id'),
            device_id   = extra_args.pop('device_id'),
            device_type = extra_args.pop('device_type'),
            **kwds)

        if extra_args:
             raise ValueError('Some extra arguments were not used:\n\t{}\n'.format(extra_args))
        
        
    def _check(self):
        super(GPUParticleAdvectionDir,self)._check()
        cls = self.__class__.__name__
        if (self._split_kernels)  and (not self.supports_kernel_splitting()):
            msg='Kernel splitting has not been implemented in {} yet !'.format(cls)
            raise NotImplementedError(msg)
        if (self._user_src is not None) and (not self.supports_user_kernels()):
            msg='{} does not support user provided kernels.'.format(cls)
            raise NotImplementedError(msg)
        if (self._has_stretching)  and (not self.supports_stretching()):
            msg='Stretching has not been implemented in {} yet !'.format(cls)
            raise NotImplementedError(msg)
        if (self._has_diffusion) and (not self.supports_diffusion()):
            msg='Diffusion has not been implemented in {} yet !'.format(cls)
            raise NotImplementedError(msg)

    def _initialize(self):
        super(GPUParticleAdvectionDir,self)._initialize()
        self._configure()
        self._initialize_cl_env()
    
    def _setup(self):
        self._device_buffer_allocations()
        
        self._collect_copy_kernels()
        self._collect_transposition_kernels()
        self._collect_advection_remesh_kernels()
        self._collect_user_kernels()
        self._collect_extra_kernels()

        self._device_buffer_initializations()
        
        self._build_exec_list()
        self._initialize_events()
        
    
    def _configure(self):
        """
        Compute resolutions for kernels for each direction.
         Resolution of the local mesh but reoganized regarding
         splitting direction (see hysop.gpu.gpu_operator.GPUOperator._reorder_vect).
           direction X : XYZ
           direction Y : YXZ
           direction Z : ZYX in parallel, ZXY in sequential.
        """

        direction=self.direction
        
        mesh_dir = MeshDirection[direction]
        if direction==MeshDirection.X:
            mesh_state = TranspositionState.XYZ
        elif direction==MeshDirection.Y:
            mesh_state = TranspositionState.YXZ
        elif direction==MeshDirection.Z:
            mesh_state = TranspositionState.ZYX if (self._is_distributed) else TranspositionState.ZXY
        else:
            raise ValueError()

        f_resol     = self.fields_topo.mesh.local_resolution
        v_resol     = self.velocity_topo.mesh.local_resolution
        f_resol_dir = self._reorder_vect(f_resol)
        v_resol_dir = self._reorder_vect(v_resol)
       
        self.mesh_dir    = mesh_dir
        self.mesh_state  = mesh_state

        self.f_resol     = f_resol
        self.v_resol     = v_resol
        self.f_resol_dir = f_resol_dir
        self.v_resol_dir = v_resol_dir
    
    def _collect_copy_kernels(self):
        self.copy = OpenClKernelLauncher(cl.enqueue_copy,
                                   self.cl_env.queue)
    
    def _collect_transposition_kernels(self):
        """
        Compile OpenCL sources for transposition kernels.
           2D: XY -> YX -> XY
           3D: 
            *sequential: XYZ -> YXZ -> ZXY -> YXZ -> XYZ
            *parallel:   XYZ -> YXZ -> ZYX -> YXZ -> XYZ
        Kernels are only collected if the dimension and direction match.
        """

        dim       = self.dim
        direction = self.direction
        
        smethod = self.method[Splitting]

        is_XY_needed, is_YX_needed = False, False
        is_XZ_needed, is_ZX_needed = False, False
        
        if dim==1:
            pass
        elif dim==2:
            if smethod==DirectionalSplitting.STRANG_FIRST_ORDER:
                is_XY_needed = (dim>1) and (direction == MeshDirection.X)
                is_YX_needed = (dim>1) and (direction == MeshDirection.Y)
            elif smethod==DirectionalSplitting.STRANG_SECOND_ORDEVR:
                is_XY_needed = (dim>1) and (direction == MeshDirection.X)
                is_YX_needed = (dim>1) and (direction == MeshDirection.Y)
        elif dim==3:
            if smethod==DirectionalSplitting.STRANG_FIRST_ORDER:
                is_XY_needed = (dim>1) and (direction == MeshDirection.X)
                is_YX_needed = (dim>1) and (direction == MeshDirection.Z)
                is_XZ_needed = (dim>2) and (direction == MeshDirection.Y) \
                        and (not self._is_distributed)
                is_ZX_needed = (dim>2) and (direction == MeshDirection.Z) \
                        and (not self._is_distributed)
            elif smethod==DirectionalSplitting.STRANG_SECOND_ORDER:
                is_XY_needed = (dim>1) and (direction == MeshDirection.X)
                is_YX_needed = (dim>1) and (direction == MeshDirection.Y)

                is_XZ_needed = (dim>2) and (direction == MeshDirection.Y) \
                        and (not self._is_distributed)
                is_ZX_needed = (dim>2) and (direction == MeshDirection.Z) \
                        and (not self._is_distributed)
        else:
            msg = 'Dimension {} not implemented yet!'.format(dim)
            raise NotImplementedError(msg)
        
        
        transpose_xy, transpose_yx = None, None
        transpose_zx, transpose_xz = None, None

        if is_XY_needed:
            transpose_xy = self._collect_transposition_kernel_xy()
        if is_YX_needed:
            transpose_yx = self._collect_transposition_kernel_yx()
        if is_XZ_needed:
            transpose_xz = self._collect_transposition_kernel_xz()
        if is_ZX_needed:
            transpose_zx = self._collect_transposition_kernel_zx()

        self._build_tranpose_funcs(transpose_xy, transpose_yx, transpose_xz, transpose_zx)
    
    def _collect_advection_remesh_kernels(self):
        self._advec, self._remesh = None, None
        self._advec_and_remesh    = None

        if self._split_kernels:
            self._collect_advec_kernel()
            self._collect_remesh_kernel()
        else:
            self._collect_advec_remesh_kernel()
        
        # point _do_compute_impl function to the right implementation
        if self._is_distributed:
            if self._split_kernels:
                if self._is_multi_scale:
                    impl = self._do_compute_2k_multiscale_comm
                else:
                    impl = self._do_compute_2k_monoscale_comm
            else:
                if self._is_multi_scale:
                    impl = self._do_compute_1k_multiscale_comm
                else:
                    impl = self._do_compute_1k_monoscale_comm
        else:
            if self._split_kernels:
                if self._is_multi_scale:
                    impl = self._do_compute_2k_multiscale
                else:
                    impl = self._do_compute_2k_monoscale
            else:
                if self._is_multi_scale:
                    impl = self._do_compute_1k_multiscale
                else:
                    impl = self._do_compute_1k_monoscale
        self._do_compute_impl = impl


    def _do_sync(self,simulation,**kargs):
        if self._is_multi_scale:
            if __VERBOSE__:
                print '_do_sync'
            self._synchronize(self.velocity.data)
            self.velocity.to_device()

    def _do_compute(self,simulation,dt_coeff,**kargs):
        if __VERBOSE__:
            print '_do_compute'
        dt = simulation.time_step * dt_coeff
        self._do_compute_impl(dt=dt,**kargs)

    
    def _device_buffer_allocations(self):
        """Allocate OpenCL buffers for velocity and advected field.
        """
        if (self._rwork is None) or (self._iwork is None):
            raise RuntimeError('_set_work_arrays has not been implemented properly.')
        
        # Velocity
        OpenClDiscreteField.fromField(self.cl_env, self.velocity,
                                   self.gpu_precision, simple_layout=False)
        if self.velocity.allocate():
            self.size_global_alloc += self.velocity.mem_size
        
        # Fields on grids
        for fg in self.fields_on_grid:
            OpenClDiscreteField.fromField(self.cl_env,
                                       fg,
                                       self.gpu_precision,
                                       layout=False)
            if fg.allocate():
                self.size_global_alloc += fg.mem_size

        # Fields on particles
        self.fields_on_part = {}
        start = 0
        for fg in self.fields_on_grid:
            for i in xrange(start, start + fg.nb_components):
                if type(self._rwork[i]) is np.ndarray:
                    self._rwork[i] = \
                        self.cl_env.global_allocation(self._rwork[i])
            self.fields_on_part[fg] = self._rwork[start: start + fg.nb_components]
            start += fg.nb_components

        # Particles positions (only used if advection and remesh kernels are splitted)
        if self._split_kernels:
            if type(self._rwork[start]) is np.ndarray:
                self._rwork[start] = \
                    self.cl_env.global_allocation(self._rwork[start])
            self._particle_position = self._rwork[start]
            start += 1
            
        if start != len(self._rwork):
            raise RuntimeError('GPU allocation error!')
    
    def _device_buffer_initializations(self):
        """
        OpenCL buffer initializations from user OpenCL kernels.
        Looking for kernels named init<FieldName>.
        """
        if (self._user_prg is None) or (self.direction != MeshDirection.X):
            return
        
        f_resolution = self.f_resolution
        v_resolution = self.v_resolution

        for gpudf in self.variables:
            #Looking for init_{field_name} kernel
            match = 'init' + '_'.join(gpudf.name.split('_')[:-1])
            for k in self.prg.all_kernels():
                k_name = k.get_info(cl.kernel_info.FUNCTION_NAME)
                if match.find(k_name) >= 0:
                    if __VERBOSE__:
                        print gpudf.name, '-> OpenCL Kernel', k_name
                    if gpudf == self.velocity:
                        workItemNumber, gwi, lwi = \
                            self.cl_env.get_work_items(v_resolution)
                    else:
                        workItemNumber, gwi, lwi = \
                            self.cl_env.get_work_items(f_resolution)
                    
                    init_field_kernel = cl.Kernel(self.prg, k_name)
                    gpudf.setInitializationKernel(
                        OpenClKernelLauncher(init_field_kernel,self.cl_env.queue,gwi,lwi)
                    )
    
   
    
    def _build_exec_list(self):
        """Prepare GPU kernels sequence
         Build execution list regarding splitting:
            
            **Splitting Strang 1st order**
               1D: X(dt)
               2D: X(dt), Y(dt)
               3D: X(dt), Y(dt), Z(dt)
             
            **Splitting Strang 2nd order**
               1D: X(dt)
               2D: X(dt/2), Y(dt), X(dt/2)
               3D: X(dt/2), Y(dt/2), Z(dt), Y(dt/2), X(dt/2)

        Parameter list is passed to kernel callers but can be completed
        or overrided during this operator apply call.
        """
        
        dim = self.dim
        smethod = self.method[Splitting]
                
        dim_msg    = 'Unsupported dimension {} for splitting method {}.'.format(dim,smethod)
        method_msg = 'Splitting method {} not implemeted yet on GPU.'.format(smethod)

        param_list,exec_list = None,None

        if smethod == DirectionalSplitting.STRANG_FIRST_ORDER:
            if self.dim == 1:
                param_list = [
                    {'dir':MeshDirection.X, 'dt_coeff':1.0},
                ]
                exec_list = [
                    [self._do_sync, self._do_copy, self._do_compute], # X(dt)
                ]
            elif dim == 2:
                param_list = [
                    {'dir':MeshDirection.X, 'dt_coeff':1.0},
                    {'dir':MeshDirection.Y, 'dt_coeff':1.0},
                ]
                exec_list = [
                    [self._do_sync, self._do_copy, self._do_compute, self._do_transpose_xy_g2p], # X(dt)
                    [                              self._do_compute, self._do_transpose_yx_g2g], # Y(dt)
                ]
            elif dim == 3:
                param_list = [
                    {'dir':MeshDirection.X, 'dt_coeff':1.0},
                    {'dir':MeshDirection.Y, 'dt_coeff':1.0},
                    {'dir':MeshDirection.Z, 'dt_coeff':1.0},
                ]
                exec_list = [
                    [self._do_sync, self._do_copy, self._do_compute, self._do_transpose_xy_g2p], # X(dt)
                    [                              self._do_compute, self._do_transpose_xz_g2p], # Y(dt)
                    [                              self._do_compute, self._do_transpose_zx_g2p,  \
                                                                     self._do_transpose_yx_p2g] # Z(dt)
                ]
            else:
                raise ValueError(dim_msg)
        elif smethod == DirectionalSplitting.STRANG_SECOND_ORDER:
            if self.dim == 1:
                param_list = [
                    {'dir':MeshDirection.X, 'dt_coeff':1.0},
                ]
                exec_list = [
                    [self._do_sync, self._do_copy, self._do_compute], # X(dt)
                ]
            elif dim == 2:
                param_list = [
                    {'dir':MeshDirection.X, 'dt_coeff':0.5},
                    {'dir':MeshDirection.Y, 'dt_coeff':1.0},
                    {'dir':MeshDirection.X, 'dt_coeff':0.5},
                ]
                exec_list = [
                    [self._do_sync, self._do_copy, self._do_compute, self._do_transpose_xy_g2p], # X(dt/2)
                    [                              self._do_compute, self._do_transpose_yx_g2p], # Y(dt)
                    [                              self._do_compute]                             # X(dt/2)
                ]
            elif dim == 3:
                param_list = [
                    {'dir':MeshDirection.X, 'dt_coeff':0.5},
                    {'dir':MeshDirection.Y, 'dt_coeff':0.5},
                    {'dir':MeshDirection.Z, 'dt_coeff':1.0},
                    {'dir':MeshDirection.Y, 'dt_coeff':0.5},
                    {'dir':MeshDirection.X, 'dt_coeff':0.5},
                ]
                exec_list = [
                    [self._do_sync, self._do_copy, self._do_compute, self._do_transpose_xy_g2p], # X(dt/2)
                    [                              self._do_compute, self._do_transpose_xz_g2p], # Y(dt/2)
                    [                              self._do_compute, self._do_transpose_zx_g2p], # Z(dt)
                    [                              self._do_compute, self._do_transpose_yx_g2p], # Y(dt/2)
                    [                              self._do_compute]                             # X(dt/2)
                ]
            else:
                raise ValueError(dim_msg)
        else:
            raise ValueError(method_msg)

        if (len(exec_list)!=len(param_list)):
            raise ValueError('param_list and exec_list length mismatch!')
        
        self._param_list = param_list
        self._exec_list  = exec_list

    def _initialize_events(self):
        # Particle initialisation OpenCL events for each field
        pass


    def step_directions(self):
        dirs = []
        for p in self._param_list:
            dirs.append(p['dir'])
        return dirs
    
## 
## Discrete operator interface
##
    def get_work_properties(self):
        # Shape of reference comes from fields, not from velocity
        res = self.f_resol
        
        iw,rw = 0,0
        for f in self.fields_on_grid:
            rw += f.nb_components
        if self._split_kernels:
            rw += 1  # work array for particles positions

        work = {}
        work['rwork'] = np.asarray((res,)*rw)
        work['iwork'] = np.asarray((res,)*iw)
        return work

    def setup(self, rwork, iwork):
        self._rwork = rwork
        self._iwork = iwork
        self._setup()
    
    @debug
    @profile
    def apply(self, simulation, step_id, **extra_params):
        """
        Apply operator along specified splitting direction.
        @param simulation : Simulation to which his operator applies
        @param step_id    : Splitting step id
        @param extra_params: Extra kernel parameters that complete or override 
            step parameters param_list[step_id].
        """

        if (simulation is None):
            raise ValueError('simulation is None.')
        if (step_id<0) or (step_id>=len(self._exec_list)): 
            raise ValueError('step_id is out of bounds.')

        # in first step, synchronize previous work on gpu fields and clean events
        if step_id == 0:
            for v in self.variables:
                v.clean_events()
        
        self._pre_apply()
    
        import copy
        parameters = copy.deepcopy(self._param_list[step_id])
        parameters.update(extra_params)
       
        my_dir  = self.direction
        req_dir = parameters['dir']
        if my_dir != req_dir:
            msg = 'Discrete splitting operator called for wrong direction, \
splitting step id {} requires direction {} but currently applied operator \
direction is {}!'.format(step_id,MeshDirection[req_dir],MeshDirection[my_dir])
            raise RuntimeError(msg)

        exec_list = self._exec_list[step_id]
        for exe in exec_list:
            exe(simulation=simulation, **parameters)
    
    def get_profiling_info(self):
        for exe in [self.copy, 
                  self.transpose_xy, self.transpose_xz,
                  self.transpose_xz, self.transpose_zx,
                  self._advec_and_remesh,
                  self._advec, self._remesh]:
            if exe is not None:
                if isinstance(exe,dict):
                    exe = exe.values()
                if not isinstance(exe,list):
                    exe = [exe]

                for k in exe:
                    for p in k.profile:
                        self.profiler += p
                
    
    @debug
    def finalize(self):
         for work in self._rwork:
             self.cl_env.global_deallocation(work)
         for work in self._iwork:
             self.cl_env.global_deallocation(work)


##  
## OpenCL kernel interface
##
    @abstractmethod
    def _initialize_cl_env(self):
        pass
    
    def _collect_transposition_kernel_xy(self):
        raise NotImplementedError()
    def _collect_transposition_kernel_yx(self):
        raise NotImplementedError()
    def _collect_transposition_kernel_xz(self):
        raise NotImplementedError()
    def _collect_transposition_kernel_zx(self):
        raise NotImplementedError()

    def _collect_advec_remesh_kernel():
        raise NotImplementedError()
    def _collect_advec_kernel():
        raise NotImplementedError()
    def _collect_remesh_kernel():
        raise NotImplementedError()

    def _collect_user_kernels():
        if self.supports_user_kernels():
            raise NotImplementedError()
        else:
            self._user_prg = None

    def _collect_extra_kernels(self):
        if self._has_stretching or self._has_diffusion:
            raise NotImplementedError()
        else: 
            pass
    
    def _do_copy(self,**kargs):
        raise NotImplementedError()
    
    def _do_compute_1k_monoscale(self, dt):
        raise NotImplementedError()
    def _do_compute_1k_multiscale(self, dt):
        raise NotImplementedError()
    def _do_compute_2k_monoscale(self, dt):
        raise NotImplementedError()
    def _do_compute_2k_multiscale(self, dt):
        raise NotImplementedError()

    def _do_compute_1k_monoscale_comm(self, dt):
        raise NotImplementedError()
    def _do_compute_1k_multiscale_comm(self, dt):
        raise NotImplementedError()
    def _do_compute_2k_monoscale_comm(self, dt):
        raise NotImplementedError()
    def _do_compute_2k_multiscale_comm(self, dt):
        raise NotImplementedError()

    def _pre_apply(self):
        pass
    
    @staticmethod
    def supports_kernel_splitting():
        return False
    @staticmethod
    def supports_user_kernels():
        return False
    @staticmethod
    def supports_stretching():
        return False
    @staticmethod
    def supports_diffusion():
        return False
    
    
    def _exec_transpose(self,kernel,mode):
        velocity = self.velocity
        all_evts = []
        evts = []
        for Vi in velocity.gpu_data:
            evt = kernel(Vi,Vi,wait_for=velocity.events)
            evts.append(evt)
        velocity.events += evts
        all_evts+=evts

        for (fg,fp) in self.fields_on_part.iteritems():
            evts = []
            for (g,p) in zip(fg.gpu_data,fp):
                if mode=='g2g':
                    evt = kernel(g,g,wait_for=fg.events)
                elif mode=='g2p':
                    evt = kernel(g,p,wait_for=fg.events)
                elif mode=='p2g':
                    evt = kernel(p,g,wait_for=fg.events)
                else:
                    raise ValueError('Unknown mode \'{}\'.'.format(mode))
                evts.append(evt)
            fg.events += evts
            all_evts  += evts
        return all_evts

    def _build_tranpose_funcs(self,Txy,Tyx,Txz,Tzx):
        
        def build_none(fname):
            def transposition_kernel(**kargs):
                msg='The transposition kernel {} has not been collected.'
                raise RuntimeError(msg.format(fname))
            return transposition_kernel
        
        def build_func(fprefix,fname,src2dst,kernel):
            def transposition_kernel(**kargs):
                if __VERBOSE__:
                    print fname
                callback_profiler.tic(fprefix)
                evts = self._exec_transpose(kernel,src2dst)
                callback_profiler.tac(fprefix,evt=evts)
                return evts
            return transposition_kernel
            
        axes    = ['xy','yx','xz','zx']
        kernels = [Txy,Tyx,Txz,Tzx]
        for axe,kernel in zip(axes,kernels):
            fprefix = 'do_transpose_{}'.format(axe)
            for src2dst in ['g2g','g2p','p2g']:
                fname = '_{}_{}'.format(fprefix,src2dst)
                if kernel is None:
                    func = build_none(fname)
                else:
                    func = build_func(fprefix,fname,src2dst,kernel)
                    callback_profiler.register_tasks(fprefix)
                setattr(self,fname,func)
       
   
    
    
