
from hysop.constants import __DEBUG__, __VERBOSE__, __KERNEL_DEBUG__, \
    DEFAULT_AUTOTUNER_FLAG, DEFAULT_AUTOTUNER_PRUNE_THRESHOLD
from hysop.tools.io_utils import IO
from hysop.tools.misc import Utils

from hysop.constants import np, AutotunerFlags
from hysop.backend.device.opencl import cl, KERNEL_DUMP_FOLDER

import os, itertools, hashlib, gzip, pickle

class KernelGenerationError(RuntimeError):
    pass
    
class OpenClKernelStatistics(object):
    """Execution statistics from kernel events.
    """
    def __init__(self,nruns=0,events=None):
        if events is not None:
            p0 = events[0].profile
            t0 = p0.end - p0.start
            total = 0
            maxi = t0 
            mini = t0
            for evt in events:
                dt = evt.profile.end - evt.profile.start
                total += dt
                if dt<mini: 
                    mini = dt
                if dt>maxi:
                    maxi = dt
            
            self.tot = total
            self.min = mini
            self.max = maxi
            self.mean = total/len(events)
        else:
            self.tot  = None
            self.min  = None
            self.max  = None
            self.mean = None
        self.nruns = nruns

    def __str__(self):
        mini  = self.min   * 1e-6
        maxi  = self.max   * 1e-6
        total = self.tot   * 1e-6
        mean  = self.mean  * 1e-6
        return 'min={:.2f}ms, max={:.2f}ms, mean={:.2f}ms (nruns={})'.format(mini,maxi,mean,self.nruns)

    @staticmethod
    def cmp(lhs,rhs):
        if lhs.mean==rhs.mean:
            if lhs.max==rhs.max:
                if lhs.min==rhs.min:
                    return 0
                elif lhs.min>rhs.min:
                    return 1
                else:
                    return -1
            elif lhs.max > rhs.max:
                return 1
            else:
                return -1
        elif lhs.mean>rhs.mean:
            return 1
        else:
            return -1
    def __lt__(self, other):
        return self.cmp(self, other) < 0
    def __gt__(self, other):
        return self.cmp(self, other) > 0
    def __eq__(self, other):
        return self.cmp(self, other) == 0
    def __le__(self, other):
        return self.cmp(self, other) <= 0
    def __ge__(self, other):
        return self.cmp(self, other) >= 0
    def __ne__(self, other):
        return self.cmp(self, other) != 0

class AutotunerConfig(object):

    _default_initial_runs = {
        AutotunerFlags.ESTIMATE: 1,
        AutotunerFlags.MEASURE: 2,
        AutotunerFlags.PATIENT: 4,
        AutotunerFlags.EXHAUSTIVE: 8
    } 

    def __init__(self, 
            autotuner_flag=DEFAULT_AUTOTUNER_FLAG,
            prune_threshold=DEFAULT_AUTOTUNER_PRUNE_THRESHOLD, 
            verbose=__VERBOSE__,
            debug=__KERNEL_DEBUG__,
            dump_folder=KERNEL_DUMP_FOLDER,
            override_cache=False,
            nruns=None):

        self.autotuner_flag  = autotuner_flag
        self.prune_threshold = prune_threshold
        self.verbose = verbose
        self.debug   = debug
        self.override_cache  = override_cache
        if (nruns is None):
            self.nruns = AutotunerConfig._default_initial_runs[autotuner_flag]
        else:
            if (nruns<1):
                raise ValueError('nruns<1.')
            self.nruns = nruns

class KernelAutotuner(object):
    cache_dir      = IO.cache_path() + '/autotune'
    config_file    = cache_dir+'/configs.pklz'
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    
    """OpenCl kernel work group size autotuner.
    """
    def __init__(self,name,work_dim,build_opts,autotuner_config,local_work_dim=None):
        """Initialize a KernelAutotuner.
        
        Parameters
        ----------
        work_dim: int
            Work dimension used in targetted OpenCL kernels.
        local_work_dim: int
            Work dimension used in workgroup sizes.
        """

        if not isinstance(autotuner_config, AutotunerConfig):
            raise ValueError('autotuner_config is not an AutotunerConfig.')
        
        self.name           = name
        self.work_dim       = work_dim
        self.local_work_dim = work_dim if (local_work_dim is None) else local_work_dim
        self.build_opts = build_opts
        self.autotuner_config = autotuner_config
        
        self.enable_variable_workload = False
        self.workloads = [(1,1,1)]

        self.extra_parameters = None

        self._init_and_load_cache()
        self._load_default_filters()
    
    @staticmethod 
    def _hash_func():
        return hashlib.new('sha256')

    @staticmethod
    def _load_configs():
        configs = {}
        if os.path.isfile(KernelAutotuner.config_file):
            with gzip.open(KernelAutotuner.config_file, 'rb') as f:
                configs.update(pickle.loads(f.read()))
        return configs

    @staticmethod
    def _make_config_key(work_dim, typegen, build_opts):
        concat_unique_list = lambda L: '['+'_'.join([str(val) for val in frozenset(L)])+']'
        hasher = KernelAutotuner._hash_func()
        hasher.update('{}_{}_{}'.format(work_dim, concat_unique_list(build_opts),
             typegen.__repr__()))
        return hasher.hexdigest() 
    
    def _update_configs(self):
        configs = KernelAutotuner._load_configs()
        if configs.keys() != self.configs.keys():
            configs.update(self.configs)
            with gzip.open(KernelAutotuner.config_file, 'wb') as f:
                pickle.dump(configs,f)
        self.configs = configs

    def _init_and_load_cache(self):
        cache_file = '{}/{}.pklz'.format(KernelAutotuner.cache_dir,self.name.replace(' ','_'))
        if os.path.isfile(cache_file):
            with gzip.open(cache_file, 'rb') as f:
                self.results = pickle.loads(f.read())
        else:
            self.results = {}
        self.configs    = KernelAutotuner._load_configs()
        self.cache_file = cache_file

    def _dump_cache(self):
        with gzip.open(self.cache_file, 'wb') as f:
            pickle.dump(self.results,f)


    def enable_variable_workitem_workload(self, max_workitem_workload):
        max_workitem_workload = np.asarray(max_workitem_workload)
        assert (max_workitem_workload>0).all()
        
        def _compute_pows(max_workload):
            pows = [1]
            workload = 1
            while(workload<max_workload):
                workload <<= 1
                pows.append(workload)
            return pows

        self.enable_variable_workload = True
        if np.isscalar(max_workitem_workload):
            workloads = _compute_pows(max_workitem_workload)
            workloads = itertools.product(workloads, repeat=3)
        else:
            workloads = [_compute_pows(mww) for mww in max_workitem_workload]
            workloads = itertools.product(*workloads)
        self.workloads = [w for w in workloads]
   
    def register_extra_parameter(self,name, values):
        if self.extra_parameters is None:
            self.extra_parameters = {}
        self.extra_parameters[name] = values

    def add_filter(self,fname, f):
        self.filters[fname] = f
        return self

    def get_candidates(self,**kargs):
        return np.asarray([c for c in self._get_wi_candidates(**kargs)])
    
    def get_workloads(self,global_size):
        def F(wl):
            return (wl[0]<=global_size[0]) and (wl[1]<=global_size[1]) and (wl[2]<=global_size[2])
        candidates = itertools.ifilter(F, self.workloads)
        return np.asarray([c for c in candidates])

    def get_extra_parameters(self):
        if self.extra_parameters is None:
            return [None]
        else:
            return itertools.product(*self.extra_parameters.values())

    def bench(self,typegen,work_size,kernel_args,
            kernel_generator    = None,
            get_max_global_size = None,
            **kargs):

        assert 'global_size' not in kargs, \
                'global_size has been replaced by work_size, due to the variable work per workitem option.'

        if np.isscalar(work_size):
            work_size = [work_size]
        else:
            work_size = list(work_size)
        work_size += [1]*(3-len(work_size))
        work_size = np.asarray(work_size)
        ws = tuple(work_size,)

        if not isinstance(kernel_args, list):
            msg='kernel_args should be a list.'
            raise ValueError(msg)
        
        if (get_max_global_size is None):
            get_max_global_size = lambda work_size, work_load, **kargs: \
                    (work_size+workload-1)/workload
        
        platform = typegen.platform
        device   = typegen.device
        ctx      = typegen.context

        verbose = self.autotuner_config.verbose
        if verbose:
            print '== Kernel {} Autotuning =='.format(self.name)
            print 'platform: {}'.format(platform.name)
            print 'device: {}'.format(device.name)
            print 'ctx: {}'.format(ctx)
            print 'work_size: {}'.format(work_size)
            print 'fbtype: {}'.format(typegen.fbtype)
            print 'build_opts: {}'.format(self.build_opts)
        
        config = KernelAutotuner._make_config_key(self.work_dim, typegen, self.build_opts)
        if config not in self.configs.keys():
            self.configs[config] = {'work_dim':self.work_dim, 
                                    'build_opts':self.build_opts, 
                                    'typegen':typegen.__repr__()}
            self._update_configs()
        if config not in self.results.keys():
            self.results[config] = {}
        results = self.results[config]
        
        best_workload     = None
        best_global_size  = None
        best_local_size   = None
        best_stats        = None
        best_extra_params = None
                
        dump_cache = False

        separator = '_'*100
        indent = lambda i: '  '*i
       
        for extra_parameters in self.get_extra_parameters():
            if self.extra_parameters is None:
                extra_parameters = {}
            else:
                extra_parameters = dict(zip(self.extra_parameters.keys(), extra_parameters))

            params_hash = hashlib.sha256(str(hash(frozenset(sorted(extra_parameters.items()))))).hexdigest()
            params_hash = params_hash[:8]
                
            best_params_workload    = None
            best_params_global_size = None
            best_params_local_size  = None
            best_params_stats       = None
                
            if verbose:
                print separator
                print '::Current tuning parameters:: {}'.format(extra_parameters)
            
            workloads = self.get_workloads(work_size) 
            for workload in workloads:
                workload=np.asarray(workload)
                max_global_size = get_max_global_size(work_size, workload, **extra_parameters)
            
                best_workload_global_size = None
                best_workload_local_size  = None
                best_workload_stats       = None
                best_workload_candidate   = None
                best_workload_ids         = None
                
                if verbose:
                    print separator
                    print indent(1)+'::Current workload {}::'.format(workload)
                    print indent(2)+'-> global_size is set to {}'.format(max_global_size)
                
                candidates = self.get_candidates(ctx=ctx,device=device,
                        max_global_size=max_global_size,**kargs)
                unpruned_candidates = np.zeros_like(candidates)
                nruns=self.autotuner_config.nruns

                step=0
                all_pruned=False
                while step==0 or candidates.shape[0]>1:
                    stats = []
                    pruned_count = 0
                    unpruned_count = 0

                    if verbose and candidates.shape[0]>0:
                        msg='\n'+indent(2)+'Step {} :: running {} candidates over {} runs:'
                        print msg.format(step,candidates.shape[0],nruns)
                    
                    for local_work_size in candidates:
                        local_work_size  = np.asarray(local_work_size)
                        lwi = tuple(local_work_size)
                        
                        try:
                            (_kernel,_kernel_args,src_hash, global_size) = kernel_generator(
                                    ctx=ctx,device=device, 
                                    work_size=work_size,work_load=workload,local_work_size=lwi,
                                    build_opts=self.build_opts, 
                                    kernel_args=kernel_args,
                                    extra_parameters=extra_parameters,
                                    **kargs)
                        except KernelGenerationError:
                            pruned_count += 1
                            continue
                
                        global_size = np.asarray(global_size) 
                        gwi         = tuple(global_size)
                        
                        update=False
                        pms = params_hash
                        
                        if (not self.autotuner_config.override_cache)    \
                            and src_hash in results.keys()               \
                            and pms in results[src_hash].keys()          \
                            and ws  in results[src_hash][pms].keys()     \
                            and gwi in results[src_hash][pms][ws].keys() \
                            and lwi in results[src_hash][pms][ws][gwi].keys():
                            
                            stat = self.results[config][src_hash][pms][ws][gwi][lwi]
                            if stat.nruns >= nruns:
                                if self.autotuner_config.verbose:
                                    print indent(3)+'{} {} => {} (cached)'.format(gwi, lwi, stat)
                                unpruned_candidates[unpruned_count,:] = local_work_size
                                unpruned_count+=1
                                stats.append(stat)
                                continue
                            else:
                                update=True
                               
                        if (best_stats is not None):
                            current_best_stats = best_stats
                        elif (best_params_stats is not None):
                            current_best_stats = best_params_stats
                        elif (best_workload_stats is not None):
                            current_best_stats = best_workload_stats
                        else:
                            current_best_stats = None

                        (stat,pruned) = self._bench_one(ctx,device,gwi,lwi,_kernel,_kernel_args,nruns,
                                current_best_stats)

                        if not pruned:
                            unpruned_candidates[unpruned_count,:] = local_work_size
                            unpruned_count+=1
                            status='update' if update else 'new'
                            stats.append(stat)
                        else:
                            pruned_count+=1
                            status='pruned'

                        if verbose:
                            print indent(3)+'{} {} => {} ({})'.format(gwi, lwi, stat, status)

                        if not pruned:
                            if src_hash not in results.keys():
                                results[src_hash] = {}
                            if pms not in results[src_hash].keys():
                                results[src_hash][pms] = {}
                            if ws not in results[src_hash][pms].keys():
                                results[src_hash][pms][ws] = {}
                            if gwi not in results[src_hash][pms][ws].keys():
                                results[src_hash][pms][ws][gwi] = {}
                            results[src_hash][pms][ws][gwi][lwi] = stat
                            dump_cache = True
                    
                    all_pruned = (pruned_count==candidates.shape[0])
                    if unpruned_count+pruned_count!=candidates.shape[0]:
                        raise RuntimeError()
                    if all_pruned:
                        break
                    
                    keep  = max(1,unpruned_count//2)
                    best_workload_ids=Utils.argsort(stats)[:keep]
                    candidates=unpruned_candidates[best_workload_ids,:]

                    nruns *= 2
                    step += 1
                   
                if all_pruned:
                    if verbose:
                        print separator
                        print indent(1)+' Workload {} winner for kernel {}:'.format(workload, self.name)
                        print indent(2)+'no winner (all candidates were pruned)'
                    continue

                if (candidates.shape[0]!=1 or len(best_workload_ids)!=1):
                    raise RuntimeError()
            
                if dump_cache:
                    self.results[config] = results
                    self._dump_cache()
                    dump_cache=False
                
                best_workload_id        = best_workload_ids[0]
                best_workload_stats     = stats[best_workload_id]
                best_workload_candidate = candidates[0]
                
                best_workload_local_size  = best_workload_candidate
                (_,_,_,best_workload_global_size) = kernel_generator(
                                    ctx=ctx,device=device, 
                                    work_size=work_size,work_load=workload,
                                    local_work_size=best_workload_local_size,
                                    build_opts=self.build_opts, 
                                    kernel_args=kernel_args,
                                    extra_parameters=extra_parameters,
                                    **kargs)
                
                if verbose:
                    print separator
                    print indent(1)+' Workload {} winner for kernel {}:'.format(workload, self.name)
                    print indent(2)+'{} {} => {}'.format(best_workload_global_size, 
                            best_workload_local_size, best_workload_stats)

                if (best_params_stats is None) or (best_workload_stats<best_workload_stats):
                    best_params_workload    = workload
                    best_params_stats       = best_workload_stats
                    best_params_global_size = best_workload_global_size
                    best_params_local_size  = best_workload_local_size
                
            if verbose:
                print separator
                print ' Current parameters winner for kernel {}:'.format(self.name)
                if (best_params_stats is None):
                    print indent(1)+'no winner (all candidates were pruned)'
                else:
                    print indent(1)+'{} {} => {}'.format(best_params_global_size, 
                            best_params_local_size, best_params_stats)

            if (best_params_stats is not None) and \
                    ((best_stats is None) or (best_params_stats<best_stats)):
                best_workload     = best_params_workload
                best_stats        = best_params_stats
                best_global_size  = best_params_global_size
                best_local_size   = best_params_local_size
                best_extra_params = extra_parameters
        
        if verbose:
            print separator
            print ' BEST OVERALL RESULT for kernel {}:'.format(self.name)
            print ' => Extra params: {}'.format(best_extra_params)
            print ' => WL={} G={} L={} => {}'.format(best_workload, best_global_size, best_local_size, 
                    best_stats)
            print separator
            print

        return (best_global_size, best_local_size, best_stats, best_workload, best_extra_params)

    
    def _bench_one(self,ctx,device,global_size,local_work_size,kernel,kernel_args,nruns,
                best_stat):
        
        assert(nruns>=1)
        profiling_enable=cl.command_queue_properties.PROFILING_ENABLE
        kernel.set_args(*kernel_args)


        evts = []
        with cl.CommandQueue(ctx,device,profiling_enable) as queue:
            evt = cl.enqueue_nd_range_kernel(queue, kernel, global_size, local_work_size)
            evts.append(evt)
        stats = OpenClKernelStatistics(events=evts, nruns=1)
        if (best_stat is None):
            pruned = False
        else:
            pruned = (stats.min > self.autotuner_config.prune_threshold*best_stat.max)
        
        if not pruned and nruns>1: 
            with cl.CommandQueue(ctx,device,profiling_enable) as queue:
                for i in xrange(nruns-1):
                    evt = cl.enqueue_nd_range_kernel(queue, kernel, global_size, local_work_size)
                    evts.append(evt)
            stats = OpenClKernelStatistics(events=evts,nruns=nruns)
        return (stats, pruned)

    def _get_wi_candidates(self,ctx,device,max_global_size,**kargs):
        pows = []
        size = 1
        
        max_size = device.max_work_group_size
        while(size<=max_size):
            pows.append(size)
            size <<= 1
        pows = np.asarray(pows)
        
        product = []
        for i in xrange(self.work_dim):
            good = (pows<=max_global_size[i])
            product.append(pows[good])
        for i in xrange(3-self.work_dim):
            product.append([1])
        candidates = itertools.product(*product)

        for fname,f in self.filters.iteritems():
            F = f(ctx=ctx,device=device,max_global_size=max_global_size,**kargs)
            candidates = itertools.ifilter(F, candidates)
        return candidates

    def _load_default_filters(self):
        self.filters = {}
        self.add_filter('dim_reqs',self._dim_filter)
        self.add_filter('minmax_wi',self._minmax_workitems_filter)

    #default filters
    def _dim_filter(self,device,**kargs):
            work_dim   = self.local_work_dim
            max_wi_dim = device.max_work_item_dimensions
            return lambda local_work_size: (work_dim<=max_wi_dim) and  \
                (work_dim==3 
                or (work_dim==2 and local_work_size[2]==1)
                or (work_dim==1 and local_work_size[1]==1 and local_work_size[2]==1))
    def _minmax_workitems_filter(self, device, min_load_factor=None, **kargs):
        def filter(local_work_size, **kargs):
            max_wg_size = device.max_work_group_size
            wi=1
            for i in xrange(3):
                wi*=local_work_size[i]
            if min_load_factor is None:
                return (wi<=max_wg_size)
            else:
                return (wi>=max_wg_size/min_load_factor) and (wi<=max_wg_size)
        return filter
    
    #user available filters
    def ordering_filter(self, **kargs):
        return lambda local_work_size: (local_work_size[2]<=local_work_size[1]) \
                and (local_work_size[1]<=local_work_size[0])
    def min_workitems_per_direction(self, min_local_size, **kargs):
        if np.isscalar(min_local_size):
            min_local_size = [min_local_size]
        else:
            min_local_size = list(min_local_size)
        min_local_size = np.asarray(min_local_size)
        wd = self.work_dim
        return lambda local_work_size,**kargs: (local_work_size[:wd]>=min_local_size).all()
    def max_workitems_per_direction(self, max_local_size, **kargs):
        if np.isscalar(max_local_size):
            max_local_size = [max_local_size]
        else:
            max_local_size = list(max_local_size)
        max_local_size = np.asarray(max_local_size)
        wd = self.work_dim
        return lambda local_work_size,**kargs: (local_work_size[:wd]<=max_local_size).all()
            
