
from hysop import __VERBOSE__

from hysop.backend.device.opencl import cl
from hysop.backend.device.opencl.opencl_kernel   import OpenClKernelLauncher
from hysop.backend.device.opencl.gpu_particle_advection_dir import GPUParticleAdvectionDir, MeshDirection

from hysop.methods import TimeIntegrator, Remesh, ExtraArgs, \
    Support, Splitting, MultiScale, Interpolation, Precision,\
    StretchingFormulation, BoundaryCondition

from hysop.numerics.odesolvers    import Euler, RK2, RK3, RK4
from hysop.numerics.interpolation.interpolation import Linear

from hysop.numerics.remeshing import L2_1, L2_2, L2_3, L2_4
from hysop.numerics.remeshing import L4_2, L4_3, L4_4
from hysop.numerics.remeshing import L6_3, L6_4, L6_5, L6_6
from hysop.numerics.remeshing import L8_4

from hysop.constants import np, DirectionLabels 
from hysop.tools.numpywrappers import npw

from hysop.backend.device.codegen.structs.mesh_info import MeshInfoStruct
from hysop.backend.device.codegen.kernels.copy_kernel import CopyKernel

from hysop.backend.device.codegen.kernels.directional_advection  import DirectionalAdvectionKernel
from hysop.backend.device.codegen.kernels.directional_stretching import DirectionalStretchingKernel

from hysop.constants import callback_profiler

class StaticGPUParticleAdvectionDir(GPUParticleAdvectionDir):
    ##  
    ## This implementation is based on GPUOperator._kernel_cfg
    ## and kernels located in hysop/gpu/cl_src.
    ##
    ## This class only overrides the particles advection kernel backend methods 
    ## (mostly kernel parsing and building, as well as kernel calls used in _exec_list).
    ##
    ## See hysop.gpu.CodegenGPUParticleAdvectionDir for the runtime generated kernels implementation.
    ##
    def __init__(self,**kargs): 
        super(self.__class__, self).__init__(**kargs)
        
        # Additional method and configuration checks
        self._check()

        # Initialize and configure everything
        self._initialize()


    def _check(self):
        super(StaticGPUParticleAdvectionDir,self)._check()
        cls = self.__class__.__name__
        dim = self.dim
    
        if dim not in [2,3]:
            msg='{} only supports 2D and 3D domains.'.format(cls)
            raise NotImplementedError(msg)

        _kernel_cfg = self._kernel_cfg
        if _kernel_cfg is None:
            msg='{} requires a kernel configuration (_kernel_cfg) to be present !'\
                    .format(cls)
            raise ValueError(msg)

        required_kernel_configs = ['advec','remesh','advec_and_remesh']
        if dim>=2: 
            required_kernel_configs.append('transpose_xy')
        if dim>=3: 
            required_kernel_configs.append('transpose_xz')

        for rkc in required_kernel_configs:
            msg = "{} requires kernel configuration '{}' but it is not present.".format(cls, rkc)
            if rkc not in _kernel_cfg.keys():
               raise KeyError(msg)

        if self._has_stretching:
            stretching = self._stretching
            if not isinstance(stretching,dict):
                msg='Stretching value should be a dict containing a formulation and an order.'
                raise ValueError(msg)
            reqkeys =  ['formulation','order']
            for k in reqkeys:
                if k not in stretching.keys():
                    msg ="Missing key '{}' in streching dictionnary.".format(k)
                    raise ValueError(msg)
            formulation = stretching['formulation']
            order = stretching['order']
            if order<0:
                msg='Stretching order < 0'
                raise ValueError(msg)
            if not isinstance(formulation,StretchingFormulation):
                msg='Stretching formulation is not one of {}.'.format(StretchingFormulation.svalues())
                raise ValueError(msg)

    def _initialize_cl_env(self):
        self._initialize_cl_build_options()
        self._initialize_cl_size_constants()
        self._initialize_cl_mesh_info()
        
        required_components = set()
        for fg in self.fields_on_grid:
            required_components.add(fg.nb_components)
        self.required_components = required_components

    def _initialize_cl_build_options(self):
        self._build_options += ""

    def _initialize_cl_size_constants(self):
        """
        Compile time constants for kernels.
        """
        self._append_size_constants(self.f_resol)
        self._append_size_constants(self.v_resol, prefix='V_NB')
        
        self._append_size_constants(
            [self.velocity_topo.ghosts()[self.direction]],
            prefix='V_GHOSTS_NB', suffix=[''])

        self._append_size_constants([self._is_multi_scale*1],
                prefix='ADVEC_IS_MULTISCALE', suffix=[''])
    
    def _initialize_cl_mesh_info(self):

        # Coordinates of the local origin
        coord_min = npw.ones(4, dtype=self.gpu_precision)
        coord_min[:self.dim] = self.fields_topo.mesh.origin

        # Space step for fields
        mesh_size = npw.ones(4, dtype=self.gpu_precision)
        mesh_size[:self.dim] = self._reorder_vect(
            self.fields_topo.mesh.space_step)

        # Space step for velocity
        v_mesh_size = npw.ones(4, dtype=self.gpu_precision)
        v_mesh_size[:self.dim] = self._reorder_vect(
            self.velocity_topo.mesh.space_step)

        mesh_info = npw.ones((12, ))
        mesh_info[:4]  = mesh_size
        mesh_info[4:8] = v_mesh_size
        mesh_info[8]   = coord_min[self.direction]
        mesh_info[9]   = 1. / mesh_size[0]
        mesh_info[10]  = 1. / v_mesh_size[0]

        cl_mesh_info   = cl.Buffer(self.cl_env.ctx, cl.mem_flags.READ_ONLY,
                                       size=mesh_info.nbytes)
        cl.enqueue_write_buffer(self.cl_env.queue,
                                cl_mesh_info, mesh_info).wait()
    
        self._mesh_info    = mesh_info
        self._cl_mesh_info = cl_mesh_info
            
        
        #TODO remove
        fields_mesh = self.fields_topo.mesh
        (fields_mesh_info,fields_mesh_info_var) = MeshInfoStruct.create_from_mesh('field_mesh_info',
                self.cl_env, self.fields_topo.mesh, self.mesh_dir, self.mesh_state)

        self._fields_mesh_info      = fields_mesh_info
        self._fields_mesh_info_var  = fields_mesh_info_var
        
        velocity_mesh = self.velocity_topo.mesh
        (velocity_mesh_info,velocity_mesh_info_var) = MeshInfoStruct.create_from_mesh('velocity_mesh_info',
                self.cl_env, self.velocity_topo.mesh, self.mesh_dir, self.mesh_state)

        self._velocity_mesh_info      = velocity_mesh_info
        self._velocity_mesh_info_var  = velocity_mesh_info_var
            
    def _collect_transposition_kernel_xy(self):
        # Only collected for direction Y (XYZ -> YXZ)
        resolution = self.f_resol_dir
        defines = ' -D NB_I=NB_Y -D NB_II=NB_X -D NB_III=NB_Z'
        return self._build_transpose_kernel_xy(resolution, defines)
    
    def _collect_transposition_kernel_xz(self):
        # Only collected for direction Z (YXZ -> ZXY)
        resolution = self.f_resol_dir
        defines = ' -D NB_I=NB_Z -D NB_II=NB_X -D NB_III=NB_Y'
        return self._build_transpose_kernel_xz(resolution, defines)
    
    def _collect_transposition_kernel_zx(self):
        # Only collected for direction Y (ZXY -> YXZ) in sequential
        resolution = self.f_resol_dir
        defines = ' -D NB_I=NB_Y -D NB_II=NB_X -D NB_III=NB_Z'
        return self._build_transpose_kernel_xz(resolution, defines)

    def _collect_transposition_kernel_yx(self):
        # Only collected for direction X (YXZ -> XYZ)
        resolution = self.f_resol_dir
        defines = ' -D NB_I=NB_X -D NB_II=NB_Y -D NB_III=NB_Z'
        return self._build_transpose_kernel_xy(resolution, defines)

    def _build_transpose_kernel_xy(self, resolution, defines):
        build_options = self._build_options + self._size_constants

        src, t_dim, b_rows, has_padding, vec, f_space = \
            self._kernel_cfg['transpose_xy']

        while t_dim > resolution[0] or t_dim > resolution[1] or \
                (resolution[0] % t_dim) > 0 or (resolution[1] % t_dim) > 0:
            t_dim /= 2
        gwi, lwi, blocs_nb = f_space(resolution, t_dim, b_rows, vec)

        if has_padding:
            build_options += " -D PADDING_XY=1"
        else:
            build_options += " -D PADDING_XY=0"
        build_options += " -D TILE_DIM_XY={0}".format(t_dim)
        build_options += " -D BLOCK_ROWS_XY={0}".format(b_rows)
        build_options += " -D NB_GROUPS_I={0}".format(blocs_nb[0])
        build_options += " -D NB_GROUPS_II={0}".format(blocs_nb[1])
        build_options += defines

        prg = self.cl_env.build_src(src, build_options, vec)
        return OpenClKernelLauncher(prg.transpose_xy, self.cl_env.queue, gwi, lwi)
    
    def _build_transpose_kernel_xz(self, resolution, defines):
        build_options = self._build_options + self._size_constants

        src, t_dim, b_rows, b_deph, is_padding, vec, f_space = \
            self._kernel_cfg['transpose_xz']

        while t_dim > resolution[0] or t_dim > resolution[2] or \
                (resolution[0] % t_dim) > 0 or (resolution[2] % t_dim) > 0:
            t_dim /= 2
        gwi, lwi, blocs_nb = f_space(resolution, t_dim, b_rows, b_deph, vec)
        if is_padding:
            build_options += " -D PADDING_XZ=1"
        else:
            build_options += " -D PADDING_XZ=0"
        build_options += " -D TILE_DIM_XZ={0}".format(t_dim)
        build_options += " -D BLOCK_ROWS_XZ={0}".format(b_rows)
        build_options += " -D BLOCK_DEPH_XZ={0}".format(b_deph)
        build_options += " -D NB_GROUPS_I={0}".format(blocs_nb[0])
        build_options += " -D NB_GROUPS_III={0}".format(blocs_nb[2])
        build_options += defines

        prg = self.cl_env.build_src(src,build_options,vec)
        return OpenClKernelLauncher(prg.transpose_xz, self.cl_env.queue, gwi, lwi)
    
    # def _collect_advec_kernel(self):
        # """
        # Compile OpenCL sources for advection and remeshing kernel.
        # """
        # Advection
        # build_options = self._build_options + self._size_constants
        # src, is_noBC, vec, f_space = self._kernel_cfg['advec']
        # gwi, lwi = f_space(self.f_resol_dir, vec)
        # WINb = lwi[0]

        # if self._is_multi_scale:
            # build_options += " -D MS_FORMULA="
            # build_options += self.method[MultiScale].__name__.upper()

        # if is_noBC:
            # build_options += " -D WITH_NOBC=1"
        # build_options += " -D WI_NB=" + str(WINb)
        # build_options += " -D PART_NB_PER_WI="
        # build_options += str(self.f_resol_dir[0] / WINb)
        
        # Build code
        # src = [s.replace('RKN', self.method[TimeIntegrator].__name__.lower())
               # for s in src]
        
        # Adding remeshing weights for the multiscale advection
        # if self._is_multi_scale:
            # src.insert(1, self._kernel_cfg['remesh'][0][1])
        
        # Euler integrator
        # if self.method[TimeIntegrator] is Euler:
            # if not self._is_multi_scale:
                # src = [s for s in src if s.find(Euler.__name__.lower()) < 0]
                # src[-1] = src[-1].replace('advection', 'advection_euler')

        # prg = self.cl_env.build_src(
            # src,
            # build_options,
            # vec,
            # nb_remesh_components=self.velocity.nb_components)

        # callback_profiler.register_tasks('advection')
        # self._advec = OpenClKernelLauncher(
            # prg.advection_kernel, self.cl_env.queue, gwi, lwi)

    def _collect_remesh_kernel(self):
        # remeshing
        build_options = self._build_options + self._size_constants
        src, is_noBC, vec, f_space = self._kernel_cfg['remesh']
        gwi, lwi = f_space(self.f_resol_dir, vec)
        WINb = lwi[0]

        build_options += " -D FORMULA=" + self.method[Remesh].__name__.upper()
        if is_noBC:
            build_options += " -D WITH_NOBC=1"
        build_options += " -D WI_NB=" + str(WINb)
        build_options += " -D PART_NB_PER_WI="
        build_options += str(self.f_resol_dir[0] / WINb)
            
        enum = ['I', 'II', 'III']
        comp  = self._reorder_vect(['NB' + d for d in DirectionLabels[:self.dim]])
        vcomp = self._reorder_vect(['V_NB' + d for d in DirectionLabels[:self.dim]])
        for i,suffix in enumerate(enum[:self.dim]):
            build_options += ' -D NB_{}={}'.format(suffix,comp[i])
        for i,suffix in enumerate(enum[:self.dim]):
            build_options += ' -D V_NB_{}={}'.format(suffix,vcomp[i])

        self._remesh = {}
        cnames = []
        for rc in self.required_components:
            prg = self.cl_env.build_src(
                src, build_options, vec,
                nb_remesh_components=rc)
            self._remesh[rc] = OpenClKernelLauncher(
                prg.remeshing_kernel, self.cl_env.queue, gwi, lwi)
            
            cname = 'remesh' if len(self.required_components)==1 \
                    else 'remesh_{}'.format(rc)
            cnames.append(cname)
            callback_profiler.register_tasks(cname)
       
        if len(cnames)>1:
            callback_profiler.register_group('remesh',cnames)

    def _collect_advec_remesh_kernel(self):
        """
        Compile OpenCL sources for advection and remeshing kernel.
        """
        build_options = self._build_options + self._size_constants

        src, is_noBC, vec, f_space = self._kernel_cfg['advec_and_remesh']
        gwi, lwi = f_space(self.f_resol_dir, vec)

        WINb = lwi[0]
        build_options += " -D FORMULA=" + self.method[Remesh].__name__.upper()

        if self._is_multi_scale:
            build_options += " -D MS_FORMULA="
            build_options += self.method[MultiScale].__name__.upper()
        if is_noBC:
            build_options += " -D WITH_NOBC=1"

        build_options += " -D WI_NB=" + str(WINb)
        build_options += " -D PART_NB_PER_WI=" + str(self.f_resol_dir[0] / WINb)
        
        # Build code
        src = [s.replace('RKN', self.method[TimeIntegrator].__name__.lower())
               for s in src]
        
        # Euler integrator
        if self.method[TimeIntegrator] is Euler:
            if not self._is_multiScale:
                src = [s for s in src if s.find(Euler.__name__.lower()) < 0]
                src[-1] = src[-1].replace('advection', 'advection_euler')
        
        self._advec_and_remesh = {}
        cnames = []
        for rc in self.required_components:
            prg = self.cl_env.build_src(
                src, build_options, vec,
                nb_remesh_components=rc)
            self._advec_and_remesh[rc] = OpenClKernelLauncher(
                prg.advection_and_remeshing, self.cl_env.queue, gwi, lwi)
            
            cname = 'advec_remesh' if len(self.required_components)==1 \
                    else 'advec_remesh_{}'.format(rc)
            callback_profiler.register_tasks(cname)
            cnames.append(cname)
        
        if len(cnames)>1:
            callback_profiler.register_group('remesh',cnames)

       
    def _collect_user_kernels(self):
        """
        Build user kernel sources.
        """

        user_src = self._user_src

        if user_src is not None:
            build_options = self._build_options + self._size_constants

            f_resol_dir = self.f_resol_dir 
            v_resol_dir = self.v_resol_dir
            
            f_work_items, gwi, lwi = self.cl_env.get_work_items(f_resol_dir)
            v_work_items, gwi, lwi = self.cl_env.get_work_items(v_resol_dir)

            build_options += " -D WI_NB="   + str(f_work_items)
            build_options += " -D V_WI_NB=" + str(v_work_items)

            self._user_prg = self.cl_env.build_src(usr_src, build_options, 1)
        else:
            self._user_prg = None

    def _collect_extra_kernels(self):
        if self._has_stretching:
            self._collect_stretching_kernels()
        if self._has_diffusion:
            self._collect_diffusion_kernels()
    
    
    def _collect_copy_kernels(self):
        
        if self._use_builtin_copy:
            super(StaticGPUParticleAdvectionDir,self)._collect_copy_kernels()
        else:
            tg = self.cl_env.typegen
            (fg,fp) = self.fields_on_part.items()[0]
            src = fp[0]
            dst = fg.gpu_data[0]
            vtype = tg.fbtype
            size = fg.mem_size // (3*tg.FLT_BYTES[vtype])

            (kernel_launcher, kernel_args, kernel_args_mapping, cached_bytes) = \
                    CopyKernel.autotune(self.cl_env,
                            src, dst, vtype, size, 
                            restrict=True, 
                            build_opts=self._build_options,
                            autotuner_config=self._autotuner_config)
            self.copy = kernel_launcher
    
    def _collect_advec_kernel(self):

        ftype              = self.cl_env.typegen.fbtype
        compute_resolution = self.fields_topo.mesh.compute_resolution
        boundary           = BoundaryCondition.PERIODIC
        rk_scheme          = self.method[TimeIntegrator]

        velocity    = self.velocity
        position    = self._particle_position

        field_mesh_info = self._fields_mesh_info_var
        velo_mesh_info  = self._velocity_mesh_info_var

        max_dt=0.01
        max_velocity=1
        velocity_dx = velo_mesh_info.value['dx'][0]

        (kernel_launcher, kernel_args, kernel_args_mapping, 
                total_work, per_work_statistic, cached_bytes) = \
                DirectionalAdvectionKernel.autotune(self.cl_env, ftype,
                        self.dim,self.direction,
                        velocity, position,
                        compute_resolution, field_mesh_info,
                        boundary, rk_scheme,
                        max_dt, max_velocity, velocity_dx,
                        build_opts=self._build_options,
                        autotuner_config=self._autotuner_config)
    
        callback_profiler.register_tasks('advection')
                # ftype=ftype,
                # total_work=total_work, 
                # per_work_statistic=per_work_statistic)

        def do_advec(dt,**kargs):
            callback_profiler.tic('advection')
            kernel_args[kernel_args_mapping['dt']] = self.cl_env.precision(dt)
            evt = kernel_launcher(*kernel_args)
            callback_profiler.tac('advection',evt=evt)
            return evt

        self._do_advec = do_advec
        

    def _collect_stretching_kernels(self):

        self.vorticity = self.fields_on_grid[0]

        ftype = self.cl_env.typegen.fbtype
        compute_resolution = self.fields_topo.mesh.compute_resolution
        boundary=self._stretching['boundary']
        order=self._stretching['order']
        formulation=self._stretching['formulation']
        rk_scheme=self._stretching['rk_scheme']

        mesh_info = self._fields_mesh_info_var
        dt=0.1

        (kernel_launcher, kernel_args, kernel_args_mapping, 
                total_work, per_work_statistic, cached_bytes) = \
                DirectionalStretchingKernel.autotune(self.cl_env, ftype,
                        self.dim,self.direction,
                        compute_resolution,
                        boundary,order,
                        formulation, rk_scheme,
                        self.velocity, self.vorticity, 
                        mesh_info, dt,
                        build_opts=self._build_options,
                        autotuner_config=self._autotuner_config)

        callback_profiler.register_tasks('stretching', 
                ftype=ftype,
                total_work=total_work, 
                per_work_statistic=per_work_statistic)

        def do_stretch(dt,**kargs):
            callback_profiler.tic('stretching')
            kernel_args[kernel_args_mapping['dt']] = self.cl_env.precision(dt)
            evt = kernel_launcher(*kernel_args)
            callback_profiler.tac('stretching',evt=evt)
            return evt

        self._do_stretch = do_stretch

    def _collect_diffusion_kernels(self):
        self._diffuse = None
        #callback_profiler.register_tasks('diffusion')


    def _exec_kernel(self,kernel):
        for (fg,fp) in self.fields_on_part.iteritems():
            evts = []
            for (g,p) in zip(fg.gpu_data,fp):
                if kernel is self.copy and self._use_builtin_copy:
                    evt = kernel.launch_sizes_in_args(g,p,wait_for=fg.events)
                else:
                    evt = kernel(p,g,wait_for=fg.events)
                evts.append(evt)
            fg.events += evts
        return evts
        
    def _do_copy(self,**kargs): 
        if __VERBOSE__:
            print '_do_copy'
        callback_profiler.tic('cpu2gpu')
        evts = self._exec_kernel(self.copy)
        callback_profiler.tac('cpu2gpu',evts=evts)
    
    def _do_compute_1k_monoscale(self, dt, **kargs):

        velocity  = self.velocity
        for (fg,fp) in self.fields_on_part.iteritems():
            nbc = fg.nb_components
            wait_evts = velocity.events + fg.events

            args = tuple(
                  [velocity.gpu_data[self.direction]] 
                + [fp[i]          for i in xrange(nbc)] 
                + [fg.gpu_data[i] for i in xrange(nbc)] 
                + [self.gpu_precision(dt)]
                + [self._cl_mesh_info]
            )
            
            cname = 'advec_remesh' if len(self.required_components)==1 \
                    else 'advec_remesh_{}'.format(nbc)
            callback_profiler.tic(cname)
            evt = self._advec_and_remesh[nbc](*args, wait_for=wait_evts)
            callback_profiler.tac(cname)

            fg.events.append(evt)
            velocity.events.append(evt)
    
    def _do_compute_2k_monoscale(self, dt, **kargs):

        velocity  = self.velocity
        nbc = velocity.nb_components

        # Advection
        advec_evt = self._do_advec(dt=dt,wait_for=velocity.events)
        velocity.events.append(advec_evt)
        
        # Remesh
        for (fg,fp) in self.fields_on_part.iteritems():
            nbc = fg.nb_components
            cname = 'remesh' if len(self.required_components)==1 \
                    else 'remesh_{}'.format(nbc)
            callback_profiler.tic(cname)
            args = tuple(
                  [self._particle_position]
                + [fp[i]          for i in xrange(nbc)]
                + [fg.gpu_data[i] for i in xrange(nbc)]
                + [self._cl_mesh_info]
            )
            remesh_evt = self._remesh[nbc](*args, wait_for=[advec_evt])
            fg.events.append(remesh_evt)
            callback_profiler.tac(cname)

        if self._has_stretching:
            evt=self._do_stretch(dt)
    
    
    def _do_compute_1k_multiscale(self, dt, **kargs):

        velocity  = self.velocity
        
        v_mesh_size            = npw.ones(4, dtype=self.gpu_precision)
        v_mesh_size[:self.dim] = self._reorder_vect(self.velocity_topo.mesh.space_step)

        for (fg,fp) in self.fields_on_part.iteritems():
            nbc = fg.nb_components
            wait_evts = velocity.events + fg.events
        
            args = tuple(
                  [velocity.gpu_data[self.direction]] 
                + [fp[i]          for i in xrange(nbc)] 
                + [fg.gpu_data[i] for i in xrange(nbc)] 
                + [self.gpu_precision(dt)]
                + [self.gpu_precision(1.0 / v_mesh_size[1])]
                + [self.gpu_precision(1.0 / v_mesh_size[2])]
                + [self._cl_mesh_info]
            )

            evt = self._advec_and_remesh[nbc](*args, wait_for=wait_evts)
            fg.events.append(evt)
    
    def _do_compute_2k_multiscale(self, dt, **kargs):

        velocity  = self.velocity
        nbc = velocity.nb_components
        
        v_mesh_size            = npw.ones(4, dtype=self.gpu_precision)
        v_mesh_size[:self.dim] = self._reorder_vect(self.velocity_topo.mesh.space_step)

        # Advection
        args = tuple([
            velocity.gpu_data[self.direction],
            self._particle_position,
            self.gpu_precision(dt),
            self.gpu_precision(1.0 / v_mesh_size[1]),
            self.gpu_precision(1.0 / v_mesh_size[2]),
            self._cl_mesh_info
        ])
        advec_evt = self._advec(*args,wait_for=velocity.events)
        velocity.events.append(advec_evt)
        
        for (fg,fp) in self.fields_on_part.iteritems():
            # Remesh
            nbc = fg.nb_components
            args = tuple(
                  [self._particle_position]
                + [fp[i]          for i in xrange(nbc)]
                + [fg.gpu_data[i] for i in xrange(nbc)]
                + [self._cl_mesh_info]
            )
            remesh_evt = self._remesh[nbc](*args, wait_for=[advec_evt])
            fg.events.append(remesh_evt)

    def _pre_apply(self):
        super(StaticGPUParticleAdvectionDir,self)._pre_apply()
        if 'cpu2gpu' not in callback_profiler.tasks:
            membytes=0
            for fg in self.fields_on_part.keys():
                 membytes += fg.mem_size
            callback_profiler.register_tasks('cpu2gpu',membytes=membytes)
            
            #callback_profiler.register_group('transpose',
                #['transpose_xy','transpose_yx','transpose_zx','transpose_xz'])
            

    @staticmethod
    def supports_multiscale():
        return True
    @staticmethod
    def supports_kernel_splitting():
        return True
    @staticmethod
    def supports_user_kernels():
        return True
    @staticmethod
    def supports_stretching():
        return True
    @staticmethod
    def supports_mpi():
        return False
