"""Classes and tools used to handle openCL interface.


* :class:`~hysop.gpu.tools.OpenClEnvironment`:
   object handling opencl platform, device ... info.
* :func:`~hysop.gpu.tools.get_opengl_shared_environment`:
   build or get an OpenCL environment with openGL properties.
* :func:`~hysop.gpu.tools.get_opencl_environment`:
   build or get an OpenCL environment.
* :func:`~hysop.gpu.tools.explore`
   explore system and display platform, devices, memory ... info.



"""
import os, re, itertools, hashlib, pickle, gzip

# from __future__ import print_function

from hysop import __VERBOSE__, __KERNEL_DEBUG__, \
        __DEFAULT_PLATFORM_ID__, __DEFAULT_DEVICE_ID__
from hysop.tools.io_utils import IO
from hysop.constants import np, HYSOP_REAL
from hysop.backend.device.opencl import cl, GPU_SRC, CL_PROFILE, KERNEL_DUMP_FOLDER

from hysop.core.mpi import MPI


FLOAT_GPU, DOUBLE_GPU = np.float32, np.float64

__cl_env = None
"""Global variable handling the OpenCL Environment instance """

class KernelError(Exception):
    """Custom exception for kernel errors.
    """
    def __init__(self, msg, err):
        super(KernelError,self).__init__(msg)
        self.msg = msg
        self.err = err

    def __str__(self):
        return self.err + ': ' + self.msg


class OpenClEnvironment(object):
    """OpenCL environment informations and useful functions.
    """

    def __init__(self, platform_id, device_id, device_type,
                 precision, gl_sharing=False, comm=None):
        """Initialize an OpenCL environment

        Parameters
        ----------
        platform_id : int
            chosen platform id.
        device_id : int
            chosen device id.
        device_type : string
            chosen device type.
        precision : int
            required precision for real data.
            Must be FLOAT_GPU or DOUBLE_GPU.
        gl_sharing : bool, optional
            True to build a context shared between OpenGL and OpenCL.
            Default=False.
        comm : mpi communicator, optional
            Communicator which handles the OpenCL env.
            Default = :data:`~hysop.core.mpi.main_comm`

        """
        self._platform_id = platform_id
        self._device_id   = device_id
        self._device_type = device_type
        self._gl_sharing  = gl_sharing
        # OpenCL platform
        self.platform = self._get_platform(platform_id)
        # OpenCL device
        self.device = self._get_device(self.platform, device_id, device_type)
        # Device available memory
        self.available_mem = self.device.global_mem_size
        # OpenCL context
        self.ctx = self._get_context(self.device, gl_sharing)
        # OpenCL queue
        self.queue = self._get_queue(self.ctx)

        # MPI sub-communicator for all processes attached to the same device
        if comm is None:
            from hysop.core.mpi import main_comm
        else:
            main_comm = comm
        
        # Splitting the mpi communicator by the device id is not enough:
        # the id of the first gpu of each node is 0
        # We build color from the processor name and the id
        import hashlib
        # The md5 sum of the proc name is tuncated to obtain an integer
        # for fortran (32bit)
        hash_name = hashlib.md5(MPI.Get_processor_name()).hexdigest()[-7:]
        self.gpu_comm = main_comm.Split(
            color=int(hash_name, 16) + device_id,
            key=main_comm.Get_rank())

        # Floating point codegeneration mode
        from hysop.backend.device.opencl.opencl_types import OpenClTypeGen
        _kargs = {'device':self.device, 'context':self.ctx, 'platform':self.platform }
        if __KERNEL_DEBUG__:
            _kargs['float_dump_mode'] = 'dec'
        else:
            _kargs['float_dump_mode'] = 'hex'

        # Floating point precision
        if precision is FLOAT_GPU:
            _kargs['fbtype'] = 'float'
        elif precision is DOUBLE_GPU:
            _kargs['fbtype'] = 'double'
        else:
            raise ValueError('Unknown floating point precision {}!'.format(precision))

        typegen = OpenClTypeGen(**_kargs)
    
        self.typegen   = typegen
        self.prec_size = typegen.FLT_BYTES[typegen.fbtype]
        self.precision = precision

        self.macros = {}
        self.default_build_opts = ""
        if CL_PROFILE and self.device.vendor.find('NVIDIA') >= 0:
            self.default_build_opts += " -cl-nv-verbose"
        #self.default_build_opts += "-Werror" + self._get_precision_opts()
        self.default_build_opts += self._get_precision_opts()

        # Kernels configuration dictionary
        if self.device.name == "Cayman":
            from hysop.backend.device.opencl.config_cayman import kernels_config as kernel_cfg
        elif self.device.name == "Tesla K20m" or \
                self.device.name == "Tesla K20Xm":
            from hysop.backend.device.opencl.config_k20m import kernels_config as kernel_cfg
        else:
            print("/!\\ Get a defautl kernels config for", self.device.name)
            from hysop.backend.device.opencl.config_default import kernels_config as kernel_cfg
        self.kernels_config = kernel_cfg
        self._locMem_Buffers = {}

    def modify(self, platform_id, device_id, device_type,
               precision, gl_sharing=False):
        """Modify OpenCL environment parameters.

        Parameters
        ----------
        platform_id : int
            chosen platform id.
        device_id : int
            chosen device id.
        device_type : string
            chosen device type.
        precision : int
            required precision for real data.
            Must be FLOAT_GPU or DOUBLE_GPU.
        gl_sharing : bool, optional
            True to build a context shared between OpenGL and OpenCL.
            Default=False.
        """
        platform_changed, device_changed = False, False
        if not platform_id == self._platform_id:
            print("platform changed")
            self._platform_id = platform_id
            self.platform = self._get_platform(platform_id)
            platform_changed = True
        if platform_changed or not (device_id is self._device_id and
                                    device_type == self._device_type):
            print("device changed")
            self._device_id = device_id
            self._device_type = device_type
            self.device = self._get_device(self.platform,
                                           device_id, device_type)
            self.available_mem = self.device.global_mem_size
            device_changed = True
        if platform_changed or device_changed or \
                (not self._gl_sharing and gl_sharing is not self._gl_sharing):
            if self._gl_sharing and not gl_sharing:
                print("Warning: Loosing Gl shared context.")
            self._gl_sharing = gl_sharing
            self.ctx = self._get_context(self.device, gl_sharing)
            self.queue = self._get_queue(self.ctx)
        if self.precision is not precision and precision is not None:
            if self.precision is not None:
                print("Warning, GPU precision is overrided from",)
                print(self.precision, 'to', precision)
            self.precision = precision
            self.default_build_opts = ""
            if CL_PROFILE and self.device.vendor.find('NVIDIA') >= 0:
                self.default_build_opts += " -cl-nv-verbose"
            self.default_build_opts += "-Werror" + self._get_precision_opts()
        
        ## update opencl typegen
        # Floating point codegeneration mode
        from hysop.backend.device.opencl.opencl_types import OpenClTypeGen
        _kargs = {'device':self.device, 'context':self.ctx, 'platform':self.platform }
        if __KERNEL_DEBUG__:
            _kargs['float_dump_mode'] = 'dec'
        else:
            _kargs['float_dump_mode'] = 'hex'

        # Floating point precision
        if precision is FLOAT_GPU:
            _kargs['fbtype'] = 'float'
        elif precision is DOUBLE_GPU:
            _kargs['fbtype'] = 'double'
        else:
            raise ValueError('Unknown floating point precision {}!'.format(precision))

        typegen = OpenClTypeGen(**_kargs)
    
        self.typegen   = typegen
        self.prec_size = typegen.FLT_BYTES[typegen.fbtype]

    @staticmethod
    def _get_platform(platform_id):
        """Returns an OpenCL platform
        :param platform_id : OpenCL platform id

        """
        try:
            # OpenCL platform
            platform = cl.get_platforms()[platform_id]
        except IndexError:
            plist = cl.get_platforms()
            platform = plist[0]
            print("  Incorrect platform_id :", platform_id, ".",)
            print(" Only ", len(plist), " available.",)
            print(" --> getting default platform ", platform.name)
        if __VERBOSE__:
            print("  Platform   ")
            print("  - Name       :", platform.name)
            print("  - Version    :", platform.version)
        return platform

    @staticmethod
    def _get_device(platform, device_id, device_type):
        """Returns an OpenCL device

        Parameters
        ----------
        platform_id : int
            chosen platform id.
        device_id : int
            chosen device id.
        device_type : string
            chosen device type.

        Try to use given parameters and in case of fails, use pyopencl context
        creation function.
        """
        display = False
        try:
            if device_type is not None:
                device_type_id = cl.device_type.__getattribute__(
                    cl.device_type, str(device_type.upper()))
                device = platform.get_devices(device_type_id)[device_id]
            else:
                device = platform.get_devices()[device_id]
        except cl.RuntimeError as e:
            print("RuntimeError:", e)
            device = cl.create_some_context().devices[0]
            display = True
        except AttributeError as e:
            print("AttributeError:", e)
            device = cl.create_some_context().devices[0]
            display = True
        except IndexError:
            print("  Incorrect device_id :", device_id, ".",)
            print(" Only ", len(platform.get_devices()), " available.",)
            if device_type is not None:
                device_type = str(device_type.upper())
                print(" Getting first device of type " + device_type)
            else:
                print(" Getting first device of the platform")
            device = platform.get_devices()[0]
            display = True
        if device_type is not None:
            assert device_type.upper() == cl.device_type.to_string(device.type)
        if display or __VERBOSE__:
            print("  Device")
            print("  - id                :", device_id)
            print("  - Name                :",)
            print(device.name)
            print("  - Type                :",)
            print(cl.device_type.to_string(device.type))
            print ("  - C Version           :",)
            print (device.opencl_c_version)
            print ("  - Global mem size     :",)
            print(device.global_mem_size / (1024 ** 3), "GB")
        return device

    def _get_context(self, device, gl_sharing):
        """Returns OpenCL context

        Parameters
        ----------
        device : OpenCL device
            which handles the context.
        gl_sharing : bool
            True to build a context shared between OpenGL and OpenCL.
            Default=False.

        """
        props = None
        if gl_sharing:
            from pyopencl.tools import get_gl_sharing_context_properties
            import sys
            if sys.platform == "darwin":
                props = get_gl_sharing_context_properties()
            else:
                # Some OSs prefer clCreateContextFromType, some prefer
                # clCreateContext. Try both.
                props = \
                    [(cl.context_properties.PLATFORM, self.platform)] \
                    + get_gl_sharing_context_properties()
            ctx = cl.Context(properties=props, devices=[device])
        else:
            ctx = cl.Context([device])
        if __VERBOSE__:
            print(" Context:")
            if props is not None:
                print("  - properties           :", props)
        return ctx

    @staticmethod
    def _get_queue(ctx):
        """Returns OpenCL queue from context

        :param ctx : OpenCL context

        """
        props = None
        if CL_PROFILE:
            props = cl.command_queue_properties.PROFILING_ENABLE
            queue = cl.CommandQueue(ctx, properties=props)
        else:
            queue = cl.CommandQueue(ctx)
        if __VERBOSE__:
            print(" Queue")
            if props is not None:
                print("  - properties           :", props)
            print("===")
        return queue

    def create_other_queue(self):
        """Create OpenCL queue from current context
        """
        return self._get_queue(self.ctx)

    @staticmethod
    def get_work_items(resolution, vector_width=1):
        """Set the optimal work-item number and OpenCL space index.

        Parameters
        ----------
        resolution : tuple
            local mesh resolution
        vector_width : int
            OpenCL vector types width

        Returns
        -------
        int : work-item number
        tuple : global space index
        tuple : local space index

        Use 64 work-items in 3D and 256 in 2D.
        \todo Use Both the number from device capability
        The problem must be a multiple of and greater
        than work-item number * vector_width
        """
        # Optimal work item number
        if len(resolution) == 3:
            workItemNumber = 64 if min(resolution) >= 64 \
                else min(resolution)
        else:
            workItemNumber = 256 if min(resolution) >= 256 \
                else min(resolution)
        # Change work-item regarding problem size
        if resolution[0] % workItemNumber > 0:
            if len(resolution) == 3:
                print("Warning : GPU best performances obtained for",)
                print("problem sizes multiples of 64")
            else:
                print("Warning : GPU best performances obtained for",)
                print("problem sizes multiples of 256")
        while resolution[0] % workItemNumber > 0:
            workItemNumber = workItemNumber / 2
        # Change work-item regarding vector_width
        if workItemNumber * vector_width > resolution[0]:
            if resolution[0] % vector_width > 0:
                raise ValueError(
                    "Resolution ({0}) must be a multiple of {1}".format(
                        resolution[0], vector_width))
            workItemNumber = resolution[0] // vector_width
        if len(resolution) == 3:
            gwi = (int(workItemNumber),
                   int(resolution[1]), int(resolution[2]))
            lwi = (int(workItemNumber), 1, 1)
        else:
            gwi = (int(workItemNumber),
                   int(resolution[1]))
            lwi = (int(workItemNumber), 1)
        return workItemNumber, gwi, lwi

    def _get_precision_opts(self):
        """Check if device is capable to work with given precision
        and returns build options considering this precision
        """
        opts = ""
        # Precision supported
        fp32_rounding_flag = True
        if self.precision is FLOAT_GPU:
            opts += " -cl-single-precision-constant"
            prec = "single"
        else:
            if self.device.double_fp_config <= 0:
                raise ValueError("Double Precision is not supported by device")
            prec = "double"
        if __VERBOSE__:
            print(" Precision capability  ",)
            print("for " + prec + " Precision: ")
        for v in ['DENORM', 'INF_NAN',
                  'ROUND_TO_NEAREST', 'ROUND_TO_ZERO', 'ROUND_TO_INF',
                  'FMA', 'CORRECTLY_ROUNDED_DIVIDE_SQRT', 'SOFT_FLOAT']:
            try:
                if eval('(self.device.' + prec + '_fp_config &' +
                        ' cl.device_fp_config.' +
                        v + ') == cl.device_fp_config.' + v):
                    if __VERBOSE__:
                        print(v)
                else:
                    if v is 'CORRECTLY_ROUNDED_DIVIDE_SQRT':
                        fp32_rounding_flag = False
            except AttributeError as ae:
                if v is 'CORRECTLY_ROUNDED_DIVIDE_SQRT':
                    fp32_rounding_flag = False
                if __VERBOSE__:
                    print(v, 'is not supported in OpenCL C 1.2.\n',)
                    print('   Exception catched : ', ae)
        if fp32_rounding_flag:
            opts += " -cl-fp32-correctly-rounded-divide-sqrt"
        return opts

    def _create_cl_program(self, file_list, 
                           vector_width=4,
                           nb_remesh_components=1,
                           options="",
                           force_verbose=None, force_debug=None):
        """Build OpenCL sources

        Parameters
        ----------
        files : string or list of strings
            user defined files names
        vector_width : int, optional
            OpenCL vector type width, default=4
        nb_remesh_components : int, optional
            number of remeshed components, default=1
        options: string
            additional OpenCL compile flags
        force_verbose: bool, optional, default=None
            force verbose mode
        force_debug: bool, optional, default=None
            force debug mode (kernel source dumping and preprocessing)
        
        Returns OpenCL kernel
        Parse the sources to handle single and double precision.
        """
        VERBOSE = __VERBOSE__       if (force_verbose is None) else force_verbose
        DEBUG   = __KERNEL_DEBUG__  if (force_debug is None)   else force_debug

        gpu_src = ""
        if cl.device_type.to_string(self.device.type) == 'GPU' and \
                self.precision is DOUBLE_GPU:
            gpu_src += '#pragma OPENCL EXTENSION cl_khr_fp64: enable \n'
        
        if isinstance(files, list):
            file_list = files
        else:
            file_list = [files]

        if VERBOSE:
            print '=== Kernel sources compiling ==='
            for sd in file_list:
                print '  - ', sf

        for sf in file_list:
            # search and open cl file.
            try:
                f = open(sf, 'r')
            except IOError as ioe:
                if ioe.errno == 2:
                    # path to cl files inside hysop.gpu package
                    f = open(GPU_SRC + sf, 'r')
                else:
                    raise ioe
            gpu_src += "".join(
                self.parse_file(f, vector_width, nb_remesh_components))
            f.close()
        
        # print gpu_src
        if self.macros is not None:
            for k in self.macros:
                gpu_src = gpu_src.replace(k, str(self.macros[k]))
        
        if self.precision is FLOAT_GPU:
            # Rexexp to add 'f' suffix to float constants
            # Match 1.2, 1.234, 1.2e3, 1.2E-05
            float_replace = re.compile(r'(?P<float>\d\.\d+((e|E)-?\d+)?)')
            gpu_src = float_replace.sub(r'\g<float>f', gpu_src)
        else:
            gpu_src = gpu_src.replace('float', 'double')
       
        # Log final opencl generated code for debug purposes
        if DEBUG:
            kernel_name = (file_list[-1].split('/')[-1]).replace('.cl','_parsed')
            def listformat(L):
                if isinstance(L,str):
                    L=L.replace('-D ','').split(' ')
                L=list(L)
                for empty in ['', ' ']:
                    if empty in L:
                        L.remove(empty)
                return '\n\t\t'+'\n\t\t'.join(L)
            dump_prefix = \
'''
/* 
Dumped OpenCL Kernel '{}'
    vector_width: {}
    nb_remesh_components: {}
    source_files: {}
    default_build_opts: {}
    additional options: {}
*/
'''.format(kernel_name,
           vector_width, 
           nb_remesh_components, 
           listformat(file_list),
           listformat(self.default_build_opts), 
           listformat(options))
   

            dumped_src = dump_prefix + gpu_src
    
            dump_folder=IO.default_path()+'/'+KERNEL_DUMP_FOLDER
            dump_file_prefix=dump_folder+'/'+kernel_name
            tmp_dump_file=dump_file_prefix+'.c'
            dump_file=dump_file_prefix+'.cl'
            if not os.path.exists(dump_folder):
                os.makedirs(dump_folder)
            with open(tmp_dump_file, 'w+') as f:
                f.write(dumped_src)

            try:
                #try to preprocess sources
                import subprocess
                opts = self.default_build_opts + options
                opts = re.sub('-cl-([a-z0-9]+-?)+ ','',opts)
                cmd = ['gcc',opts,'-E','-c',tmp_dump_file,'-o',dump_file_prefix+'_preprocessed.cl']
                subprocess.check_call(' '.join(cmd), shell=True);  
            finally:
                os.rename(tmp_dump_file,dump_file)

            if VERBOSE:
                msg = 'OpenCL kernel {} source dumped to {}.'.format(kernel_name, dump_file)
                print msg

        # OpenCL program
        prg = cl.Program(self.ctx, gpu_src)
        return prg

    def build_src(self, files, 
                  options="", 
                  vector_width=4,
                  nb_remesh_components=1):
        """Build OpenCL sources

        Parameters
        ----------
        files : string or list of strings
            user defined file names
        options : string, optional
            Compiler options, default=""
        vector_width : int, optional
            OpenCL vector type width, default=4
        nb_remesh_components : int, optional
            number of remeshed components, default=1
        force_verbose: bool, optional
            force verbose mode
        force_debug: bool, optional
            force debug mode (kernel dumping)

        Returns OpenCL binaries

        Parse the sources to handle single and double precision.
        """

        if isinstance(files, list):
            file_list = files
        else:
            file_list = [files]

        if __VERBOSE__:
            print("=== Kernel sources compiling ===")
            for sf in file_list:
                print("   - ", sf)

        # --- create kernel from cl files ---
        prg = self._create_cl_program(files=file_list, 
                                      options=options,
                                      vector_width=vector_width,
                                      nb_remesh_components=nb_remesh_components,
                                      force_verbose=force_verbose,
                                      force_debug=force_debug)
        
        # --- Build kernel ---
        try:
            build = prg.build(self.default_build_opts + options)
        except Exception, e:
            print("Build files : ")
            for sf in file_list:
                print("   - ", sf)
            print("Build options : ", self.default_build_opts + options)
            print("Vectorization : ", vector_width)
            raise e

        # display post-build info
        if __VERBOSE__:
            # print options
            print "Build options : ",
            print build.get_build_info(
                self.device, cl.program_build_info.OPTIONS)
            print "Compiler status : ",
            print build.get_build_info(
                self.device, cl.program_build_info.STATUS)
            print "Compiler log : ",
            print build.get_build_info(self.device,
                             cl.program_build_info.LOG)
            print "===\n"
        return build
    
    def build_raw_src(self, src, options="", 
            kernel_name=None,
            force_verbose=None, force_debug=None):
        """Build raw OpenCL sources

        Parameters
        ----------
        src : string 
            OpenCL source code
        options : string
            Compiler options to use for buildind
        Returns OpenCL binaries
        """
        VERBOSE = __VERBOSE__      if (force_verbose is None) else force_verbose
        DEBUG   = __KERNEL_DEBUG__ if (force_debug is None)   else force_debug

        gpu_src = src
        
        src_hash = hashlib.sha1(gpu_src).hexdigest()
        if (kernel_name is None):
            kernel_name = src_hash
        else:
            kernel_name += '_{}'.format(src_hash[:4])

        soptions = ' '.join(options)
        if VERBOSE:
            print "=== Kernel raw source compiling ==="
        prg = cl.Program(self.ctx, gpu_src)
            
            
        dump_folder=IO.default_path()+'/'+KERNEL_DUMP_FOLDER
        if not os.path.exists(dump_folder):
            os.makedirs(dump_folder)

        # Build OpenCL program
        try:
            build = prg.build(self.default_build_opts + soptions)
        except Exception, e:
            # always dump source when build fails
            dump_file=dump_folder+'/'+'{}_build_fail.cl'.format(kernel_name)
            with open(dump_file, 'w+') as f:
                f.write(gpu_src)
            print "Build options : ", self.default_build_opts + soptions
            print 'Build Failed: dumped source to {}.'.format(dump_file) 
            raise e

        if VERBOSE:
            #print options
            print "Build options : ",
            print build.get_build_info(
                self.device, cl.program_build_info.OPTIONS)
            print "Compiler status : ",
            print build.get_build_info(
                self.device, cl.program_build_info.STATUS)
            print "Compiler log : ",
            print build.get_build_info(self.device,
                                       cl.program_build_info.LOG)
        if DEBUG:
            # dump kernel source while in debug mode
            dump_file=dump_folder+'/'+'{}_dump.cl'.format(kernel_name)
            print 'Dumping kernel src at \'{}\'.'.format(dump_file)
            with open(dump_file, 'w+') as f:
                f.write(gpu_src)

        if VERBOSE:
            print "===\n"

        return build

    @staticmethod
    def parse_file(f, n=8, nb_remesh_components=1):
        """Parse a file containing OpenCL sources.

        Parameters
        ----------
        f : string
            file name
        n : int, optional
            vector width, default=8
        nb_remesh_components : int
            number of remeshed components

        Returns
        -------
        string, the parsed sources.

        Notes
        -----
        * __N__ is expanded as an integer corresponding to vector width.
        * __NN__ instruction is duplicated to operate on each vector component:

          * if line ends with ';', the whole instruciton is
            duplicated.
          * if line ends with ',' and contains
            '(float__N__)(', the float element is duplicated

        * Remeshing fields components are expanded as follows :
          All code between '__RCOMPONENT_S__' and
          '__RCOMPONENT_E__' flags are duplicated n times with n
          the number of components to compute. In this duplicated code, the
          flag '__ID__' is replaced by index of a range of lenght
          the number of components. A flag '__RCOMPONENT_S__P__'
          may be used and the duplicated elements are separated with ','
          (for function parameters expanding).

        Examples with a 4-width vector code::

            float__N__ x;           ->  float4 x;

            x.s__NN__ = 1.0f;       ->  x.s0 = 1.0f;
                                        x.s1 = 1.0f;
                                        x.s2 = 1.0f;
                                        x.s3 = 1.0f;

            x = (int__N__)(__NN__,  ->  x = (int4)(0,
                            );                      1,
                                                    2,
                                                    3,
                                                    );

        Examples with a 2 components expansion code::

            __RCOMP_P __global const float* var__ID__,
            -> __global const float* var0,__global const float* var1,

            __RCOMP_I var__ID__[i] = 0.0;
            -> var0[i] = 0.0;var1[i] = 0.0;

            aFunction(__RCOMP_P var__ID__, __RCOMP_P other__ID__);
            -> aFunction(var0, var1, other0, other1);

        """
        src = ""
        # replacement for floatN elements
        vec_floatn = re.compile(r'\(float__N__\)\(')
        vec_nn = re.compile('__NN__')
        vec_n = re.compile('__N__')
        for l in f.readlines():
            # Expand floatN items
            if vec_floatn.search(l) and vec_nn.search(l) and \
                    l[-2] == ',':
                sl = l.split("(float__N__)(")
                l = sl[0] + "(float" + str(n) + ")("
                el = sl[1].rsplit(',', 1)[0]
                for i in xrange(n):
                    l += vec_nn.sub(str(i), el) + ','
                l = l[:-1] + '\n'
            # Expand floatN elements access
            elif vec_nn.search(l) and l[-2] == ';':
                el = ""
                for i in xrange(n):
                    el += vec_nn.sub(str(i), l)
                l = el
            # Replace vector length
            src += vec_n.sub(str(n), l)

        # Replacement for remeshed components
        re_instr = re.compile(r'__RCOMP_I([\w\s\.,()\[\]+*/=-]+;)')
        # __RCOMP_I ...;

        def repl_instruction(m):
            return ''.join(
                [m.group(1).replace('__ID__', str(i))
                 for i in xrange(nb_remesh_components)])
        # __RCOMP_P ..., ou __RCOMP_P ...)
        re_param = re.compile(r'__RCOMP_P([\w\s\.\[\]+*/=-]+(?=,|\)))')

        def repl_parameter(m):
            return ', '.join(
                [m.group(1).replace('__ID__', str(i))
                 for i in xrange(nb_remesh_components)])

        src = re_instr.sub(repl_instruction, src)
        src = re_param.sub(repl_parameter, src)
        return src

    def global_allocation(self, array):
        """Allocate and returns an opencl buffer

        Parameters
        ----------
        array : numpy array
            source buffer, on host
        """
        # create an opencl buffer from input array
        clBuff = cl.Buffer(self.ctx,
                           cl.mem_flags.ALLOC_HOST_PTR, size=array.nbytes)
        # Touch the buffer on device to performs the allocation
        # Transfers a single element in device (the precision does not matter)
        e = np.zeros((1,), dtype=np.float64)
        cl.enqueue_copy(self.queue, clBuff, e,
                        buffer_origin=(0, 0, 0), host_origin=(0, 0, 0),
                        region=(e.nbytes,)).wait()
        # update memory counter
        self.available_mem -= clBuff.size
        return clBuff

    def global_deallocation(self, cl_mem):
        self.available_mem += cl_mem.size
        cl_mem.release()

    # def LocalMemAllocator(self, sizes_list, type_list=None):
    #     """
    #     Allocates spaces in device local memory.
    #     @param sizes_list : list of sizes.
    #     @param type_list : list of corresponding types
    #     It returns a list of buffers of given size (one per size specified in
    #     in the list) and the size of new buffers.
    #     @remark : Buffers are stored and could be reused.
    #     @remark : it assumes that all returned buffers are different
    #     """
    #     new_alloc = 0
    #     if type_list is None:
    #         type_list = [HYSOP_REAL] * len(sizes_list)
    #     buff_list = []  # Returned list
    #     keys_list = []
    #     for s, t in zip(sizes_list, type_list):
    #         keys_list.append(int(t(0).nbytes * s))

    #     for size, key, t in zip(sizes_list, keys_list, type_list):
    #         buff = None
    #         try:
    #             # List of existing buffers not already in the list
    #             avail_buff = [b for b in self._locMem_Buffers[key]
    #                           if b not in buff_list]
    #             if len(avail_buff) > 0:
    #                 # adding the first buffer
    #                 buff = avail_buff[0]
    #             else:
    #                 # Allocate a new buffer
    #                 buff = cl.LocalMemory(int(t(0).nbytes * size))
    #                 new_alloc += buff.size
    #                 self._locMem_Buffers[key].append(buff)
    #         except KeyError:
    #             # Allocate a fist buffer of given size
    #             buff = cl.LocalMemory(int(t(0).nbytes * size))
    #             new_alloc += buff.size
    #             self._locMem_Buffers[key] = [buff]
    #         buff_list.append(buff)
    #     return buff_list, new_alloc


def get_opengl_shared_environment(platform_id=None,
                                  device_id=None,
                                  device_type=None, precision=HYSOP_REAL,
                                  comm=None):
    """Build or get an OpenCL environment with openGL properties.

    Parameters
    ----------
    platform_id : int
        chosen platform id.
    device_id : int
       chosen device id.
    device_type : string
       chosen device type.
    precision : int, optional
       required precision for real data.
       Default : HYSOP_REAL
    comm : mpi communicator, optional
        Communicator which handles the OpenCL env.
        Default = hysop.core.mpi.main_comm

    Returns
    -------
    :class:`~hysop.gpu.tools.OpenClEnvironment`
        object handling OpenCL platform, device, context and queue

    The context is obtained with gl-shared properties depending on the OS.
    """
    if platform_id is None:
        platform_id = __DEFAULT_PLATFORM_ID__
    if device_id is None:
        device_id = __DEFAULT_DEVICE_ID__
    global __cl_env
    if __cl_env is None:
        __cl_env = OpenClEnvironment(platform_id, device_id, device_type,
                                     precision, gl_sharing=True, comm=comm)
    else:
        __cl_env.modify(platform_id, device_id, device_type,
                        precision, gl_sharing=True)
    return __cl_env


def get_opencl_environment(platform_id=None,
                           device_id=None,
                           device_type=None, precision=HYSOP_REAL,
                           comm=None):
    """Build or get an OpenCL environment.

    Parameters
    ----------
    platform_id : int
        chosen platform id.
    device_id : int
       chosen device id.
    device_type : string
       chosen device type.
    precision : int, optional
       required precision for real data.
       Default : HYSOP_REAL
    comm : mpi communicator, optional
        Communicator which handles the OpenCL env.
        Default = hysop.core.mpi.main_comm

    Returns
    -------

    :class:`~hysop.gpu.tools.OpenClEnvironment`
        object handling OpenCL platform, device, context and queue

    """
    if platform_id is None:
        platform_id = __DEFAULT_PLATFORM_ID__
    if device_id is None:
        device_id = __DEFAULT_DEVICE_ID__
    global __cl_env
    if __cl_env is None:
        __cl_env = OpenClEnvironment(platform_id, device_id, device_type,
                                     precision, comm=comm)
    else:
        __cl_env.modify(platform_id, device_id, device_type,
                        precision)
    return __cl_env


def explore():
    """Scan system and print OpenCL environment details"""
    print("OpenCL exploration : ")
    platforms = cl.get_platforms()
    platforms_info = ["name", "version", "vendor", "profile", "extensions"]
    devices_info = ["name",
                    "version",
                    "vendor",
                    "profile",
                    "extensions",
                    "available",
                    "type",
                    "compiler_available",
                    "double_fp_config",
                    "single_fp_config",
                    "global_mem_size",
                    "global_mem_cache_type",
                    "global_mem_cache_size",
                    "global_mem_cacheline_size",
                    "local_mem_size",
                    "local_mem_type",
                    "max_clock_frequency",
                    "max_compute_units",
                    "max_constant_buffer_size",
                    "max_mem_alloc_size",
                    "max_work_group_size",
                    "max_work_item_dimensions",
                    "max_work_item_sizes",
                    "preferred_vector_width_double",
                    "preferred_vector_width_float",
                    "preferred_vector_width_int"]
    for pltfm in platforms:
        print("Platform:", pltfm.name)
        for pltfm_info in platforms_info:
            print("  |-", pltfm_info, ':', eval("pltfm." + pltfm_info))
        devices = pltfm.get_devices()
        for dvc in devices:
            print("  |- Device:", dvc.name)
            for dvc_info in devices_info:
                print("    |-", dvc_info, ':', eval("dvc." + dvc_info))
