# -*- coding: utf-8 -*-
from hysop import Field, Box
import numpy as np
import hysop.numerics.differential_operations as diffop
from hysop.tools.numpywrappers import npw
from hysop.tools.parameters import Discretization
from hysop.domain.subsets import SubBox
import math as m
pi = m.pi
cos = np.cos
sin = np.sin


def compute_vel(res, x, y, z, t):
    res[0][...] = sin(x) * cos(y) * cos(z)
    res[1][...] = - cos(x) * sin(y) * cos(z)
    res[2][...] = 0.
    return res


def compute_vort(res, x, y, z, t):
    res[0][...] = - cos(x) * sin(y) * sin(z)
    res[1][...] = - sin(x) * cos(y) * sin(z)
    res[2][...] = 2. * sin(x) * sin(y) * cos(z)
    return res


def compute_vel2(res, x, y, t):
    res[0][...] = sin(x) * cos(y)
    res[1][...] = - cos(x) * sin(y)
    return res


def compute_vort2(res, x, y, t):
    res[0][...] = 2. * sin(x) * sin(y)
    return res


def analyticalDivRhoV(res, x, y, z, t):
    res[0][...] = cos(y) * cos(z) * sin(y) * sin(z)
    return res


def analyticalDivRhoV2D(res, x, y, t):
    res[0][...] = 0.
    return res


def analyticalDivWV(res, x, y, z, t):
    res[0][...] = - cos(y) * cos(z) * sin(y) * sin(z)
    res[1][...] = cos(x) * cos(z) * sin(z) * sin(x)
    res[2][...] = 0.
    return res


def analyticalGradVxW(res, x, y, z, t):
    res[0][...] = - sin(y) * cos(y) * sin(z) * cos(z)
    res[1][...] = sin(x) * cos(x) * sin(z) * cos(z)
    res[2][...] = 0.
    return res


def analyticalDivStressTensor(res, x, y, z, t):
    res[0][...] = - 3. * sin(x) * cos(y) * cos(z)
    res[1][...] = 3. * cos(x) * sin(y) * cos(z)
    res[2][...] = 0.
    return res

def analyticalDivAdvection(res, x, y, z, t):
    res[0][...] = - cos(z) * cos(z) * (cos(x) * cos(x) - sin(x) * sin(x)) - \
        cos(z) * cos(z) * (cos(y) * cos(y) - sin(y) * sin(y))
    return res


Nx = 88
Ny = 50
Nz = 64
g = 2
discr3D = Discretization([Nx + 1, Ny + 1, Nz + 1], [g, g, g])
discr2D = Discretization([Nx + 1, Ny + 1], [g, g])
ldom = npw.asrealarray([2., pi, 3.5])
xdom = [0., -0.3, 0.4]


def init(discr, vform, wform):
    dim = len(discr.resolution)
    dom = Box(dimension=dim, length=ldom[:dim],
              origin=xdom[:dim])
    topo = dom.create_topology(discr)
    work = []
    shape = tuple(topo.mesh.local_resolution)
    subsize = np.prod(shape)
    for _ in xrange(6):
        work.append(npw.zeros(subsize,))
    velo = Field(domain=dom, formula=vform,
                 name='Velocity', is_vector=True)
    vorti = Field(domain=dom, formula=wform,
                  name='Vorticity', is_vector=dim == 3)
    vd = velo.discretize(topo)
    wd = vorti.discretize(topo)
    velo.initialize(topo=topo)
    vorti.initialize(topo=topo)
    return vd, wd, topo, work


# Init 2D and 3D
v3d, w3d, topo3, work3 = init(discr3D, compute_vel, compute_vort)
v2d, w2d, topo2, work2 = init(discr2D, compute_vel2, compute_vort2)


def build_op(op_class, len_result, topo, work=None, reduced=False,
             reduce_output=False):
    if reduced:
        # computations will be done on a subset of the domain
        sl = topo.domain.length * 0.5
        orig = topo.domain.origin + 0.1 * topo.domain.length
        subbox = SubBox(parent=topo.domain, origin=orig, length=sl)
        indices = subbox.discretize(topo)[0]
        if reduce_output:
            outputshape = subbox.mesh[topo].resolution
        else:
            outputshape = np.asarray(topo.mesh.local_resolution).copy()
    else:
        indices = None
        if reduce_output:
            outputshape = np.asarray(topo.mesh.local_resolution).copy()
            gh = topo.ghosts()
            outputshape -= 2 * gh
            outputshape = tuple(outputshape)
        else:
            outputshape = tuple(topo.mesh.local_resolution)

    if work is not None:
        wk_prop = op_class.get_work_properties(topo, indices)
        if wk_prop['rwork'] is not None:
            lwork = len(wk_prop['rwork'])
        else:
            lwork = 0
        op = op_class(topo=topo, indices=indices, work=work[:lwork],
                      reduce_output_shape=reduce_output)
    else:
        op = op_class(topo=topo, indices=indices,
                      reduce_output_shape=reduce_output)
    result = [npw.zeros(outputshape) for _ in xrange(len_result)]
    return op, result


def assert_curl(wdim, topo, velo, vorti, work=None, reduced=False,
                reduce_output=False):
    op, result = build_op(diffop.Curl, wdim, topo, work, reduced,
                          reduce_output)
    ic_out = op.output_indices
    ic_in = op.in_indices
    result = op(velo.data, result)
    atol = np.max(topo.mesh.space_step ** 2)
    for i in xrange(wdim):
        assert np.allclose(vorti.data[i][ic_in], result[i][ic_out], atol=atol)


def test_curl():
    # 2d
    assert_curl(1, topo2, v2d, w2d)
    # 2d work
    assert_curl(1, topo2, v2d, w2d, work2)
    # 2d reduced
    assert_curl(1, topo2, v2d, w2d, reduced=True)
    # 2d reduced and work
    assert_curl(1, topo2, v2d, w2d, work2, reduced=True)
    # 2d reduce output
    assert_curl(1, topo2, v2d, w2d, reduce_output=True)
    # 2d reduce output and work
    assert_curl(1, topo2, v2d, w2d, work2, reduce_output=True)
    # 2d reduce output, on subset, with work
    assert_curl(1, topo2, v2d, w2d, work2, reduced=True, reduce_output=True)
    # 3d
    assert_curl(3, topo3, v3d, w3d)
    # 3d work
    assert_curl(3, topo3, v3d, w3d, work3)
    # 3d reduced
    assert_curl(3, topo3, v3d, w3d, reduced=True)
    # 3d reduced and work
    assert_curl(3, topo3, v3d, w3d, work3, reduced=True)
    # 3d reduce output
    assert_curl(3, topo3, v3d, w3d, reduce_output=True)
    # 3d reduce output and work
    assert_curl(3, topo3, v3d, w3d, work3, reduce_output=True)
    # 3d, work, reduce output, on subset
    assert_curl(3, topo3, v3d, w3d, work3, reduced=True, reduce_output=True)


def assert_div_rho_v(topo, velo, scal, formula, work=None, reduced=False,
                     reduce_output=False):
    ref = Field(domain=topo.domain, formula=formula,
                name='Analytical', is_vector=False)
    rd = ref.discretize(topo)
    ref.initialize(topo=topo)
    op, result = build_op(diffop.DivRhoV, 1, topo, work, reduced,
                          reduce_output)
    ic_out = op.output_indices
    ic_in = op.in_indices
    tol = np.max(topo.mesh.space_step ** 4)
    result = op(velo.data, [scal], result)
    assert np.allclose(rd[0][ic_in], result[0][ic_out], atol=tol)


def test_div_rho_v():
    # 2d
    assert_div_rho_v(topo2, v2d, w2d[0], analyticalDivRhoV2D)
    # 2d work
    assert_div_rho_v(topo2, v2d, w2d[0], analyticalDivRhoV2D, work2)
    # 2d reduced
    assert_div_rho_v(topo2, v2d, w2d[0], analyticalDivRhoV2D, reduced=True)
    # 2d reduced and work
    assert_div_rho_v(topo2, v2d, w2d[0], analyticalDivRhoV2D, work2,
                     reduced=True)
    # 2d reduce output
    assert_div_rho_v(topo2, v2d, w2d[0], analyticalDivRhoV2D,
                     reduce_output=False)
    # 2d reduce output and work
    assert_div_rho_v(topo2, v2d, w2d[0], analyticalDivRhoV2D, work2,
                     reduce_output=False)
    # 2d reduce output, on subset, with work
    assert_div_rho_v(topo2, v2d, w2d[0], analyticalDivRhoV2D, work2,
                     reduce_output=True, reduced=True)
    # 3d
    assert_div_rho_v(topo3, v3d, w3d[0], analyticalDivRhoV)
    # 3d work
    assert_div_rho_v(topo3, v3d, w3d[0], analyticalDivRhoV, work3)
    # 3d reduced
    assert_div_rho_v(topo3, v3d, w3d[0], analyticalDivRhoV, reduced=True)
    # 3d reduced and work
    assert_div_rho_v(topo3, v3d, w3d[0], analyticalDivRhoV, work3,
                     reduced=True)
    # 3d reduce output
    assert_div_rho_v(topo3, v3d, w3d[0], analyticalDivRhoV,
                     reduce_output=False)
    # 3d reduce output and work
    assert_div_rho_v(topo3, v3d, w3d[0], analyticalDivRhoV, work3,
                     reduce_output=False)
    # 3d, work, reduce output, on subset
    assert_div_rho_v(topo3, v3d, w3d[0], analyticalDivRhoV, work3,
                     reduce_output=False, reduced=True)


def assert_div_w_v(wdim, topo, velo, vorti, formula, work=None, reduced=False,
                   reduce_output=False):
    # Reference field
    ref = Field(domain=topo.domain, formula=formula,
                name='Analytical', is_vector=True)
    rd = ref.discretize(topo)
    ref.initialize(topo=topo)
    op, result = build_op(diffop.DivWV, wdim, topo, work, reduced,
                          reduce_output)
    result = op(vorti.data, velo.data, result)
    ic_out = op.output_indices
    ic_in = op.in_indices

    # Numerical VS analytical
    tol = np.max(topo.mesh.space_step ** 2)
    for i in xrange(wdim):
        assert np.allclose(rd[i][ic_in], result[i][ic_out], atol=tol)


def test_div_w_v():
    # 3d
    assert_div_w_v(3, topo3, v3d, w3d, analyticalDivWV)
    # 3d work
    assert_div_w_v(3, topo3, v3d, w3d, analyticalDivWV, work3)
    # 3d reduced
    assert_div_w_v(3, topo3, v3d, w3d, analyticalDivWV, reduced=True)
    # 3d reduced and work
    assert_div_w_v(3, topo3, v3d, w3d, analyticalDivWV, work3,
                   reduced=True)
    # 3d reduce output
    assert_div_w_v(3, topo3, v3d, w3d, analyticalDivWV, reduce_output=True)
    # 3d reduce output and work
    assert_div_w_v(3, topo3, v3d, w3d, analyticalDivWV, work3,
                   reduce_output=True)
    # 3d, work, reduce output, on subset
    assert_div_w_v(3, topo3, v3d, w3d, analyticalDivWV, work3,
                   reduce_output=True, reduced=True)


def assert_laplacian(wdim, topo, vorti, formula, work=None, reduced=False,
                     reduce_output=False):
    ref = Field(domain=topo.domain, formula=formula,
                name='Analytical', is_vector=True)
    rd = ref.discretize(topo)
    ref.initialize(topo=topo)
    op, result = build_op(diffop.Laplacian, wdim, topo, work, reduced,
                          reduce_output)
    ic_out = op.output_indices
    ic_in = op.in_indices
    result = op(vorti.data, result)
    tol = np.max(topo.mesh.space_step ** 2)
    for i in xrange(wdim):
        assert np.allclose(rd.data[i][ic_in], result[i][ic_out], rtol=tol)


def laplacian_func(res, x, y, z, t):
    res[0][...] = 3 * cos(x) * sin(y) * sin(z)
    res[1][...] = 3 * sin(x) * cos(y) * sin(z)
    res[2][...] = -6. * sin(x) * sin(y) * cos(z)
    return res


def laplacian_func_2d(res, x, y, t):
    res[0][...] = -4. * sin(x) * sin(y)
    return res


def test_laplacian():
    # 2d
    assert_laplacian(1, topo2, w2d, laplacian_func_2d)
    # 2d work
    assert_laplacian(1, topo2, w2d, laplacian_func_2d, work2)
    # 2d reduced
    assert_laplacian(1, topo2, w2d, laplacian_func_2d, reduced=True)
    # 2d reduced and work
    assert_laplacian(1, topo2, w2d, laplacian_func_2d, work2,
                     reduced=True)
    # 2d reduce output
    assert_laplacian(1, topo2, w2d, laplacian_func_2d, reduce_output=True)
    # 2d reduce output and work
    assert_laplacian(1, topo2, w2d, laplacian_func_2d, work2,
                     reduce_output=True)
    # 2d, work, reduce output, on subset
    assert_laplacian(1, topo2, w2d, laplacian_func_2d, work2,
                     reduce_output=True, reduced=True)

    # 3d
    assert_laplacian(3, topo3, w3d, laplacian_func)
    # 3d work
    assert_laplacian(3, topo3, w3d, laplacian_func, work3)

    # 3d reduced
    assert_laplacian(3, topo3, w3d, laplacian_func, reduced=True)
    # 3d reduced and work
    assert_laplacian(3, topo3, w3d, laplacian_func, work3,
                     reduced=True)
    # 3d reduce output
    assert_laplacian(3, topo3, w3d, laplacian_func, reduce_output=True)
    # 3d reduce output and work
    assert_laplacian(3, topo3, w3d, laplacian_func, work3,
                     reduce_output=True)
    # 3d, work, reduce output, on subset
    assert_laplacian(3, topo3, w3d, laplacian_func, work3,
                     reduce_output=True, reduced=True)


def assert_grad_s(topo, scal, formula, work=None, reduced=False,
                  reduce_output=False):
    # Reference field
    ref = Field(domain=topo.domain, formula=formula,
                name='Analytical', is_vector=True)
    rd = ref.discretize(topo)
    ref.initialize(topo=topo)
    op, result = build_op(diffop.GradS, topo.domain.dim,
                          topo, work, reduced, reduce_output)
    ic_out = op.output_indices
    ic_in = op.in_indices
    result = op([scal], result)
    # Numerical VS analytical
    tol = np.max(topo.mesh.space_step ** 4)
    for i in xrange(topo.domain.dim):
        assert np.allclose(rd[i][ic_in], result[i][ic_out], atol=tol)


def grad_s_func_2d(res, x, y, t):
    res[0][...] = cos(x) * cos(y)
    res[1][...] = -sin(x) * sin(y)
    return res


def grad_s_func_3d(res, x, y, z, t):
    res[0][...] = cos(x) * cos(y) * cos(z)
    res[1][...] = -sin(x) * sin(y) * cos(z)
    res[2][...] = -sin(x) * cos(y) * sin(z)
    return res


def test_grad_s():
    # 2d
    assert_grad_s(topo2, v2d[0], grad_s_func_2d)
    # 2d work
    assert_grad_s(topo2, v2d[0], grad_s_func_2d, work2)
    # 2d reduced
    assert_grad_s(topo2, v2d[0], grad_s_func_2d, reduced=True)
    # 2d reduced and work
    assert_grad_s(topo2, v2d[0], grad_s_func_2d, work2,
                  reduced=True)
    # 2d reduce output
    assert_grad_s(topo2, v2d[0], grad_s_func_2d,
                  reduce_output=True)
    # 2d reduce output and work
    assert_grad_s(topo2, v2d[0], grad_s_func_2d, work2,
                  reduce_output=True)
    # 2d reduce output, on subset, with work
    assert_grad_s(topo2, v2d[0], grad_s_func_2d, work2,
                  reduce_output=True, reduced=True)
    # 3d
    assert_grad_s(topo3, v3d[0], grad_s_func_3d)
    # 3d work
    assert_grad_s(topo3, v3d[0], grad_s_func_3d, work3)
    # 3d reduced
    assert_grad_s(topo3, v3d[0], grad_s_func_3d, reduced=True)
    # 3d reduced and work
    assert_grad_s(topo3, v3d[0], grad_s_func_3d, work3,
                  reduced=True)
    # 3d reduce output
    assert_grad_s(topo3, v3d[0], grad_s_func_3d, reduce_output=True)
    # 3d reduce output and work
    assert_grad_s(topo3, v3d[0], grad_s_func_3d, work3,
                  reduce_output=True)
    # 3d, work, reduce output, on subset
    assert_grad_s(topo3, v3d[0], grad_s_func_3d, work3,
                  reduce_output=True, reduced=True)


def assert_grad_v(topo, velo, formula, work=None, reduced=False,
                  reduce_output=False):
    # Reference field
    resdim = topo.domain.dim ** 2
    ref = Field(domain=topo.domain, formula=formula,
                name='Analytical', nb_components=resdim)
    rd = ref.discretize(topo)
    ref.initialize(topo=topo)
    op, result = build_op(diffop.GradV, resdim, topo, work, reduced,
                          reduce_output)
    result = op(velo.data, result)
    ic_out = op.output_indices
    ic_in = op.in_indices
    # Numerical VS analytical
    tol = np.max(topo.mesh.space_step ** 4)
    for i in xrange(resdim):
        assert np.allclose(rd[i][ic_in], result[i][ic_out], atol=tol)


def grad_v_func_2d(res, x, y, t):
    res[0][...] = cos(x) * cos(y)
    res[1][...] = -sin(x) * sin(y)
    res[2][...] = sin(x) * sin(y)
    res[3][...] = -cos(x) * cos(y)
    return res


def grad_v_func_3d(res, x, y, z, t):
    res[0][...] = cos(x) * cos(y) * cos(z)
    res[1][...] = -sin(x) * sin(y) * cos(z)
    res[2][...] = -sin(x) * cos(y) * sin(z)
    res[3][...] = sin(x) * sin(y) * cos(z)
    res[4][...] = - cos(x) * cos(y) * cos(z) 
    res[5][...] = cos(x) * sin(y) * sin(z)
    res[6][...] = 0.
    res[7][...] = 0.
    res[8][...] = 0.
    return res


def test_grad_v():
    # 2d
    assert_grad_v(topo2, v2d, grad_v_func_2d)
    # 2d work
    assert_grad_v(topo2, v2d, grad_v_func_2d, work2)
    # 2d reduced
    assert_grad_v(topo2, v2d, grad_v_func_2d, reduced=True)
    # 2d reduced and work
    assert_grad_v(topo2, v2d, grad_v_func_2d, work2,
                  reduced=True)
    # 2d reduce output
    assert_grad_v(topo2, v2d, grad_v_func_2d,
                  reduce_output=True)
    # 2d reduce output and work
    assert_grad_v(topo2, v2d, grad_v_func_2d, work2,
                  reduce_output=True)
    # 2d reduce output, on subset, with work
    assert_grad_v(topo2, v2d, grad_v_func_2d, work2,
                  reduce_output=True, reduced=True)
    # 3d
    assert_grad_v(topo3, v3d, grad_v_func_3d)
    # 3d work
    assert_grad_v(topo3, v3d, grad_v_func_3d, work3)
    # 3d reduced
    assert_grad_v(topo3, v3d, grad_v_func_3d, reduced=True)
    # 3d reduced and work
    assert_grad_v(topo3, v3d, grad_v_func_3d, work3,
                  reduced=True)
    # 3d reduce output
    assert_grad_v(topo3, v3d, grad_v_func_3d, reduce_output=True)
    # 3d reduce output and work
    assert_grad_v(topo3, v3d, grad_v_func_3d, work3,
                  reduce_output=True)
    # 3d, work, reduce output, on subset
    assert_grad_v(topo3, v3d, grad_v_func_3d, work3,
                  reduce_output=True, reduced=True)


def assert_grad_vxw(topo, velo, vorti, work=None, reduced=False,
                    reduce_output=False):
    # Reference field
    ref = Field(domain=topo.domain, formula=analyticalGradVxW,
                name='Analytical', is_vector=True)
    rd = ref.discretize(topo)
    ref.initialize(topo=topo)
    op, result = build_op(diffop.GradVxW, velo.nb_components, topo,
                          work, reduced, reduce_output)
    diag = npw.zeros(2)
    result, diag = op(velo.data, vorti.data, result, diag)
    # Numerical VS analytical
    ic_out = op.output_indices
    ic_in = op.in_indices
    tol = np.max(topo3.mesh.space_step ** 2)
    for i in xrange(topo.domain.dim):
        assert np.allclose(rd[i][ic_in], result[i][ic_out], atol=tol)


def test_grad_vxw():
    # 3d
    assert_grad_vxw(topo3, v3d, w3d)
    # 3d work
    assert_grad_vxw(topo3, v3d, w3d, work3)
    # 3d reduced
    assert_grad_vxw(topo3, v3d, w3d, reduced=True)
    # 3d reduced and work
    assert_grad_vxw(topo3, v3d, w3d, work3,
                    reduced=True)
    # 3d reduce output
    assert_grad_vxw(topo3, v3d, w3d, reduce_output=True)
    # 3d reduce output and work
    assert_grad_vxw(topo3, v3d, w3d, work3,
                    reduce_output=True)
    # 3d, work, reduce output, on subset
    assert_grad_vxw(topo3, v3d, w3d, work3,
                    reduce_output=True, reduced=True)


# test for RHS of Pressure-Poisson equation
def assert_div_advection(topo, velo, work=None, reduced=False,
                         reduce_output=False):
    # Reference scalar field
    ref = Field(domain=topo.domain, formula=analyticalDivAdvection,
                name='Analytical')
    rd = ref.discretize(topo)
    ref.initialize(topo=topo)
    op, result = build_op(diffop.DivAdvection, 1, topo, work, reduced,
                          reduce_output)
    result = op(velo.data, result)
    ic_out = op.output_indices
    ic_in = op.in_indices

    # Numerical VS analytical
    tol = np.max(topo.mesh.space_step ** 4)
    assert np.allclose(rd[0][ic_in], result[0][ic_out], rtol=tol)


def test_div_advection():
    # 3d
    assert_div_advection(topo3, v3d)
    # 3d work
    assert_div_advection(topo3, v3d, work3)
    # 3d reduced
    assert_div_advection(topo3, v3d, reduced=True)
    # 3d reduced and work
    assert_div_advection(topo3, v3d, work3,
                         reduced=True)
    # 3d reduce output
    assert_div_advection(topo3, v3d, reduce_output=True)
    # 3d reduce output and work
    assert_div_advection(topo3, v3d, work3,
                         reduce_output=True)
    # 3d, work, reduce output, on subset
    assert_div_advection(topo3, v3d, work3,
                         reduce_output=True, reduced=True)


def assert_strain(topo, velo, formula, work=None, reduced=False,
                  reduce_output=False):
    # Reference scalar field
    resdim = topo.domain.dim * (topo.domain.dim + 1) / 2
    ref = Field(domain=topo.domain, formula=formula,
                name='Analytical', nb_components=resdim)
    rd = ref.discretize(topo)
    ref.initialize(topo=topo)
    op, result = build_op(diffop.Strain, resdim, topo, work, reduced,
                          reduce_output)
    result = op(velo.data, result)

    ic_out = op.output_indices
    ic_in = op.in_indices
    # Numerical VS analytical
    tol = np.max(topo.mesh.space_step ** 4)
    for d in xrange(resdim):
        assert np.allclose(rd[d][ic_in], result[d][ic_out], atol=tol)


def strain_f_2d(res, x, y, t):
    res[0][...] = cos(x) * cos(y)
    res[1][...] = - cos(x) * cos(y)
    res[2][...] = 0.
    return res


def strain_f_3d(res, x, y, z, t):
    res[0][...] = cos(x) * cos(y) * cos(z)
    res[1][...] = - cos(x) * cos(y) * cos(z)
    res[2][...] = 0.
    res[3][...] = 0.
    res[4][...] = -0.5 * sin(x) * cos(y) * sin(z)
    res[5][...] = 0.5 * cos(x) * sin(y) * sin(z)
    return res


def test_strain():
    # 2d
    assert_strain(topo2, v2d, strain_f_2d)
    # 2d work
    assert_strain(topo2, v2d, strain_f_2d, work2)
    # 2d reduced
    assert_strain(topo2, v2d, strain_f_2d, reduced=True)
    # 2d reduced and work
    assert_strain(topo2, v2d, strain_f_2d, work2,
                  reduced=True)
    # 2d reduce output
    assert_strain(topo2, v2d, strain_f_2d,
                  reduce_output=True)
    # 2d reduce output and work
    assert_strain(topo2, v2d, strain_f_2d, work2,
                  reduce_output=True)
    # 2d reduce output, on subset, with work
    assert_strain(topo2, v2d, strain_f_2d, work2,
                  reduce_output=True, reduced=True)
    # 3d
    assert_strain(topo3, v3d, strain_f_3d)
    # 3d work
    assert_strain(topo3, v3d, strain_f_3d, work3)
    # 3d reduced
    assert_strain(topo3, v3d, strain_f_3d, reduced=True)
    # 3d reduced and work
    assert_strain(topo3, v3d, strain_f_3d, work3,
                  reduced=True)
    # 3d reduce output
    assert_strain(topo3, v3d, strain_f_3d, reduce_output=True)
    # 3d reduce output and work
    assert_strain(topo3, v3d, strain_f_3d, work3,
                  reduce_output=True)
    # 3d, work, reduce output, on subset
    assert_strain(topo3, v3d, strain_f_3d, work3,
                  reduce_output=True, reduced=True)


def compute_max(field, ic):
    """Compute max of sum_i (abs(field_i[ic])) for i in len(field)
    """
    temp = npw.zeros_like(field[0])
    for d in xrange(len(field)):
        temp[ic] += np.abs(field[d][ic])
    return np.max(temp)


def assert_strain_criteria(topo, velo, formula, work=None, reduced=False,
                           reduce_output=False):
    # Reference scalar field
    resdim = topo.domain.dim * (topo.domain.dim + 1) / 2
    ref = Field(domain=topo.domain, formula=formula,
                name='Analytical', nb_components=resdim)
    rd = ref.discretize(topo)
    ref.initialize(topo=topo)
    op, result = build_op(diffop.StrainCriteria, 0, topo, work, reduced,
                          reduce_output)
    result = npw.zeros(topo.domain.dim)
    result = op(velo.data, result)
    ic_in = op.in_indices

    dd = topo.domain.dim
    if dd == 3:
        ll = [None, ] * dd
        ll[0] = [rd[0], rd[3], rd[4]]
        ll[1] = [rd[1], rd[3], rd[5]]
        ll[2] = [rd[2], rd[4], rd[5]]
    elif dd == 2:
        ll = [None, ] * dd
        ll[0] = [rd[0], rd[2]]
        ll[1] = [rd[1], rd[2]]

    for d in xrange(dd):
        refmax = compute_max(ll[d], ic_in)
        assert np.allclose(refmax, result[d])


def test_strain_criteria():
    # 2d
    assert_strain_criteria(topo2, v2d, strain_f_2d)
    # 2d work
    assert_strain_criteria(topo2, v2d, strain_f_2d, work2)
    # 2d reduced
    assert_strain_criteria(topo2, v2d, strain_f_2d, reduced=True)
    # 2d reduced and work
    assert_strain_criteria(topo2, v2d, strain_f_2d, work2,
                           reduced=True)
    # 2d reduce output
    assert_strain_criteria(topo2, v2d, strain_f_2d,
                           reduce_output=True)
    # 2d reduce output and work
    assert_strain_criteria(topo2, v2d, strain_f_2d, work2,
                           reduce_output=True)
    # 2d reduce output, on subset, with work
    assert_strain_criteria(topo2, v2d, strain_f_2d, work2,
                           reduce_output=True, reduced=True)
    # 3d
    assert_strain_criteria(topo3, v3d, strain_f_3d)
    # 3d work
    assert_strain_criteria(topo3, v3d, strain_f_3d, work3)
    # 3d reduced
    assert_strain_criteria(topo3, v3d, strain_f_3d, reduced=True)
    # 3d reduced and work
    assert_strain_criteria(topo3, v3d, strain_f_3d, work3,
                           reduced=True)
    # 3d reduce output
    assert_strain_criteria(topo3, v3d, strain_f_3d, reduce_output=True)
    # 3d reduce output and work
    assert_strain_criteria(topo3, v3d, strain_f_3d, work3,
                           reduce_output=True)
    # 3d, work, reduce output, on subset
    assert_strain_criteria(topo3, v3d, strain_f_3d, work3,
                           reduce_output=True, reduced=True)


def assert_max_diag_grad(topo, velo, formula, work=None, reduced=False,
                         reduce_output=False):
    # Reference field
    ref = Field(domain=topo.domain, formula=formula,
                name='Analytical', is_vector=True)
    rd = ref.discretize(topo)
    ref.initialize(topo=topo)
    op, result = build_op(diffop.MaxDiagGradV, 0, topo, work, reduced,
                          reduce_output)
    result = npw.zeros(topo.domain.dim)
    result = op(velo.data, result)
    ic_in = op.in_indices
    for d in xrange(topo.domain.dim):
        refmax = compute_max([rd[d]], ic_in)
        assert np.allclose(refmax, result[d])


def diag_grad_func_2d(res, x, y, t):
    res[0][...] = cos(x) * cos(y)
    res[1][...] = -cos(x) * cos(y)
    return res


def diag_grad_func_3d(res, x, y, z, t):
    res[0][...] = cos(x) * cos(y) * cos(z)
    res[1][...] = - cos(x) * cos(y) * cos(z)
    res[2][...] = 0.
    return res


def test_max_diag_grad():
    # 2d
    assert_max_diag_grad(topo2, v2d, diag_grad_func_2d)
    # 2d work
    assert_max_diag_grad(topo2, v2d, diag_grad_func_2d, work2)
    # 2d reduced
    assert_max_diag_grad(topo2, v2d, diag_grad_func_2d, reduced=True)
    # 2d reduced and work
    assert_max_diag_grad(topo2, v2d, diag_grad_func_2d, work2,
                         reduced=True)
    # 2d reduce output
    assert_max_diag_grad(topo2, v2d, diag_grad_func_2d,
                         reduce_output=True)
    # 2d reduce output and work
    assert_max_diag_grad(topo2, v2d, diag_grad_func_2d, work2,
                         reduce_output=True)
    # 2d reduce output, on subset, with work
    assert_max_diag_grad(topo2, v2d, diag_grad_func_2d, work2,
                         reduce_output=True, reduced=True)
    # 3d
    assert_max_diag_grad(topo3, v3d, diag_grad_func_3d)
    # 3d work
    assert_max_diag_grad(topo3, v3d, diag_grad_func_3d, work3)
    # 3d reduced
    assert_max_diag_grad(topo3, v3d, diag_grad_func_3d, reduced=True)
    # 3d reduced and work
    assert_max_diag_grad(topo3, v3d, diag_grad_func_3d, work3,
                         reduced=True)
    # 3d reduce output
    assert_max_diag_grad(topo3, v3d, diag_grad_func_3d, reduce_output=True)
    # 3d reduce output and work
    assert_max_diag_grad(topo3, v3d, diag_grad_func_3d, work3,
                         reduce_output=True)
    # 3d, work, reduce output, on subset
    assert_max_diag_grad(topo3, v3d, diag_grad_func_3d, work3,
                         reduce_output=True, reduced=True)


def assert_stretch_like(topo, velo, formula, work=None, reduced=False,
                        reduce_output=False):
    # Reference scalar field
    resdim = topo.domain.dim ** 2
    ref = Field(domain=topo.domain, formula=formula,
                name='Analytical', nb_components=resdim)
    rd = ref.discretize(topo)
    ref.initialize(topo=topo)
    op, result = build_op(diffop.StretchLike, 0, topo, work, reduced,
                          reduce_output)
    result = npw.zeros(topo.domain.dim)
    result = op(velo.data, result)
    ic_in = op.in_indices

    dd = topo.domain.dim
    for d in xrange(dd):
        pos = d * dd
        refmax = compute_max(rd[pos: pos + dd], ic_in)
        assert np.allclose(refmax, result[d])


def test_stretch_like():
    # 2d
    assert_stretch_like(topo2, v2d, grad_v_func_2d)
    # 2d work
    assert_stretch_like(topo2, v2d, grad_v_func_2d, work2)
    # 2d reduced
    assert_stretch_like(topo2, v2d, grad_v_func_2d, reduced=True)
    # 2d reduced and work
    assert_stretch_like(topo2, v2d, grad_v_func_2d, work2,
                        reduced=True)
    # 2d reduce output
    assert_stretch_like(topo2, v2d, grad_v_func_2d,
                        reduce_output=True)
    # 2d reduce output and work
    assert_stretch_like(topo2, v2d, grad_v_func_2d, work2,
                        reduce_output=True)
    # 2d reduce output, on subset, with work
    assert_stretch_like(topo2, v2d, grad_v_func_2d, work2,
                        reduce_output=True, reduced=True)
    # 3d
    assert_stretch_like(topo3, v3d, grad_v_func_3d)
    # 3d work
    assert_stretch_like(topo3, v3d, grad_v_func_3d, work3)
    # 3d reduced
    assert_stretch_like(topo3, v3d, grad_v_func_3d, reduced=True)
    # 3d reduced and work
    assert_stretch_like(topo3, v3d, grad_v_func_3d, work3,
                        reduced=True)
    # 3d reduce output
    assert_stretch_like(topo3, v3d, grad_v_func_3d, reduce_output=True)
    # 3d reduce output and work
    assert_stretch_like(topo3, v3d, grad_v_func_3d, work3,
                        reduce_output=True)
    # 3d, work, reduce output, on subset
    assert_stretch_like(topo3, v3d, grad_v_func_3d, work3,
                        reduce_output=True, reduced=True)


def assert_diag_and_stretch(topo, velo, formula, work=None, reduced=False,
                            reduce_output=False):
    # Reference scalar field
    resdim = topo.domain.dim ** 2
    ref = Field(domain=topo.domain, formula=formula,
                name='Analytical', nb_components=resdim)
    rd = ref.discretize(topo)
    ref.initialize(topo=topo)
    op, result = build_op(diffop.DiagAndStretch, 0, topo, work, reduced,
                          reduce_output)
    result = npw.zeros(2 * topo.domain.dim)
    result = op(velo.data, result)
    ic_in = op.in_indices

    dd = topo.domain.dim
    ll = [rd[i * (dd + 1)] for i in xrange(dd)]

    for d in xrange(dd):
        refmax = compute_max([ll[d]], ic_in)
        assert np.allclose(refmax, result[d])
        pos = d * dd
        refmax = compute_max(rd[pos: pos + dd], ic_in)
        assert np.allclose(refmax, result[d + dd])


def test_diag_and_stretch():
    # 2d
    assert_diag_and_stretch(topo2, v2d, grad_v_func_2d)
    # 2d work
    assert_diag_and_stretch(topo2, v2d, grad_v_func_2d, work2)
    # 2d reduced
    assert_diag_and_stretch(topo2, v2d, grad_v_func_2d, reduced=True)
    # 2d reduced and work
    assert_diag_and_stretch(topo2, v2d, grad_v_func_2d, work2,
                            reduced=True)
    # 2d reduce output
    assert_diag_and_stretch(topo2, v2d, grad_v_func_2d,
                            reduce_output=True)
    # 2d reduce output and work
    assert_diag_and_stretch(topo2, v2d, grad_v_func_2d, work2,
                            reduce_output=True)
    # 2d reduce output, on subset, with work
    assert_diag_and_stretch(topo2, v2d, grad_v_func_2d, work2,
                            reduce_output=True, reduced=True)
    # 3d
    assert_diag_and_stretch(topo3, v3d, grad_v_func_3d)
    # # 3d work
    assert_diag_and_stretch(topo3, v3d, grad_v_func_3d, work3)
    # 3d reduced
    assert_diag_and_stretch(topo3, v3d, grad_v_func_3d, reduced=True)
    # 3d reduced and work
    assert_diag_and_stretch(topo3, v3d, grad_v_func_3d, work3,
                            reduced=True)
    # 3d reduce output
    assert_diag_and_stretch(topo3, v3d, grad_v_func_3d, reduce_output=True)
    # 3d reduce output and work
    assert_diag_and_stretch(topo3, v3d, grad_v_func_3d, work3,
                            reduce_output=True)
    # 3d, work, reduce output, on subset
    assert_diag_and_stretch(topo3, v3d, grad_v_func_3d, work3,
                            reduce_output=True, reduced=True)


def assert_strain_stretch_criteria(topo, velo, formula1, formula2, work=None,
                                   reduced=False, reduce_output=False):
    # Reference scalar field
    resdim = topo.domain.dim * (topo.domain.dim + 1) / 2
    graddim = topo.domain.dim ** 2
    refstrain = Field(domain=topo.domain, formula=formula2,
                      name='Strain', nb_components=resdim)
    refgrad = Field(domain=topo.domain, formula=formula1,
                    name='Gradl', nb_components=graddim)

    rd = refstrain.discretize(topo)
    gd = refgrad.discretize(topo)
    refstrain.initialize(topo)
    refgrad.initialize(topo=topo)
    op, result = build_op(diffop.StrainAndStretch, 0, topo, work, reduced,
                          reduce_output)
    result = npw.zeros(2 * topo.domain.dim)
    result = op(velo.data, result)
    ic_in = op.in_indices

    dd = topo.domain.dim
    if dd == 3:
        ll = [None, ] * dd
        ll[0] = [rd[0], rd[3], rd[4]]
        ll[1] = [rd[1], rd[3], rd[5]]
        ll[2] = [rd[2], rd[4], rd[5]]
    elif dd == 2:
        ll = [None, ] * dd
        ll[0] = [rd[0], rd[2]]
        ll[1] = [rd[1], rd[2]]

    if dd == 3:
        for d in xrange(dd):
            refmaxstr = compute_max(ll[d], ic_in)
            assert np.allclose(refmaxstr, result[d])
            pos = d * dd
            refmaxgrad = compute_max(gd[pos: pos + dd], ic_in)
            assert np.allclose(refmaxgrad, result[d + dd])


def test_strain_stretch():
    # 2d
    assert_strain_stretch_criteria(topo2, v2d, grad_v_func_2d, strain_f_2d)
    # 2d work
    assert_strain_stretch_criteria(topo2, v2d, grad_v_func_2d, strain_f_2d,
                                   work2)
    # 2d reduced
    assert_strain_stretch_criteria(topo2, v2d, grad_v_func_2d, strain_f_2d,
                                   reduced=True)
    # 2d reduced and work
    assert_strain_stretch_criteria(topo2, v2d, grad_v_func_2d, strain_f_2d,
                                   work2, reduced=True)
    # 2d reduce output
    assert_strain_stretch_criteria(topo2, v2d, grad_v_func_2d, strain_f_2d,
                                   reduce_output=True)
    # 2d reduce output and work
    assert_strain_stretch_criteria(topo2, v2d, grad_v_func_2d, strain_f_2d,
                                   work2, reduce_output=True)
    # 2d reduce output, on subset, with work
    assert_strain_stretch_criteria(topo2, v2d, grad_v_func_2d, strain_f_2d,
                                   work2, reduce_output=True, reduced=True)
    # 3d
    assert_strain_stretch_criteria(topo3, v3d, grad_v_func_3d, strain_f_3d)
    # 3d work
    assert_strain_stretch_criteria(topo3, v3d, grad_v_func_3d, strain_f_3d,
                                   work3)
    # 3d reduced
    assert_strain_stretch_criteria(topo3, v3d, grad_v_func_3d, strain_f_3d,
                                   reduced=True)
    # 3d reduced and work
    assert_strain_stretch_criteria(topo3, v3d, grad_v_func_3d, strain_f_3d,
                                   work3, reduced=True)
    # 3d reduce output
    assert_strain_stretch_criteria(topo3, v3d, grad_v_func_3d, strain_f_3d,
                                   reduce_output=True)
    # 3d reduce output and work
    assert_strain_stretch_criteria(topo3, v3d, grad_v_func_3d, strain_f_3d,
                                   work3, reduce_output=True)
    # 3d, work, reduce output, on subset
    assert_strain_stretch_criteria(topo3, v3d, grad_v_func_3d, strain_f_3d,
                                   work3, reduce_output=True, reduced=True)

if __name__ == "__main__":
    test_curl()
    test_div_rho_v()
    test_div_w_v()
    test_laplacian()
    test_grad_s()
    test_grad_v()
    test_grad_vxw()
    test_div_advection()
    test_strain()
    test_strain_criteria()
    test_max_diag_grad()
    test_stretch_like()
    test_diag_and_stretch()
    test_strain_criteria()
    test_strain_stretch()
