"""FD schemes tests
"""
from hysop.numerics.finite_differences import FDC2, FDC4, FD2C2
from hysop.tools.numpywrappers import npw
from hysop import Box, Discretization, Field
import numpy as np
from hysop.domain.subsets import SubBox


Nx = Ny = Nz = 128
g = 2
d2d = Discretization([Nx, Ny], [g, g])
d3d = Discretization([Nx, Ny, Nz], [g, g, g])


def f2d(res, x, y, t):
    res[0][...] = np.cos(x) * y ** 2 + x * y ** 3
    return res


def ref_f2d(res, x, y, t):
    res[0][...] = -np.sin(x) * y ** 2 + y ** 3
    return res


def ref2_f2d(res, x, y, t):
    res[0][...] = 2 * np.cos(x) + 6 * x * y
    return res


def f3d(res, x, y, z, t):
    res[0][...] = np.cos(x) * y ** 2 + x * y ** 3
    return res


def ref_f3d(res, x, y, z, t):
    res[0][...] = -np.sin(x) * y ** 2 + y ** 3
    return res


def ref2_f3d(res, x, y, z, t):
    res[0][...] = 2 * np.cos(x) + 6 * x * y
    return res


lengthes = [1., 1., 1.]


def init(discr, formulas):
    dimension = len(discr.resolution)
    box = Box(length=lengthes[:dimension])
    f1 = Field(name='input', domain=box, formula=formulas[0])
    ref = Field(name='ref', domain=box, formula=formulas[1])
    topo = box.create_topology(discr)
    f1d = f1.discretize(topo).data[0]
    refd = ref.discretize(topo).data[0]
    f1.initialize(topo=topo)
    ref.initialize(topo=topo)
    ref2 = Field(name='ref', domain=box, formula=formulas[2])
    ref2d = ref2.discretize(topo).data[0]
    ref2.initialize(topo=topo)
    return topo, f1d, refd, ref2d


def run_schemes(schemes, f1d, result, ind, hh, iout=None):

    work = npw.zeros(result.size)
    if iout is None:
        iout = ind
    for sc in schemes:
        ref = schemes[sc][0]
        order = schemes[sc][2]
        direction = schemes[sc][1]
        result = sc(f1d, direction, result)
        assert np.allclose(result[iout], ref[ind], atol=hh[0] ** order)
        result[...] = 1.
        result = sc.compute_and_add(f1d, direction, result, work)
        assert np.allclose(result[iout], 1. + ref[ind], atol=hh[0] ** order)
        result[...] = 1.
        result = sc.compute_and_add_abs(f1d, direction, result, work)
        assert np.allclose(result[iout], 1. + np.abs(ref[ind]),
                           atol=hh[0] ** order)


def check_fd_schemes(discr, formulas):
    topo, f1d, refd, ref2d = init(discr, formulas)
    ind = topo.mesh.compute_index
    hh = topo.mesh.space_step
    sc = {
        FDC2(hh, ind): [refd, 0, 2],
        FDC4(hh, ind): [refd, 0, 4],
        FD2C2(hh, ind): [ref2d, 1, 2]}
    shape = topo.mesh.local_resolution
    result = npw.zeros(shape)
    run_schemes(sc, f1d, result, ind, hh)


def check_fd_schemes_reduce_input(discr, formulas):
    topo, f1d, refd, ref2d = init(discr, formulas)
    sl = topo.domain.length * 0.5
    orig = topo.domain.origin + 0.1 * topo.domain.length
    subbox = SubBox(parent=topo.domain, origin=orig, length=sl)
    ind = subbox.discretize(topo)[0]
    hh = topo.mesh.space_step
    sc = {
        FDC2(hh, ind): [refd, 0, 2],
        FDC4(hh, ind): [refd, 0, 4],
        FD2C2(hh, ind): [ref2d, 1, 2]}
    shape = topo.mesh.local_resolution
    result = npw.zeros(shape)
    run_schemes(sc, f1d, result, ind, hh)


def check_fd_schemes_reduce_output(discr, formulas):
    topo, f1d, refd, ref2d = init(discr, formulas)
    ind = topo.mesh.compute_index
    hh = topo.mesh.space_step
    sc = {
        FDC2(hh, ind, output_indices=True): [refd, 0, 2],
        FDC4(hh, ind, output_indices=True): [refd, 0, 4],
        FD2C2(hh, ind, output_indices=True): [ref2d, 1, 2]
        }
    shape = np.asarray(topo.mesh.local_resolution).copy()
    shape -= 2 * g
    shape = tuple(shape)
    result = npw.zeros(shape)
    iout = [slice(0, shape[i]) for i in xrange(len(shape))]
    run_schemes(sc, f1d, result, ind, hh, iout)


def check_fd_schemes_reduce_all(discr, formulas):
    topo, f1d, refd, ref2d = init(discr, formulas)
    hh = topo.mesh.space_step
    sl = topo.domain.length * 0.5
    orig = topo.domain.origin + 0.1 * topo.domain.length
    subbox = SubBox(parent=topo.domain, origin=orig, length=sl)
    ind = subbox.discretize(topo)[0]
    sc = {
        FDC2(hh, ind, output_indices=True): [refd, 0, 2],
        FDC4(hh, ind, output_indices=True): [refd, 0, 4],
        FD2C2(hh, ind, output_indices=True): [ref2d, 1, 2]
        }
    shape = subbox.mesh[topo].resolution
    result = npw.zeros(shape)
    iout = [slice(0, shape[i]) for i in xrange(len(shape))]
    run_schemes(sc, f1d, result, ind, hh, iout)


def check_fd_schemes_setiout(discr, formulas):
    topo, f1d, refd, ref2d = init(discr, formulas)
    hh = topo.mesh.space_step
    sl = topo.domain.length * 0.5
    orig = topo.domain.origin + 0.1 * topo.domain.length
    subbox = SubBox(parent=topo.domain, origin=orig, length=sl)
    ind = subbox.discretize(topo)[0]
    shape = np.asarray(subbox.mesh[topo].resolution).copy()
    shape += 6
    shape = tuple(shape)
    result = npw.zeros(shape)
    iout = [slice(3, shape[i] - 3) for i in xrange(len(shape))]
    sc = {
        FDC2(hh, ind, output_indices=iout): [refd, 0, 2],
        FDC4(hh, ind, output_indices=iout): [refd, 0, 4],
        FD2C2(hh, ind, output_indices=iout): [ref2d, 1, 2]
        }
    run_schemes(sc, f1d, result, ind, hh, iout)


def test_fd_2d():
    check_fd_schemes(d2d, [f2d, ref_f2d, ref2_f2d])


def test_fd_3d():
    check_fd_schemes(d3d, [f3d, ref_f3d, ref2_f3d])


def test_fd_3d_reduce_input():
    check_fd_schemes_reduce_input(d3d, [f3d, ref_f3d, ref2_f3d])


def test_fd_3d_reduce_output():
    check_fd_schemes_reduce_output(d3d, [f3d, ref_f3d, ref2_f3d])


def test_fd_3d_reduce_all():
    check_fd_schemes_reduce_all(d3d, [f3d, ref_f3d, ref2_f3d])


def test_fd_3d_setiout():
    check_fd_schemes_setiout(d3d, [f3d, ref_f3d, ref2_f3d])
