"""Test ghost points synchronization
"""
from hysop.numerics.update_ghosts import UpdateGhosts, UpdateGhostsFull
from hysop.domain.box import Box
from hysop.fields.continuous_field import Field
from hysop.tools.parameters import Discretization
from hysop.core.mpi import main_rank
from hysop.tools.numpywrappers import npw
import numpy as np


def get_gh_slices(topo):
    """Return indices of ghost points"""
    tab = Field(topo.domain, name='temp')
    tabd = tab.discretize(topo)[0]
    tabd[...] = 0.
    tabd[topo.mesh.compute_index] = 1.
    return np.where(tabd == 0.)


def get_ghosts_indices(topo):
    """Return indices of points inside the ghost layers
    for a given topology

    gh_s[d] : for points at the beginning of the domain
    in direction d.
    gh_e[d] : for points at the end of the domain
    in direction d.

    Corner points are not included.
    """
    gh_s = []
    gh_e = []
    ghosts = topo.ghosts()
    sl_start = [slice(0, ghosts[d]) for d in xrange(topo.domain.dim)]
    sl_end = [slice(-ghosts[d], None, None)
              for d in xrange(topo.domain.dim)]
    for j in xrange(topo.domain.dim):
        gh_s.append(list(topo.mesh.compute_index))
        gh_e.append(list(topo.mesh.compute_index))
        gh_s[j][j] = sl_start[j]
        gh_e[j][j] = sl_end[j]
    return gh_s, gh_e


def get_corners_indices(topo):
    """Return indices of points inside the ghost layers
    for a given topology

    gh_s[d] : for points at the beginning of the domain
    in direction d.
    gh_e[d] : for points at the end of the domain
    in direction d.

    Corner points are included.
    """
    gh_s = []
    gh_e = []
    ghosts = topo.ghosts()
    sl_start = [slice(0, ghosts[d]) for d in xrange(topo.domain.dim)]
    sl_end = [slice(-ghosts[d], None, None)
              for d in xrange(topo.domain.dim)]
    for j in xrange(topo.domain.dim):
        gh_s.append(list(topo.mesh.local_index))
        gh_e.append(list(topo.mesh.local_index))
        gh_s[j][j] = sl_start[j]
        gh_e[j][j] = sl_end[j]

    start, end = get_ghosts_indices(topo)
    tab = np.zeros((topo.mesh.local_resolution), dtype=np.int16)
    tab[topo.mesh.compute_index] = 1
    for d in xrange(topo.domain.dim):
        tab[start[d]] = 1
        tab[end[d]] = 1
    corners = np.where(tab == 0)

    corners = [list(topo.mesh.local_index), ] * (2 ** (topo.domain.dim))

    dimension = topo.domain.dim
    corners_size = 2 ** dimension
    corners = [None, ] * corners_size
    val = npw.int_zeros(corners_size)
    if dimension == 1:
        return [], 0.
    elif dimension == 2:
        corners[0] = [sl_start[0], sl_start[1]]
        corners[1] = [sl_start[0], sl_end[1]]
        corners[2] = [sl_end[0], sl_start[1]]
        corners[3] = [sl_end[0], sl_end[1]]
        val[0] = topo.comm.Get_cart_rank(
            [topo.proc_coords[0] - 1, topo.proc_coords[1] - 1])
        val[1] = topo.comm.Get_cart_rank(
            [topo.proc_coords[0] - 1, topo.proc_coords[1] + 1])
        val[2] = topo.comm.Get_cart_rank(
            [topo.proc_coords[0] + 1, topo.proc_coords[1] - 1])
        val[3] = topo.comm.Get_cart_rank(
            [topo.proc_coords[0] + 1, topo.proc_coords[1] + 1])
    elif dimension == 3:
        corners[0] = [sl_start[0], sl_start[1], sl_start[2]]
        corners[1] = [sl_start[0], sl_start[1], sl_end[2]]
        corners[2] = [sl_start[0], sl_end[1], sl_start[2]]
        corners[3] = [sl_start[0], sl_end[1], sl_end[2]]
        corners[4] = [sl_end[0], sl_start[1], sl_start[2]]
        corners[5] = [sl_end[0], sl_start[1], sl_end[2]]
        corners[6] = [sl_end[0], sl_end[1], sl_start[2]]
        corners[7] = [sl_end[0], sl_end[1], sl_end[2]]
        val[0] = topo.comm.Get_cart_rank(
            [topo.proc_coords[0] - 1,
             topo.proc_coords[1] - 1,
             topo.proc_coords[2] - 1])
        val[1] = topo.comm.Get_cart_rank(
            [topo.proc_coords[0] - 1,
             topo.proc_coords[1] - 1,
             topo.proc_coords[2] + 1])
        val[2] = topo.comm.Get_cart_rank(
            [topo.proc_coords[0] - 1,
             topo.proc_coords[1] + 1,
             topo.proc_coords[2] - 1])
        val[3] = topo.comm.Get_cart_rank(
            [topo.proc_coords[0] - 1,
             topo.proc_coords[1] + 1,
             topo.proc_coords[2] + 1])
        val[4] = topo.comm.Get_cart_rank(
            [topo.proc_coords[0] + 1,
             topo.proc_coords[1] - 1,
             topo.proc_coords[2] - 1])
        val[5] = topo.comm.Get_cart_rank(
            [topo.proc_coords[0] + 1,
             topo.proc_coords[1] - 1,
             topo.proc_coords[2] + 1])
        val[6] = topo.comm.Get_cart_rank(
            [topo.proc_coords[0] + 1,
             topo.proc_coords[1] + 1,
             topo.proc_coords[2] - 1])
        val[7] = topo.comm.Get_cart_rank(
            [topo.proc_coords[0] + 1,
             topo.proc_coords[1] + 1,
             topo.proc_coords[2] + 1])

#    print "youou ", corners
    return corners, val


def run_synchro(ghosts, cutdir=None, full=False):
    """Build domain, fields and topology
    for a given ghost layer size and
    distributed directions.
    Then call synchronization and check.
    """
    dimension = len(ghosts)
    dom = Box(length=[1., ] * dimension)
    f = Field(dom, is_vector=True, name='f')
    discr = Discretization([33, ] * dimension, ghosts=ghosts)
    topo = dom.create_topology(discr, cutdir=cutdir)
    df = f.discretize(topo)
    # initialize + 0 in ghost layer
    nbc = f.nb_components
    for i in xrange(nbc):
        df[i][...] = main_rank
        #df[i][topo.mesh.compute_index] = topo.rank

    for i in xrange(df.nb_components):
        assert np.allclose(df[i], main_rank)

    if full:
        sync = UpdateGhostsFull(topo, nbc)

    else:
        sync = UpdateGhosts(topo, nbc)

    start, end = get_ghosts_indices(topo)
    corners, val = get_corners_indices(topo)

    sync.apply(df.data)
    nghb = topo.neighbours_tmp
    for i in xrange(df.nb_components):
        assert np.allclose(df[i][topo.mesh.compute_index], main_rank)
        for d in xrange(topo.domain.dim):
            assert np.allclose(df[i][start[d]], nghb[d][0])
            assert np.allclose(df[i][end[d]], nghb[d][1])
        if full:
            for d in xrange(len(corners)):
                assert np.allclose(df[i][corners[d]], val[d])


def test_1_1_simple():
    """ 1D domain, 1D topo, simple update
    """
    run_synchro([1, ])


def test_2_1_simple():
    """ 2D domain, 1D topo, simple update
    """
    run_synchro([1, 1], [True, False])


def test_2_2_simple():
    """ 2D domain, 2D topo, simple update
    """
    run_synchro([1, 1])


def test_3_1_simple():
    """ 3D domain, 1D topo, simple update
    """
    run_synchro([1, 1, 1], [True, False, False])


def test_3_2_simple():
    """ 3D domain, 1D topo, simple update
    """
    run_synchro([1, 1, 1], [True, False, True])


def test_3_3_simple():
    """ 3D domain, 1D topo, simple update
    """
    run_synchro([1, 1, 1])


def test_3_3_bis():
    """ 3D domain, 3D topo, different ghost size in each dir
    """
    run_synchro([1, 3, 2])


def test_1_1_full():
    """ 1D domain, 1D topo, corners update
    """
    run_synchro([1, ], full=True)


def test_2_1_full():
    """ 2D domain, 1D topo, corners update
    """
    run_synchro([1, 1], [True, False], full=True)


def test_2_2_full():
    """ 2D domain, 2D topo, corners update
    """
    run_synchro([1, 1], full=True)


def test_3_1_full():
    """ 3D domain, 1D topo, corners update
    """
    run_synchro([1, 1, 1], [True, False, False], full=True)


def test_3_2_full():
    """ 3D domain, 2D topo, corners update
    """
    run_synchro([1, 1, 1], [True, False, True], full=True)


def test_3_3_full():
    """ 3D domain, 3D topo, corners update
    """
    run_synchro([1, 1, 1], full=True)

def test_3_2_bis_full():
    """ 3D domain, 2D topo, corners update, different ghost size in each dir
    """
    run_synchro([1, 3, 2], [True, False, True], full=True)
