"""Advection of a field in a given direction

See :ref:`advection` in user guide.

"""
from hysop.constants import debug, DirectionLabels
from hysop.methods import Support, TimeIntegrator, Interpolation
from hysop.numerics.remeshing import Remeshing
from hysop.operator.computational import Computational
import numpy as np
from hysop.operator.continuous import opsetup, opapply
from hysop.operator.advection import Advection


class AdvectionDir(Computational):
    """Advection of a scalar or vector field in a given direction,
    assuming incompressible flow.
    """

    @debug
    def __init__(self, parent, direction, **kwds):
        """"
        Parameters
        ----------
        parent : :class:`hysop.operator.advection.Advection`
            main advection operator
        direction: int
            direction of advection
        kwds : base class arguments.
        """
        assert isinstance(parent, Advection)
        self.parent = parent
        # advected_fields :
        self.advected_fields = parent.advected_fields
        # velocity field
        self.velocity = parent.velocity
        self._kwds = kwds
        super(AdvectionDir, self).__init__(variables=parent.variables, **kwds)

        self.method = parent.method
        self.input = parent.input
        self.output = parent.output
        self.name = parent.name + DirectionLabels[direction]

        # direction to advect
        self.direction = direction

        # Fields on particles
        self.particle_fields = None

        # Positions of the particles
        self.particle_positions = None

        # number of components of the rhs (time integration)
        self._rhs_size = 1

        # check if advection is done on gpu
        self._gpu_advection = self.method[Support].find('gpu') >= 0

    @debug
    def discretize(self):
        if self._is_discretized:
            return

        # everything is done in parent ...
        self.variables = self.parent.variables
        self.discrete_fields = self.parent.discrete_fields
        self._is_discretized = True

    @opsetup
    def setup(self, rwork=None, iwork=None):
        # select discretization of the advected fields
        advected_discrete_fields = [self.discrete_fields[v]
                                    for v in self.variables
                                    if v is not self.velocity]
        # GPU advection ...
        if self._gpu_advection:
            topo_shape = advected_discrete_fields[0].topology.shape
            if topo_shape[self.direction] == 1:
                from hysop.backend.device.opencl.gpu_particle_advection \
                    import GPUParticleAdvection as advec
            else:
                from hysop.backend.device.opencl.multi_gpu_particle_advection \
                    import MultiGPUParticleAdvection as advec
        else:
            # pure-python advection
            from hysop.operator.discrete.particle_advection \
                import ParticleAdvection as advec

        self.discrete_op = advec(
            velocity=self.discrete_fields[self.velocity],
            fields_on_grid=advected_discrete_fields,
            direction=self.direction,
            rwork=rwork, iwork=iwork,
            **self._kwds)

        self._is_uptodate = True

    def get_work_properties(self):
        super(AdvectionDir, self).get_work_properties()
        # Shape of reference comes from fields, not from velocity
        advected_discrete_fields = [self.discrete_fields[v]
                                    for v in self.variables
                                    if v is not self.velocity]
        topo = advected_discrete_fields[0].topology
        # Find number and shape of required work arrays
        if not self._gpu_advection:
            # -- pure python advection --
            #  work array shape depends on the time integrator
            #  interpolation scheme and remeshing scheme
            ti_work = self.method[TimeIntegrator].get_work_properties(
                self._rhs_size, topo)
            ti_rwork_length = len(ti_work['rwork'])
            iw_prop = self.method[Interpolation].get_work_properties(topo)
            rw_prop = Remeshing.get_work_properties(topo)
            interp_iwork_length = len(iw_prop['iwork'])
            interp_rwork_length = len(iw_prop['rwork'])
            remesh_iwork_length = len(rw_prop['iwork'])
            remesh_rwork_length = len(rw_prop['rwork'])
            iwork_length = max(interp_iwork_length, remesh_iwork_length)
            rwork_length = max(ti_rwork_length + interp_rwork_length,
                               remesh_rwork_length)
        else:
            # -- GPU advection --
            # no work array
            iwork_length, rwork_length = 0, 0

        # buffers for fields on particles
        rwork_length += np.sum([f.nb_components for f in self.advected_fields])
        if not self._gpu_advection or \
           self.method[Support].find('gpu_2k') >= 0:
            rwork_length += 1  # work array for positions
        memsize = np.prod(topo.mesh.local_resolution)
        return {'rwork': [memshape] * rwork_length,
                'iwork': [memshape] * iwork_length}

    @debug
    @opapply
    def apply(self, simulation=None, dt_coeff=1.0, split_id=0, old_dir=0):
        """

        Parameters
        ----------

        simulation : `:class::~hysop.problem.simulation.Simulation`
        dt_coeff : double
        split_id : int, optional
        old_dir : int, optional
        """
        if not self._single_topo and not self._gpu_advection:
            raise ValueError("Multiscale advection is not yet supported "
                             "in pure Python, use Scales or GPU.")

        self.discrete_op.apply(simulation,
                               dt_coeff, split_id, old_dir)
