# -*- coding: utf-8 -*-
"""Interface to scales solver (fortran)

"""
try:
    from hysop.f2hysop import scales2py
except ImportError:
    msgE = 'scales package not available for your hysop install.'
    msgE += 'Try to recompile with WITH_SCALES=ON'
    raise ImportError(msgE)
from hysop.operator.discrete.discrete import DiscreteOperator
from hysop.methods import MultiScale
from hysop.constants import debug
import math
ceil = math.ceil


class ScalesAdvection(DiscreteOperator):
    """Advection process, based on scales library.
    """

    _scales_remesh_kernels = ['p_M4', 'p_M8', 'p_44', 'p_64',
                              'p_66', 'p_84', 'p_M6']
    """Kernels available in scales"""
    _scales_remesh_corrected_kernels = ['p_O2', 'p_O4', 'p_L2']
    """Corrected kernels available in scales"""

    @debug
    def __init__(self, velocity, advected_fields, **kwds):
        """
        Parameters
        ----------
        velocity : :class:`~hysop.fields.discrete_field.DiscreteField`
            advection velocity (discretized)
        advected_fields : list of :class:`~hysop.fields.discrete_field.DiscreteField`
            discrete field(s) to be advected
        direction : int
            advection direction
        kwds : extra parameters for base class
        """
        msg = 'Scales advection init : variables parameter is useless.'
        msg += 'See user guide for details on the proper'
        msg += ' way to build the operator.'
        assert 'variables' not in kwds, msg
        # Advection velocity
        self.velocity = velocity
        variables = [self.velocity]
        if isinstance(advected_fields, list):
            self.advected_fields = advected_fields
        else:
            self.advected_fields = [advected_fields]
        variables += self.advected_fields

        msg = 'Scales advection init : method parameter is compulsory.'
        assert 'method' in kwds, msg

        super(ScalesAdvection, self).__init__(variables=variables, **kwds)

        self.input = self.variables
        self.output = self.advected_fields

        # Connection to the proper scales functions.
        self._scales_func = []
        is_multi_scale = self.method[MultiScale] is not None
        for adF in self.advected_fields:
            if adF.nb_components == 3:
                if is_multi_scale:
                    # 3D interpolation of the velocity before advection
                    self._scales_func.append(
                        scales2py.solve_advection_inter_basic_vect)
                    # Other interpolation only 2D interpolation first and
                    # 1D interpolations before advections in each direction
                    # (slower than basic): solve_advection_inter
                else:
                    self._scales_func.append(scales2py.solve_advection_vect)
            else:
                if is_multi_scale:
                    self._scales_func.append(
                        scales2py.solve_advection_inter_basic)
                else:
                    self._scales_func.append(scales2py.solve_advection)

    @debug
    def apply(self, simulation=None):
        assert simulation is not None, \
            "Missing simulation value for computation."

        dt = simulation.time_step
        # Call scales advection
        for adF, fun in zip(self.advected_fields, self._scales_func):
            adF = fun(dt, self.velocity.data[0],
                      self.velocity.data[1],
                      self.velocity.data[2],
                      *adF)

    def finalize(self):
        """
        \todo check memory deallocation in scales???
        """
        DiscreteOperator.finalize(self)
