# -*- coding: utf-8 -*-
"""Compute Energy and Enstrophy from velocity and vorticity


See :ref:`energy_enstrophy`.
"""
from hysop.operator.discrete.energy_enstrophy import EnergyEnstrophy as DEE
from hysop.operator.computational import Computational
from hysop.operator.continuous import opsetup


class EnergyEnstrophy(Computational):
    """
    Computes enstrophy and kinetic energy
    """

    def __init__(self, velocity, vorticity, is_normalized=True, **kwds):
        """

        Parameters
        ----------
        velocity, vorticity : :class:`~hysop.operator.field.continuous.Field`
            velocity and vorticity fields
        isNormalized : boolean
            true if enstrophy and energy values have to be normalized
            by the domain lengths.

        Notes
        -----
        By default, values at each time step are appended
        to energy_enstrophy file. See :ref:`operators_io`
        for more details.
        """
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(EnergyEnstrophy, self).__init__(variables=[velocity, vorticity],
                                              **kwds)
        # velocity field
        self.velocity = velocity
        # vorticity field
        self.vorticity = vorticity
        # are the energy end enstrophy values normalized by domain lengths ?
        self.is_normalized = is_normalized
        # self._buffer_1 = 0.
        # self._buffer_2 = 0.
        self.input = [velocity, vorticity]
        self.output = []

    def get_work_properties(self):
        super(EnergyEnstrophy, self).get_work_properties()
        vd = self.discrete_fields[self.velocity]
        wd = self.discrete_fields[self.vorticity]
        v_ind = vd.topology.mesh.compute_index
        w_ind = wd.topology.mesh.compute_index
        size_v = vd[0][v_ind].size
        size_w = wd[0][w_ind].size
        size_work = max(size_v, size_w)
        # work array is a flat array, shared between energy and enstrophy
        return {'rwork': [size_work], 'iwork': None}

    @opsetup
    def setup(self, rwork=None, iwork=None):
        if not self._is_uptodate:

            self.discrete_op = DEE(self.discrete_fields[self.velocity],
                                   self.discrete_fields[self.vorticity],
                                   self.is_normalized,
                                   rwork=rwork)
            # Output setup
            self._set_io('energy_enstrophy', (1, 3))
            self.discrete_op.set_writer(self._writer)
            self._is_uptodate = True

    def energy(self):
        """Return last computed value of the energy
        """
        return self.discrete_op.energy

    def enstrophy(self):
        """Return last computed value of the enstrophy
        """
        return self.discrete_op.enstrophy
