"""Test 'boundaries absorption' operator
"""
from hysop.operator.absorption_bc import AbsorptionBC
from hysop import Box, Discretization, Field, VariableParameter,\
    Simulation
import numpy as np
from hysop.domain.subsets import SubBox
from hysop.operator.analytic import Analytic


d2d = Discretization([33, ] * 2)
d3d = Discretization([33, ] * 3)

uinf = 10.


def init_test(dim, discr, ext_filter, filter_func=None):
    """Create fields and operator
    """
    dom = Box([1., ] * dim)
    #topo = dom.create_topology(discr)
    v = Field(dom, name='velo', is_vector=True)
    w = Field(dom, name='vorti', is_vector=True)
    wref = Field(v.domain, name='wref', is_vector=True)

    topo = dom.create_topology(discr)
    # Build and apply absorption operator on w
    flowrate = VariableParameter(data=uinf)
    x_range = [0.9, 1.]
    op = AbsorptionBC(v, w, flowrate, x_range,
                      discretization=discr, filter_func=filter_func)
    op.discretize()
    op.setup()
    wk_prop = op.get_work_properties()['rwork']
    dop = op.discrete_op
    assert np.prod(dop.absorption_box.mesh[topo].resolution) == wk_prop[0]

    orig = [0.9, 0., 0.]
    ll = [0.1, 1., 1.]
    boxend = SubBox(origin=orig, length=ll, parent=dom)
    ll = [0.9, 1., 1.]
    boxend.discretize(topo)
    v.randomize(topo)
    vd = v.discretize(topo)
    w.randomize(topo)
    wref.copy(w, topo)
    wdref = wref.discretize(topo)
    wd = w.discretize(topo)

    for d in xrange(dom.dimension):
        assert np.allclose(wdref[d], wd[d])
    op.apply()
    # Apply filter on wref
    if dop.absorption_box.on_proc[topo]:
        ext_filter(wdref, vd, boxend, topo)
    for d in xrange(dom.dimension):
        assert np.allclose(wdref[d], wd[d])


def apply_filter_default(res, velo, subbox, topo):
    """Default values for the filter in the absorption box
    """
    xcend = subbox.mesh[topo].coords[0]
    iend = subbox.ind[topo][0]
    xb = xcend[0]
    xe = xcend[-1]
    xc = xb + (xe - xb) / 2.0
    eps = 10.
    form = np.tanh(eps * (xcend - xc))
    absorption_filter = form - np.tanh(eps * (xe - xc))
    coeff = 1.0 / (np.tanh(eps * (xb - xc)) - np.tanh(eps * (xe - xc)))
    absorption_filter *= coeff
    res[0][iend] *= absorption_filter
    res[1][iend] *= absorption_filter
    res[2][iend] *= absorption_filter
    dfilter = eps * (1.0 - form ** 2)
    dfilter *= coeff
    res[1][iend] = res[1][iend] + (uinf - velo[2][iend]) * dfilter
    res[2][iend] = res[2][iend] - (uinf - velo[1][iend]) * dfilter
    return res


def apply_user_filter(res, velo, subbox, topo):
    """Linear filter
    """
    xcend = subbox.mesh[topo].coords[0]
    iend = subbox.ind[topo][0]
    xb = xcend[0]
    xe = xcend[-1]
    absorption_filter = xcend / (xb - xe) - 1. / (xb - xe)
    res[0][iend] *= absorption_filter
    res[1][iend] *= absorption_filter
    res[2][iend] *= absorption_filter
    dfilter = 1. / (xb - xe)
    res[1][iend] = res[1][iend] + (uinf - velo[2][iend]) * dfilter
    res[2][iend] = res[2][iend] - (uinf - velo[1][iend]) * dfilter
    return res


def test_default_filter_3d():
    """Filter using default internal
    filter
    """
    init_test(3, d3d, apply_filter_default)


def user_filter(x):
    """Linear filter, just for test"""
    xb = x[0]
    xe = x[-1]
    return x / (xb - xe) - 1. / (xb - xe)


def user_diff_filter(x):
    """Linear filter, just for test"""
    return 1. / (x[0] - x[-1])


def test_user_filter_3d():
    """Filter using external filter
    """
    filter_f = [user_filter, user_diff_filter]
    init_test(3, d3d, apply_user_filter, filter_f)
