# -*- coding: utf-8 -*-
"""Tests for stretching component resolution
"""
from hysop import Field, Box
import numpy as np
from hysop.operator.stretching import Stretching, StretchingLinearized, GradUW
from hysop.problem.simulation import Simulation
from hysop.methods import TimeIntegrator, Formulation,\
    SpaceDiscretization
from hysop.methods import RK3, FDC4
from hysop.tools.parameters import Discretization
from hysop.tools.misc import WorkSpaceTools
pi = np.pi
cos = np.cos
sin = np.sin


# 3d discretisation with size 2 ghosts layer
d3d = Discretization([65, ] * 3, [2, ] * 3)


def compute_vel(res, x, y, z, t):
    """3d vector field
    """
    amodul = cos(pi * 1. / 3.)
    pix = pi * x
    piy = pi * y
    piz = pi * z
    pi2x = 2. * pix
    pi2y = 2. * piy
    pi2z = 2. * piz
    res[0][...] = 2. * sin(pix) * sin(pix) \
        * sin(pi2y) * sin(pi2z) * amodul
    res[1][...] = - sin(pi2x) * sin(piy) \
        * sin(piy) * sin(pi2z) * amodul
    res[2][...] = - sin(pi2x) * sin(piz) \
        * sin(piz) * sin(pi2y) * amodul
    return res


def compute_vort(res, x, y, z, t):
    """3d vector field
    """
    amodul = cos(pi * 1. / 3.)
    pix = pi * x
    piy = pi * y
    piz = pi * z
    pi2x = 2. * pix
    pi2y = 2. * piy
    pi2z = 2. * piz
    res[0][...] = 2. * pi * sin(pi2x) * amodul *\
        (- cos(pi2y) * sin(piz) * sin(piz) +
         sin(piy) * sin(piy) * cos(pi2z))

    res[1][...] = 2. * pi * sin(pi2y) * amodul *\
        (2. * cos(pi2z) * sin(pix) * sin(pix) +
         sin(piz) * sin(piz) * cos(pi2x))

    res[2][...] = -2. * pi * sin(pi2z) * amodul *\
        (cos(pi2x) * sin(piy) * sin(piy) +
         sin(pix) * sin(pix) * cos(pi2y))

    return res


def init(method=None, work=False):
    """Build, init, setup for operator
    """
    # Domain
    box = Box()

    # Fields
    velo = Field(
        domain=box, formula=compute_vel,
        name='Velocity', is_vector=True)
    vorti = Field(
        domain=box, formula=compute_vort,
        name='Vorticity', is_vector=True)
    # Stretching operator
    op = Stretching(velo, vorti, discretization=d3d, method=method)
    op.discretize()
    rwork = None
    if work:
        # Find required size for internal work vector, if needed
        wkp = op.get_work_properties()['rwork']
        # Allocate work space
        rwork = WorkSpaceTools.check_work_array(len(wkp), wkp[0])
    topo = op.discrete_fields[velo].topology
    op.setup(rwork=rwork)
    simulation = Simulation(start=0, end=1., time_step=0.05)
    # initialize fields and simu
    velo.initialize(topo=topo)
    vorti.initialize(topo=topo)
    simulation.initialize()
    return op, simulation


def test_stretching():
    """Default case
    """
    #method = {TimeIntegrator: RK3, Formulation: Conservative,
    #          SpaceDiscretization: FDC4}
    op, simu = init()
    op.apply(simu)


def test_stretching_external_work():
    """Default setup but with user-defined work arrays
    """
    op, simu = init(work=True)
    op.apply(simu)


def test_stretching_graduw():
    """GradUW formulation
    """
    method = {TimeIntegrator: RK3, Formulation: GradUW,
              SpaceDiscretization: FDC4}
    op, simu = init(method=method)
    op.apply(simu)


# def test_compare_stretching():
#     """Run conservative and graduv form,
#     check if results are close enough
#     """
#     op1, simu = init()
#     op1.apply(simu)
#     method = {TimeIntegrator: RK3, Formulation: GradUW,
#               SpaceDiscretization: FDC4}
#     op2, simu = init(method=method)
#     op2.apply(simu)
#     w1 = op1.discrete_op.vorticity
#     w2 = op2.discrete_op.vorticity
#     h = w1.topology.mesh.space_step.max()
#     for d in xrange(op1.domain.dim):
#         print np.abs((w1[d] - w2[d])).max()
#         assert np.allclose(w1[d], w2[d], atol=h ** 2)


def init_linearized(method=None, work=False):
    """Build, init, setup for linearized stretching
    """
    # Domain
    box = Box()

    # Base fields:
    velo_bf = Field(
        domain=box, formula=compute_vel,
        name='VelocityBF', is_vector=True)
    vorti_bf = Field(
        domain=box, formula=compute_vort,
        name='VorticityBF', is_vector=True)
    # Perturbations
    velo = Field(
        domain=box, formula=compute_vel,
        name='Velocity', is_vector=True)
    vorti = Field(
        domain=box, formula=compute_vort,
        name='Vorticity', is_vector=True)

    # Usual stretching operator
    stretch1 = Stretching(velo, vorti, discretization=d3d)
    # Linearized stretching
    stretch2 = StretchingLinearized(velocity=velo, vorticity=vorti,
                                    velocity_BF=velo_bf,
                                    vorticity_BF=vorti_bf,
                                    discretization=d3d, method=method)
    stretch1.discretize()
    stretch2.discretize()
    rwork = None
    if work:
        wkp = stretch2.get_work_properties()['rwork']
        rwork = WorkSpaceTools.check_work_array(len(wkp), wkp[0])
    topo = stretch1.discrete_fields[velo].topology
    # initialize all fields
    velo.initialize(topo=topo)
    vorti.initialize(topo=topo)
    velo_bf.initialize(topo=topo)
    vorti_bf.initialize(topo=topo)
    stretch1.setup()
    stretch2.setup(rwork=rwork)
    simulation = Simulation(start=0, end=1., time_step=0.05)
    simulation.initialize()
    return stretch1, stretch2, simulation


def test_stretching_linearized():
    """Linearized formulation for stretching
    """
    str1, str2, simu = init_linearized()
    str1.apply(simu)
    str2.apply(simu)


def test_stretching_external_work_graduv():
    """User-defined work arrays for GradUW formulation
    """
    method = {TimeIntegrator: RK3, Formulation: GradUW,
              SpaceDiscretization: FDC4}
    op, simu = init(work=True, method=method)
    op.apply(simu)
