"""Initialize fields on a grid, with a user-defined function
"""
from hysop.constants import Backend, Implementation
from hysop.fields.continuous_field import Field
from hysop.tools.types import check_instance
from hysop.tools.decorators import debug
from hysop.topology.cartesian_descriptor import CartesianTopologyDescriptors
from hysop.core.graph.computational_node_frontend import ComputationalGraphNodeFrontend

class AnalyticField(ComputationalGraphNodeFrontend):
    """
    Applies an analytic formula, given by user, on its field.
    Formula may be given in different formats depending on the
    chosen implementation backend.
    """

    @classmethod
    def implementations(cls):
        from hysop.backend.host.python.operator.analytic   import PythonAnalyticField
        from hysop.backend.device.opencl.operator.analytic import OpenClAnalyticField
        __implementations = {
                Implementation.PYTHON: PythonAnalyticField,
                Implementation.OPENCL: OpenClAnalyticField
        }
        return __implementations
    
    @classmethod
    def default_implementation(cls):
        return Implementation.PYTHON

    @debug
    def __init__(self, field, formula, variables, extra_input_kwds=None,
            implementation=None, base_kwds=None, **kwds):
        """
        AnalyticField operator frontend.

        Apply a user-defined formula onto a field, possibly 
        dependent on space variables and external fields/parameters.

        Parameters
        ----------
        field: hysop.field.continuous_field.Field
            Continuous field to be modified.
        formula : python function or sympy expression
            The formula to be applied onto the field.
        variables: dict
            Dictionary of fields as keys and topology descriptors as values.
        implementation: Implementation, optional, defaults to None
            target implementation, should be contained in available_implementations().
            If None, implementation will be set to default_implementation().
        extra_input_kwds: dict, optional
            Extra inputs that will be forwarded to the formula.
            Fields and Parameters are handled correctly as input requirements.
            Only used for Implementation.PYTHON, discarded for other implementations.
        base_kwds: dict, optional
            Base class keyword arguments.
        kwds: dict, optional
            Extra parameters passed towards operator implementation.
        """
        check_instance(field, Field)
        check_instance(variables, dict, keys=Field, values=CartesianTopologyDescriptors)

        if (implementation is Implementation.PYTHON):
            kwds['extra_input_kwds'] = extra_input_kwds

        super(AnalyticField, self).__init__(field=field, formula=formula, 
                variables=variables, implementation=implementation, 
                base_kwds=base_kwds, **kwds)
