
from hysop.tools.types import check_instance, first_not_None
from hysop.tools.decorators import debug
from hysop.fields.continuous_field import Field
from hysop.topology.cartesian_descriptor import CartesianTopologyDescriptors
from hysop.operator.base.spectral_operator import SpectralOperatorBase
from hysop.symbolic.relational import Assignment
from hysop.symbolic.field import laplacian

class PoissonOperatorBase(SpectralOperatorBase):
    """
    Solves the poisson equation using a specific implementation.
    """
    
    @debug
    def __init__(self, Fin, Fout, variables, 
            name=None, pretty_name=None,
            **kwds): 
        """
        Initialize a n-dimensional Poisson operator base (using spectral methods).

        Solves:
            Laplacian(Fout) = Fin
        
        Parameters
        ----------
        Fout: Field
            Input continuous field (rhs).
        Fin: Field
            Output continuous field (lhs), possibly inplace, same number of components as Fin.
        variables: dict
            Dictionary of fields as keys and topology descriptors as values.
        kwds: dict, optional
            Base class arguments.
        """
        check_instance(Fin,  Field)
        check_instance(Fout, Field)
        check_instance(variables, dict, keys=Field, values=CartesianTopologyDescriptors)
        assert Fin.nb_components == Fout.nb_components
        
        input_fields  = { Fin:  variables[Fin]  }
        output_fields = { Fout: variables[Fout] }
        
        default_name = 'Poisson_{}_{}'.format(Fin.name, Fout.name)
        default_pretty_name = 'Poisson_{}_{}'.format(Fin.pretty_name, Fout.pretty_name)
        name = first_not_None(name, default_name)
        pretty_name = first_not_None(name, default_pretty_name)

        super(PoissonOperatorBase, self).__init__(name=name, pretty_name=pretty_name,
                input_fields=input_fields, output_fields=output_fields, **kwds)

        forward_transforms  = ()
        backward_transforms = ()
        wave_numbers        = ()
        
        tg = self.new_transform_group()
        for (Fi,Fo) in zip(Fin.fields, Fout.fields):
            Ft = tg.require_forward_transform(Fi, custom_output_buffer='auto')
            Bt = tg.require_backward_transform(Fo, custom_input_buffer='auto',
                                                matching_forward_transform=Ft)
            assert (Ft.output_dtype == Bt.input_dtype)
            expr = Assignment(Bt.s, laplacian(Ft.s, Ft.s.frame))
            kds  = tg.push_expressions(expr)
            
            forward_transforms  += (Ft,)
            backward_transforms += (Bt,)
            wave_numbers        += (kds,)

        self.Fin  = Fin
        self.Fout = Fout
        self.tg   = tg
        self.forward_transforms  = forward_transforms
        self.backward_transforms = backward_transforms
        self.wave_numbers        = wave_numbers
    
    @debug
    def discretize(self):
        super(PoissonOperatorBase, self).discretize()
        self.dFin  = self.get_input_discrete_field(self.Fin)
        self.dFout = self.get_output_discrete_field(self.Fout)
        all_dkds, all_nd_dkds = (), ()
        for wn in self.wave_numbers:
            dkds, nd_dkds = [None,]*len(wn), [None,]*len(wn)
            for wi in wn:
                idx, dkd, nd_dkd = self.tg.discrete_wave_numbers[wi]
                dkds[idx]    = dkd
                nd_dkds[idx] = nd_dkd
            all_dkds += (dkds,)
            all_nd_dkds += (all_dkds,)
        self.all_dkds    = all_dkds
        self.all_nd_dkds = all_nd_dkds

