
from abc import abstractmethod
from hysop.constants import FieldProjection
from hysop.tools.numpywrappers import npw
from hysop.tools.types import check_instance, first_not_None, to_tuple
from hysop.tools.decorators import debug
from hysop.tools.numerics import float_to_complex_dtype
from hysop.fields.continuous_field import Field
from hysop.parameters.scalar_parameter import ScalarParameter
from hysop.operator.base.spectral_operator import SpectralOperatorBase
from hysop.topology.cartesian_descriptor import CartesianTopologyDescriptors
from hysop.fields.continuous_field import Field
from hysop.symbolic.field import laplacian, grad
from hysop.symbolic.relational import Assignment
from hysop.core.memory.memory_request import MemoryRequest

class PoissonCurlOperatorBase(object):
    """
    Solves the poisson-rotational equation using a specific implementation.
    """
    
    @debug
    def __init__(self, vorticity, velocity, variables, 
            diffusion=None, dt=None, projection=None, **kwds): 
        """
        PoissonCurl operator to solve incompressible flows using various fft backends.
        
        Parameters
        ----------
        velocity : :class:`~hysop.fields.continuous_field.Field
            Output solution velocity field.
        vorticity:  :class:`~hysop.fields.continuous_field.Field`
            Input vorticity to be diffused, projected.
            If diffused and/or projected, vorticity is also an output.
        variables: dict
            Dictionary of Fields as keys and topologies as values.
        diffusion: ScalarParameter, optional, defaults to None.
            Diffuse the vorticity field before applying projection and poisson velocity.
        dt: ScalarParameter, optional, defaults to None
            Timestep is only required for diffusion.
            If diffusion is not enabled, this parameter is ignored.
        projection: hysop.constants.FieldProjection or positive integer, optional
            Project vorticity such that resolved velocity is divergence free (for 3D fields).
            When active, projection is done prior to every solve, unless projection is 
            an integer in which case it is done every given steps.
            This parameter is ignored for 2D fields and defaults to no projection.
        kwds : 
            Base class parameters.
        """
        
        check_instance(velocity,  Field)
        check_instance(vorticity, Field)
        check_instance(variables, dict, keys=Field, values=CartesianTopologyDescriptors)
        check_instance(diffusion, ScalarParameter, allow_none=True)
        check_instance(dt, ScalarParameter, allow_none=True)
        check_instance(projection, (FieldProjection, int), allow_none=True)

        assert velocity.domain is vorticity.domain, 'only one domain is supported'
        assert variables[velocity] is variables[vorticity], 'only one topology is supported'

        # check for diffusion
        should_diffuse = (diffusion is not None)
        if should_diffuse:
            if (dt is None):
                msg='Diffusion has been specified but no timestep was given.'
                raise RuntimeError(msg)
        else:
            dt = None
        
        # check for projection
        if (projection == FieldProjection.NONE) or (projection is None) \
                or (projection==0) or (velocity.dim!=3):
            projection = FieldProjection.NONE
        should_project = (projection is not FieldProjection.NONE)
        
        if (projection == FieldProjection.NONE):
            do_project = lambda simu: False
        elif (projection == FieldProjection.EVERY_STEP):
            do_project = lambda simu: True
        else: # projection is an integer representing frequency 
            freq = projection
            if (freq>=1):
                do_project = lambda simu: ((simu.current_iteration % freq)==0)
            else:
                msg='Got negative projection frequency {}.'.format(freq)
                raise ValueError(msg)
        
        # check fields
        dim=velocity.dim
        wcomp = vorticity.nb_components
        if (dim==2) and (wcomp!=1):
            msg='Vorticity component mistmach, got {} components but expected 1.'.format(wcomp)
            raise RuntimeError(msg)
        if (dim==3) and (wcomp!=3):
            msg='Vorticity component mistmach, got {} components but expected 3.'.format(wcomp)
            raise RuntimeError(msg)
        if (dim!=3) and (projection!=FieldProjection.NONE):
            raise RuntimeError('Velocity projection only available in 3D.')
        if (velocity.dtype != vorticity.dtype):
            raise RuntimeError('Datatype mismatch between velocity and vorticity.')
        
        # input and output fields
        vtopology = variables[velocity]
        wtopology = variables[vorticity]
        input_params  = {}
        input_fields  = { vorticity: wtopology }
        output_fields = { velocity:  vtopology }
        if should_diffuse:
            assert (dt is not None), 'Diffusion timestep has not been given.'
            input_params[diffusion.name] = diffusion
            input_params[dt.name] = dt
        if (should_diffuse or should_project):
            output_fields[vorticity] = wtopology

        super(PoissonCurlOperatorBase, self).__init__(input_fields=input_fields, 
                output_fields=output_fields, input_params=input_params, **kwds)

        self.W = vorticity
        self.U = velocity
        self.dim = dim
        
        self.should_diffuse = should_diffuse
        self.nu = diffusion
        self.dt = dt

        self.should_project = should_project
        self.projection = projection
        self.do_project = do_project
    
    @debug
    def discretize(self):
        if self.discretized:
            return
        super(PoissonCurlOperatorBase, self).discretize()
        self.dW = self.get_input_discrete_field(self.W)
        self.dU = self.get_output_discrete_field(self.U)
        

class SpectralPoissonCurlOperatorBase(PoissonCurlOperatorBase, SpectralOperatorBase):
    @debug
    def __init__(self, vorticity, velocity, variables, 
            diffusion=None, dt=None, projection=None, **kwds): 

        super(SpectralPoissonCurlOperatorBase, self).__init__(
                vorticity=vorticity, velocity=velocity, variables=variables,
                diffusion=diffusion, dt=dt, projection=projection, **kwds)

        dim = self.dim
        should_diffuse, should_project = self.should_diffuse, self.should_project
        

        # build spectral transforms
        tg = self.new_transform_group()
        W_forward_transforms  = to_tuple(tg.require_forward_transform(vorticity))
        U_backward_transforms = to_tuple(tg.require_backward_transform(velocity,
                                                            custom_input_buffer='B0'))
        if (should_diffuse or should_project):
            W_backward_transforms = to_tuple(tg.require_backward_transform(vorticity))
        else:
            W_backward_transforms = (None,)*dim
        
        W_Fts = npw.asarray([x.s for x in W_forward_transforms])
        U_Bts = npw.asarray([x.s for x in U_backward_transforms])
        W_Bts = npw.asarray([None if (x is None) else x.s for x in W_backward_transforms])

        # generate wavenumbers
        kd1s = ()
        for Wi in W_Fts:
            expr1 = grad(Wi, Wi.frame)
            kd1 = sorted(tg.push_expressions(*to_tuple(expr1)), key=lambda k: k.axis)
            kd1s += (kd1,)

        # laplacian
        kd2s = ()
        for Wi in W_Fts:
            expr2 = laplacian(Wi, Wi.frame)
            kd2 = sorted(tg.push_expressions(*to_tuple(expr2)), key=lambda k: k.axis)
            kd2s += (kd2,)

        # curl 
        if (dim==2):
            W2,    = W_forward_transforms
            U0, U1 = U_backward_transforms
            exprs = (Assignment(U0.s, +W2.s.diff(W2.s.frame.coords[1])),
                     Assignment(U1.s, -W2.s.diff(W2.s.frame.coords[0])))
            Uin  = (W2, W2)
            Uout = (U0, U1)
            Uk   = tuple(tg.push_expressions(e)[0] for e in exprs)
        elif (dim==3):
            W0, W1, W2 = W_forward_transforms
            U0, U1, U2 = U_backward_transforms
            exprs = (Assignment(U0.s, +W2.s.diff(W2.s.frame.coords[1])),
                     Assignment(U0.s, -W1.s.diff(W1.s.frame.coords[2])),
                     Assignment(U1.s, +W0.s.diff(W0.s.frame.coords[2])),
                     Assignment(U1.s, -W2.s.diff(W2.s.frame.coords[0])),
                     Assignment(U2.s, +W1.s.diff(W1.s.frame.coords[0])),
                     Assignment(U2.s, -W0.s.diff(W0.s.frame.coords[1])))
            Uin  = (W2, W1, W0, W2, W1, W0)
            Uout = (U0, U0, U1, U1, U2, U2)
            Uk   = tuple(tg.push_expressions(e)[0] for e in exprs)
        else: 
            raise NotImplementedError
        
        self.tg = tg
        self.W_forward_transforms  = W_forward_transforms
        self.U_backward_transforms = U_backward_transforms
        self.W_backward_transforms = W_backward_transforms
        
        self.W_Fts = W_Fts
        self.U_Bts = U_Bts
        self.W_Bts = W_Bts

        self.kd1s = kd1s
        self.kd2s = kd2s

        self.Uin  = Uin
        self.Uout = Uout
        self.Uk   = Uk
    
    @debug
    def discretize(self):
        if self.discretized:
            return
        super(SpectralPoissonCurlOperatorBase, self).discretize()
        
        kd1s, kd2s = self.kd1s, self.kd2s
        if self.should_project:
            dkd1s = ()
            for kd1 in kd1s:
                dkd1 = [None,]*len(kd1)
                for wi in kd1:
                    idx, dkd, nd_dkd = self.tg.discrete_wave_numbers[wi]
                    dkd1[idx] = dkd
                dkd1s += (tuple(dkd1),)
        else:
            dkd1s = None

        dkd2s = ()
        for kd2 in kd2s:
            dkd2 = [None,]*len(kd2)
            for wi in kd2:
                idx, dkd, nd_dkd = self.tg.discrete_wave_numbers[wi]
                dkd2[idx] = dkd
            dkd2s += (tuple(dkd2),)
        
        dUk = ()
        for ki in self.Uk:
            _, dki, _ = self.tg.discrete_wave_numbers[ki]
            dUk += (dki,)
        
        self.dkd1s = dkd1s
        self.dkd2s = dkd2s
        self.dUk   = dUk
    
    def get_work_properties(self):
        requests = super(SpectralPoissonCurlOperatorBase, self).get_work_properties()
        Ut = self.U_backward_transforms
        dtypes = tuple(tr.input_dtype for tr in Ut)
        shapes = tuple(tr.input_shape for tr in Ut)
        assert all(d==dtypes[0] for d in dtypes)
        # we request one buffer per vorticity component (1 in 2D, 3 in 3D)
        for (i,(Ft,Bt)) in enumerate(zip(self.W_forward_transforms,
                                         self.W_backward_transforms)):
            assert Ft.output_dtype == dtypes[0]
            if (Bt is not None):
                assert (Ft.backend is Bt.backend)
                assert (Ft.output_dtype == Bt.input_dtype), (Ft.output_dtype, Bt.input_dtype)
                assert (Ft.output_shape == Bt.input_shape), (Ft.output_shape, Bt.input_shape)
            shape = Ft.output_shape
            dtype = Ft.output_dtype
            request = MemoryRequest(backend=self.tg.backend, dtype=dtype, 
                                    shape=shape, nb_components=1,
                                    alignment=self.min_fft_alignment)
            requests.push_mem_request('fft_buffer_{}'.format(i), request)
        return requests

    
    def setup(self, work):
        dim = self.dim
        Ks, KKs = self.dkd1s, self.dkd2s

        W_forward_transforms  = self.W_forward_transforms 
        W_backward_transforms = self.W_backward_transforms
        U_backward_transforms = self.U_backward_transforms

        for (i,(W_Ft,W_Bt)) in enumerate(zip(W_forward_transforms, 
                                             W_backward_transforms)):
            dtmp, = work.get_buffer(self, 'fft_buffer_{}'.format(i))
            W_Ft.configure_output_buffer(dtmp)
            if (W_Bt is not None):
                W_Bt.configure_input_buffer(dtmp)
            output_axes = W_Ft.output_axes

        super(SpectralPoissonCurlOperatorBase, self).setup(work)

        # extract buffers
        reorder_fields = tuple(dim-1-i for i in output_axes)
        if (dim==2):
            K   = Ks
            KK  = KKs
            WIN = (W_forward_transforms[0].full_output_buffer,)
        elif (dim==3):
            K   = tuple(Ks[i]  for i in reorder_fields) if Ks else None
            KK  = tuple(KKs[i] for i in reorder_fields)
            WIN = tuple(W_forward_transforms[i].full_output_buffer for i in reorder_fields)
        else:
            raise NotImplementedError
        self.WIN  = WIN
        self.K  = K
        self.KK = KK
                
        UIN, UOUT, UK = (), (), ()
        assert len(self.Uin) == len(self.Uout) == len(self.dUk)
        for i,(Uin, Uout, Uk) in enumerate(zip(self.Uin, self.Uout, self.dUk)):
            Uin  = Uin.full_output_buffer
            Uout = Uout.full_input_buffer
            UIN  += (Uin,)
            UOUT += (Uout,)
            UK   += (Uk,)
        
        self.UIN  = UIN
        self.UOUT = UOUT
        self.UK   = UK

