
import datetime, sys
from hysop.constants import Backend, MemoryOrdering
from hysop.tools.string_utils import vprint_banner
from hysop.tools.contexts import Timer
from hysop.tools.decorators import debug
from hysop.core.graph.computational_graph import ComputationalGraph
from hysop.tools.string_utils import vprint_banner, vprint

class Problem(ComputationalGraph):

    def __init__(self, name=None, method=None, 
                       check_unique_clenv=True, **kwds):
        super(Problem,self).__init__(name=name, method=method, **kwds)
        self._do_check_unique_clenv = check_unique_clenv

    @debug
    def insert(self, *ops):
        self.push_nodes(*ops)

    @debug
    def build(self, args=None, allow_subbuffers=False):
        with Timer() as tm:
            msg = self.build_problem(args=args, allow_subbuffers=allow_subbuffers)
            if msg:
                msg=' Problem {} achieved, exiting ! '.format(msg)
                vprint_banner(msg, at_border=2)
                sys.exit(0)
        msg=' Problem building took {} ({}s) '
        msg=msg.format(datetime.timedelta(seconds=round(tm.interval)), 
                        tm.interval)
        vprint_banner(msg, spacing=True, at_border=2)

        if (args is not None) and args.stop_at_build:
            msg=' Problem has been built, exiting. '
            vprint_banner(msg, at_border=2)
            sys.exit(0)
           
    def build_problem(self, args, allow_subbuffers):
        if (args is not None) and args.stop_at_initialization:
            return 'initialization'
        vprint('\nInitializing problem...')
        self.initialize(outputs_are_inputs=True, topgraph_method=None)
        
        if (args is not None) and args.stop_at_discretization:
            return 'discretization'
        vprint('\nDiscretizing problem...')
        self.discretize()
        
        if (args is not None) and args.stop_at_work_properties:
            return 'work properties retrieval'
        vprint('\nGetting work properties...')
        work = self.get_work_properties()

        if (args is not None) and args.stop_at_work_allocation:
            return 'work allocation'
        vprint('\nAllocating work...')
        work.allocate(allow_subbuffers=allow_subbuffers)
        
        if (args is not None) and args.stop_at_setup:
            return 'setup'
        vprint('\nSetting up problem...')
        self.setup(work)

    def discretize(self):
        super(Problem, self).discretize()
        if self._do_check_unique_clenv:
            self.check_unique_clenv()

    def check_unique_clenv(self):
        reduced_graph = self.reduced_graph
        operators = reduced_graph.vertex_properties['operators']

        cl_env, op = None, None
        for vid in self.sorted_nodes:
            vertex = reduced_graph.vertex(vid)
            op     = operators[vertex]
            for topo in set(op.input_fields.values() + op.output_fields.values()):
                if (topo.backend.kind == Backend.OPENCL):
                    if (cl_env is None):
                        first_op = op
                        cl_env = topo.backend.cl_env
                    elif (topo.backend.cl_env is not cl_env):
                        msg=''
                        msg+='\nOpenCl environment mismatch between operator {} and operator {}.'
                        msg=msg.format(first_op.name, op.name)
                        msg+='\n{}'.format(cl_env)
                        msg+='\n and'
                        msg+='\n{}'.format(topo.backend.cl_env)
                        msg+='\n If this is required, override check_unique_clenv().'
                        raise RuntimeError(msg)

    def initialize_field(self, field, **kwds):
        """Initialize a field on all its input and output topologies."""
        reduced_graph = self.reduced_graph
        operators = reduced_graph.vertex_properties['operators']

        initialized = set()
        for vid in self.sorted_nodes:
            vertex = reduced_graph.vertex(vid)
            op     = operators[vertex]
            for op_fields in (self.input_discrete_tensor_fields, self.output_discrete_tensor_fields,
                              op.input_discrete_tensor_fields, op.output_discrete_tensor_fields,
                              op.input_discrete_fields, op.output_discrete_fields):
                # give priority to tensor field initialization
                if (field in op_fields):
                    dfield = op_fields[field]
                    if all((df in initialized) for df in dfield.discrete_fields()):
                        # all contained scalar fields were already initialized
                        continue
                    elif dfield.has_unique_topology() and \
                            all((df not in initialized) for df in dfield.discrete_fields()):
                        # dfield has a unique topology and
                        # none of its scalar fields has ever been initialized
                        topo_view = dfield.topology
                        topo   = topo_view.topology
                        wstate = topo_view.topology_state.copy(memory_order=MemoryOrdering.C_CONTIGUOUS,
                                                               is_read_only=False) # get a writtable state
                        dfield = field.discretize(topo, wstate)
                        dfield.initialize(**kwds)
                        initialized.update(dfield.discrete_fields())
                    else:
                        # dfield has not unique topology 
                        # or some of its scalar fields were already initialized
                        for (component, dfield_i) in dfield.nd_iter():
                            if (dfield_i._dfield in initialized):
                                continue
                            topo_view = dfield_i.topology
                            topo      = topo_view.topology
                            wstate    = topo_view.topology_state.copy(memory_order=MemoryOrdering.C_CONTIGUOUS,
                                                                      is_read_only=False) # get a writtable state
                            dfield_i  = dfield_i.field.discretize(topo, wstate)
                            dfield_i.initialize(component=component, **kwds)
                            initialized.add(dfield_i._dfield)
        if not initialized:
            msg='FATAL ERROR: Could not initialize field {}.'.format(field.name)
            raise RuntimeError(msg)

    @debug
    def solve(self, simu, dry_run=False, dbg=None, 
                report_freq=10, plot_freq=10, **kargs):
        if dry_run:
            vprint()
            vprint_banner('** Dry-run requested, skipping simulation. **')
            return
        vprint('\nSolving problem...')
        simu.initialize()
        with Timer() as tm:
            while not simu.is_over:
                 vprint()
                 simu.print_state()
                 self.apply(simulation=simu, dbg=dbg, **kargs)
                 simu.advance(dbg=dbg, plot_freq=plot_freq)
                 if (simu.current_iteration % report_freq) == 0:
                     self.profiler_report()

        msg=' Simulation took {} ({}s) '
        msg+='\n  for {} iterations ({}s per iteration) '
        msg=msg.format(datetime.timedelta(seconds=round(tm.interval)), 
                       tm.interval, max(simu.current_iteration+1,1),
                       tm.interval/max(simu.current_iteration+1,1))
        vprint_banner(msg, spacing=True, at_border=2)
        
        simu.finalize()
        self.final_report()

        if (dbg is not None):
            dbg('final iteration', nostack=True)

    def final_report(self):
        self.profiler_report()

    @debug
    def finalize(self):
        vprint('Finalizing problem...')
        super(Problem, self).finalize()
        
