
from abc import ABCMeta, abstractmethod
from hysop.constants import Backend
from hysop.symbolic.base import DummySymbolicScalar, sm
from hysop.tools.types import check_instance, to_tuple, first_not_None
from hysop.tools.numpywrappers import npw
from hysop.backend.device.opencl import clArray
from hysop.backend.device.opencl.opencl_array import OpenClArray
from hysop.backend.host.host_array import HostArray

class SymbolicMemoryObject(DummySymbolicScalar):
    def __new__(cls, memory_object, name, **kwds):
        obj = super(SymbolicMemoryObject, cls).__new__(cls, name=name, **kwds)
        obj._memory_object = None
        if (memory_object is not None):
            obj.bind_memory_object(memory_object)
        return obj

    @property
    def is_bound(self):
        return (self._memory_object is not None)

    def assert_bound(self):
        if not self.is_bound:
            msg='{}::{} memory object has not been bound yet.'
            msg=msg.format(self.__class__.__name__, self.name)
            raise RuntimeError(msg)
    
    @property
    def memory_object(self):
        if (self._memory_object is None):
            msg='Symbolic memory_object {} has not been setup yet.'.format(self.name)
            raise RuntimeError(msg)
        return self._memory_object
    
    @abstractmethod
    def bind_memory_object(self, memory_object):
        if (self._memory_object is not None):
            msg='A memory_object has already been bound to SymbolicArray {}.'.format(self.name)
            raise RuntimeError(msg)
        if isinstance(memory_object, (OpenClArray, HostArray)):
            memory_object = memory_object.handle
        self._memory_object = memory_object
        return self

    @property
    def shape(self):
        self.assert_bound()
        return self._memory_object.shape
    @property
    def size(self):
        self.assert_bound()
        return self._memory_object.size
    @property
    def strides(self):
        self.assert_bound()
        return self._memory_object.strides
    @property
    def dtype(self):
        self.assert_bound()
        return self._memory_object.dtype
    @property
    def ctype(self):
        from hysop.backend.device.codegen.base.variables import dtype_to_ctype
        self.assert_bound()
        return dtype_to_ctype(self._memory_object.dtype)
    @property
    def dim(self):
        self.assert_bound()
        return self._memory_object.ndim

    def short_description(self):
        self.assert_bound()
        return '{}[dim={}, shape=[], strides={}, dtype={}]'.format(
                self.__class__.__name__,
                self.dim, self.shape, self.strides, self.dtype)
    
    def __eq__(self, other):
        return id(self) == id(other)
    def __hash__(self):
        return id(self)
    def _hashable_content(self):
        """See sympy.core.basic.Basic._hashable_content()"""
        hc = super(SymbolicMemoryObject, self)._hashable_content()
        hc += (str(id(self)),)
        return hc


class IndexedBuffer(sm.Indexed):
    """
    Tag for indexed SymbolicBuffers.
    """
    @property
    def indexed_object(self):
        return self.args[0].args[0]

    @property
    def index(self):
        assert len(self.args)==2
        return self.args[1]

class SymbolicArray(SymbolicMemoryObject):
    """
    An array is considered to be indexed by local indices in 
    autogenerated vectorized code (in symbolic code generation
    framework).
    """
    class ArrayRequirements(object):
        def __init__(self, sarray):
            self._symbolic_array = sarray
            self.min_ghosts = npw.int_zeros(shape=(sarray.dim,))
        def update_requirements(self, other):
            check_instance(other, self.__class__)
            assert (other.min_ghosts.size == self.min_ghosts.size)
            self.min_ghosts += other.min_ghosts

    def __new__(cls, memory_object, name, dim=None, **kwds):
        obj = super(SymbolicArray, cls).__new__(cls, memory_object=memory_object,
                    name=name, **kwds)
        msg='Dimension could not be deduced from memory_object, '
        msg+='please specify a dimension for symbolic array {}.'
        msg=msg.format(name.encode('utf-8'))
        if (memory_object is None):
            if (dim is None):
                raise RuntimeError(msg)
        elif hasattr(memory_object, 'ndim'):
            dim = memory_object.ndim
        elif hasattr(memory_object, 'dim'):
            dim = memory_object.dim
        else:
            raise RuntimeError(msg)
        check_instance(dim, int)
        obj._dim = dim
        return obj

    def __getitem__(self, key):
        msg='Symbolic array {} cannot be indexed.'.format(self.name)
        raise RuntimeError(msg)
    
    def new_requirements(self):
        return self.ArrayRequirements(self)

    @property
    def array(self):
        return self._memory_object

    @property
    def dim(self):
        return self._dim

    def to_backend(self, backend):
        if (backend is Backend.HOST):
            self.__class__ = HostSymbolicArray
        elif (backend is Backend.OPENCL):
            self.__class__ = OpenClSymbolicArray
        else:
            msg='Unknown backend kind {}.'.format(backend)
            raise NotImplementedError(msg)
        return self


class SymbolicBuffer(SymbolicMemoryObject):
    """
    A buffer will not be indexed by local indices by default.
    The user has to index a SymbolicBuffer so that it can be 
    used for code generation.
    """
    def __getitem__(self, key):
        assert isinstance(key, (int, npw.integer, sm.Expr))
        return IndexedBuffer(self, key)
    
    @property
    def buffer(self):
        return self._memory_object
    
    def to_backend(self, backend):
        if (backend is Backend.HOST):
            self.__class__ = HostSymbolicBuffer
        elif (backend is Backend.OPENCL):
            self.__class__ = OpenClSymbolicBuffer
        else:
            msg='Unknown backend kind {}.'.format(backend)
            raise NotImplementedError(msg)
        return self


class SymbolicHostMemoryObject(object):
    def bind_memory_object(self, memory_object):
        check_instance(memory_object, (HostArray, npw.ndarray))
        return super(SymbolicHostMemoryObject, self).bind_memory_object(memory_object)


class SymbolicDeviceMemoryObject(object):
    def bind_memory_object(self, memory_object):
        check_instance(memory_object, (OpenClArray, clArray.Array))
        return super(SymbolicDeviceMemoryObject, self).bind_memory_object(memory_object)
    @property
    def base_data(self):
        self.assert_bound()
        return self._memory_object.base_data
    @property
    def offset(self):
        self.assert_bound()
        return self._memory_object.offset


class HostSymbolicArray(SymbolicHostMemoryObject, SymbolicArray):
    pass
class OpenClSymbolicArray(SymbolicDeviceMemoryObject, SymbolicArray):
    pass
class HostSymbolicBuffer(SymbolicHostMemoryObject, SymbolicBuffer):
    pass
class OpenClSymbolicBuffer(SymbolicDeviceMemoryObject, SymbolicBuffer):
    pass
    

if __name__ == '__main__':
    from hysop.deps import sm
    from hysop.core.arrays.all import default_host_array_backend
    a = npw.ones(shape=(10,10), dtype=npw.int8)
    b = default_host_array_backend.zeros(shape=(10,), dtype=npw.uint16)
    
    A = HostSymbolicArray(a, name='a')
    B = b.as_symbolic_array('b')
    
    C = HostSymbolicBuffer(a, name='c')
    D = b.as_symbolic_buffer('d')

    print 7*A + 9*B + 10*C - 4*D

    assert A.array  is a
    assert B.array  is b.handle
    assert C.buffer is a
    assert D.buffer is b.handle
    
    try:
        A[5]
        assert False
    except RuntimeError as e:
        pass 

    print C[5]
    print D[C]

    print 
