
import os, shutil, datetime, inspect, argparse, sys
import numpy as np

from hysop.tools.types import check_instance
from hysop.fields.discrete_field import DiscreteScalarFieldView

class DebugDumper(object):
    def __init__(self, name, path='/tmp/hysop/debug', force_overwrite=False,
            enable_on_op_apply=False):
        directory = path+'/'+name
        blobs_directory = directory + '/data'
        
        if os.path.exists(directory):
            if force_overwrite:
                shutil.rmtree(directory)
            else:
                msg='Directory \'{}\' already exists.'.format(directory)
                raise RuntimeError(msg)

        os.makedirs(blobs_directory)

        runfile = '{}/run.txt'.format(directory)
        runfile = open(runfile, 'a')

        self.name = name
        self.directory = directory
        self.blobs_directory = blobs_directory
        self.runfile = runfile
        self.dump_id = 0
        self.enable_on_op_apply = enable_on_op_apply

        self.print_header()

    def __del__(self):
        if hasattr(self, 'runfile') and (self.runfile is not None):
            self.runfile.close()
            self.runfile = None
    
    @classmethod
    def lformat(cls, id_, iteration, time, tag, dtype, shape, min_, max_, mean, variance, description=''):
        return '{:<4}  {:<10}  {:<20}  {:<20}  {:<20}  {:<20}  {:<20}  {:<20}  {:<20}  {:<20}  {}'.format(
                id_, iteration, time, tag, dtype, shape, min_, max_, mean, variance, description)
    
    def print_header(self):
        now = datetime.datetime.now()
        self.runfile.write('DEBUG DUMP {} ({})\n'.format(
            self.name,
            now.strftime("%Y-%m-%d %H:%M")))
        self.runfile.write(self.lformat('id', 'iteration', 'time', 'tag', 'dtype', 'shape', 'min', 'max', 'mean', 'variance',
                                        'description'))

    @classmethod
    def get_arrays(self, arrays):
        if isinstance(arrays, DiscreteScalarFieldView):
            return tuple(d.get()[arrays.compute_slices] for d in arrays.data)
        elif isinstance(arrays, np.ndarray):
            return (arrays,)
        else:
            check_instance(arrays, tuple, values=np.ndarray)
            return arrays
        

    def __call__(self, iteration, time, tag, arrays, description=''):
        check_instance(iteration, int)
        arrays = self.get_arrays(arrays)
        N = len(arrays)
        for (i, data) in enumerate(arrays):
            if (N>1):
                tag_ = '{}_{}'.format(tag, i)
            else:
                tag_ = tag
            if (description != '') and (N>1):
                description_ = '{} (component {})'.format(description, i)
            else:
                _file,_line = inspect.stack()[1][1:3]
                description_='{}:{}'.format(_file, _line)
            dtype = data.dtype
            if dtype in (np.complex64, np.complex128):
                data = (data.real, data.imag)
                tags = (tag_+'_real', tag_+'_imag')
            else:
                data = (data,)
                tags = (tag_,)
            for (d,tag_) in zip(data, tags):
                dtype = d.dtype
                shape = d.shape
                id_ = self.dump_id
                _min, _max = np.nanmin(d), np.nanmax(d)
                mean, variance = np.nanmean(d), np.nanvar(d)
                entry = '\n'+self.lformat(id_, iteration, time, tag_, dtype, shape, _min, _max, mean, variance, description_)
            
                self.runfile.write(entry)

                dst = '{}/{}'.format(self.blobs_directory, self.dump_id)
                np.savez_compressed(dst, data=d)

                self.dump_id += 1

