"""Tools related to i/o in HySoP.

.. currentmodule hysop.tools.io_utils

* :class:`~IO`
* :class:`~IOParams`
* :class:`~Writer`
* :class:`~XMF`, tools to prepare/write xmf files.

"""
import os, h5py, psutil, warnings, tempfile, socket
import subprocess32 as subprocess
from collections import namedtuple
from inspect import getouterframes, currentframe
from re import findall

from hysop.tools.parameters import MPIParams
from hysop.tools.warning import HysopWarning
from hysop.tools.decorators import requires_cmd
import hysop.core.mpi as mpi

class IO(object):
    """
    Static class with utilities to set/find the place where i/o files
    will be read/written.
    """

    _default_path = None
    _cache_path = None

    HDF5 = 998
    """HDF5 format id"""

    ASCII = 997
    """ascii format id"""


    @staticmethod
    @requires_cmd('stat')
    def get_fs_type(path):
        cmd = ['stat', '-f', '-c', '%T', path]
        fs_type = subprocess.check_output(cmd)
        return fs_type.replace('\n','')

    @classmethod
    def is_shared_fs(cls, path):
        return (cls.get_fs_type(path) in ['nfs'])

    @classmethod
    def default_path(cls):
        """Get the current default path used for io.

        Returns
        -------
        string
            the default value of the current i/o path.
        """
        assert (cls._default_path is not None), 'default path has not been set.'
        return cls._default_path

    @staticmethod
    def check_dir(filename, io_rank=0, comm=None):
        """Check if the directory of 'filename' exists and creates it if not.

        Parameters
        -----------
        filename : string
            file name with full or relative path
        io_rank : int
            processus rank that does the check.
        comm : mpi communicator
            the mpi communicator that does the check.
        """
        # Create output dir if required
        if (comm is None):
            comm = mpi.main_comm
        if (comm.Get_rank() == io_rank):
            d = os.path.dirname(filename)
            if not os.path.exists(d):
                os.makedirs(d)

    @staticmethod
    def set_default_path(pathdir):
        """Set a new default path for hysop i/o.

        Parameters
        -----------
        pathdir : string
            the new path

        Notes
        ------
        pN will be add to path name, N being the number of MPI process
        used for the simulation.

        """
        assert isinstance(pathdir, str)
        IO._default_path = os.path.join(pathdir,
                                        'p' + str(mpi.main_size))
        IO.check_dir(IO._default_path)

    @classmethod
    def default_cache_path(cls):
        from hysop import get_env
        home = os.path.expanduser('~')
        tmp  = tempfile.gettempdir()
        candidates = [get_env('CACHE_DIR', None), '{}/.cache'.format(home), home, '{}'.format(tmp)]
        cpath = None
        for c in candidates:
            if (c is None) or (not os.path.isdir(c)) or cls.is_shared_fs(c):
                continue
            if (c == home):
                cpath = '{}/.hysop'.format(home)
            else:
                cpath = '{}/hysop'.format(c)
            break
        if (cpath is None):
            msg='No suitable caching directory was found in {}.'
            msg=msg.format(candidates)
            raise RuntimeError(msg)
        if not os.path.exists(cpath):
            try:
                os.makedirs(cpath)
            except IOError:
                pass
        return cpath

    @staticmethod
    def cache_path():
        if IO._cache_path is None:
            IO.set_cache_path(IO.default_cache_path())
        return IO._cache_path
    
    @classmethod
    def set_cache_path(cls, path):
        if cls.is_shared_fs(path):
            new_path += '/{}'.format(socket.gethostname())
            msg='\nSpecified cache path \'{}\' is stored on a network filesystem '  
            msg += 'which does not correctly support file locking.'
            msg += '\nSetting cache_path to \'{}\'.'
            msg=msg.format(path, new_path)
            warnings.warn(msg, HysopWarning)
            path = new_path
        IO._cache_path = path
        IO.check_dir(path)

    @staticmethod
    def set_datasetname(field_name, topo, direction=None):
        """Return the dataset name of a given continuous field,
        saved for a given topology
        """
        val = field_name + '_' + str(topo.get_id())
        if direction is not None:
            val += DirectionLabels[direction]
        return val

    @staticmethod
    def get_datasetnames(filename):
        """Return the list of dataset names present
        in hdf input file

        Parameters
        ----------
        filename : string
            hdf file

        Returns
        -------
            a list of strings
        """
        hdf_file = h5py.File(filename, 'r')
        keys = hdf_file.keys()
        hdf_file.close()
        return keys


class IOParams(namedtuple("IOParams", ['filename', 'filepath',
                                       'frequency', 'fileformat',
                                       'io_leader'])):
    """
    A struct to handle I/O files parameters

    Parameters
    -----------
    filename : string
        name of the file (absolute or relative path)
    filepath : string
        location of the file
    frequency : int
        frequency of output or input (e.g. every N times steps)
    fileformat : int
        format of the file. See notes for available format. Default=HDF5.
    io_leader : int
        rank of the mpi process dealing with the io. Default is 0.

    See examples in hysop.operator.hdf_io

    Notes
    -----
    Format parameter must be one of the following :
      - :class:`~IO.HDF5`
      - :class:`~IO.ASCII`

    """
    def __new__(cls, filename, filepath=None, frequency=1,
                fileformat=None, io_leader=0):

        # Filename is absolute path, filepath arg is ignored.
        if os.path.isabs(filename):
            filepath = os.path.dirname(filename)

        else:
            if (filepath is not None):
                filename = os.path.join(filepath, filename)
                filepath = os.path.abspath(os.path.dirname(filename))
            else:
                filepath = os.path.dirname(filename)
                if filepath == '':
                    # Get default output path
                    filepath = IO.default_path()
                    filename = os.path.join(filepath, filename)
                else:
                    filepath = os.path.abspath(filepath)
                    filename = os.path.join(filepath,
                                            os.path.basename(filename))
        if (fileformat is None):
            fileformat = IO.HDF5

        IO.check_dir(filename)
        return super(IOParams, cls).__new__(cls, filename, filepath,
                                            frequency, fileformat, io_leader)


class Writer(object):
    """
    To write data from a 2D numpy array into an ascii file.

    Examples
    --------

    >>> from hysop.tools.io_utils import IOParams, IO
    >>> params = IOParams(filename='r.dat', fileformat=IO.ASCII)
    >>> wr = Writer(params, buffshape=(1, 2))
    >>> ite = 3 # current iteration number
    >>> if wr.do_write(ite):
    ...    wr.buffer[...] = 3.
    ...    wr.write()
    >>> wr.finalize()

    result : buffer is written into r.dat
    """
    def __init__(self, io_params, buffshape=None, mpi_params=None,
                 safe_io=True):
        """

        Parameters
        ----------
        io_params : hysop.tools.io_utils.IOParams
            setup for file ouput (name, location ...)
        buffshape : tuple
            2D numpy.array.shape like tuple, shape of the output/input buffer.
        mpi_params : hysop.tools.parameters.MPIParams
            mpi setup (comm that owns the writer)
        safe_io : boolean
            True --> open/close file everytime data are written.
            False --> open at init and close during finalize.
            Cost less but if simu crashes, data are lost.

        """
        # Absolute path + name for i/o file
        # Note that if filename contains absolute path
        # filepath is ignored
        msg = 'wrong type for io_params arg.'
        assert isinstance(io_params, IOParams), msg
        assert io_params.fileformat == IO.ASCII
        self.io_params = io_params

        # A reference communicator, just to identify a
        # process rank for io.
        if mpi_params is None:
            mpi_params = MPIParams()
        else:
            msg = 'wrong type for mpi_params arg.'
            assert isinstance(mpi_params, MPIParams), msg
        self.mpi_params = mpi_params

        # check if output dir exists, create it if not.
        IO.check_dir(self.io_params.filename, self.io_params.io_leader,
                     self.mpi_params.comm)

        # Shape of the output buffer (must be a 2D numpy array)
        if buffshape is None:
            buffshape = (1, 1)
        self._buffshape = buffshape
        assert len(self._buffshape) == 2,\
            '2D shape required : set arg buffshape as a 2D tuple: (x,y)'

        # The buffer (numpy array) that will be printed to a file
        from hysop.tools.numpywrappers import npw
        self.buffer = npw.zeros(self._buffshape)
        " buffer used to save printed data"

        # Defines how often
        # output file is written :
        # True --> open/close file everytime
        # data are written.
        # False --> open at init and close
        # during finalize. Cost less but if simu
        # crashes, data are lost.
        if safe_io:
            self.write = self._fullwrite
        else:
            self.write = self._partialwrite

        # Force synchro to be sure that all output dirs
        # have been created.
        self.mpi_params.comm.barrier()
        if self.mpi_params.rank == self.io_params.io_leader:
            self._file = open(self.io_params.filename, 'w')

    def do_write(self, ite):
        """Returns true if output is required
        for iteration ite

        Parameters
        ----------
        ite : int
            current iteration number

        """
        num = ite + 1  # Number of iterations done
        rk = self.mpi_params.rank
        return rk == self.io_params.io_leader and \
            (num % self.io_params.frequency) == 0

    def _fullwrite(self):
        """open, write and close"""
        import scitools.filetable as ft
        self._file = open(self.io_params.filename, 'a')
        ft.write(self._file, self.buffer)
        self._file.close()

    def _partialwrite(self):
        """just write, no open, nor close"""
        import scitools.filetable as ft
        ft.write(self._file, self.buffer)

    def finalize(self):
        """close, if required"""
        if self.mpi_params.rank == self.io_params.io_leader:
            if not self._file.closed:
                self._file.close()

    def __str__(self):
        if self.mpi_params.rank == self.io_params.io_leader:
            s = ' === Writer === \n'
            s += ' - filename = ' + self.io_params.filename
            s += '\n - buffshape = ' + str(self._buffshape)
            s += '\n - frequ = ' + str(self.io_params.frequency)
            return s


class XMF(object):
    """Static class - Tools to prepare and write xmf file
    """

    @staticmethod
    def _list_format(l):
        """Format a list to the xml output.
        Removes the '[]()' and replace ',' with ' ' in default str.

        Parameters
        ----------
        l : list to format

        """
        buff = str(l).replace(',', ' ').replace('[', '')
        return buff.replace(']', '').replace('(', '').replace(')', '')

    @staticmethod
    def prepare_grid_attributes(dataset_names,
            resolution, origin, step):
        """
        Prepare XDMF header as a string.

        Parameters
        -----------
        dataset_names : list
            all datasets names
        resolution: 3d tuple
        origin: 3d tuple
        step: 3d tuple
        subset : :class:`hysop.domain.subsets.Subset`, optional
            to define a grid only on this subset.
            If None, grid on the whole domain (from topo)

        Returns:
        --------
        string
            the xml-like header formattable with the following keywords:
                niteration : iteration number
                time: time in seconds
                filename: target file name

        """
        # The header (xml-like), saved in a string.
        # always use a 3D mesh because paraview does not like 2D meshes (uses axe (Y,Z) instead of (X,Y)).
        xml_grid = ""
        topo_type = "3DCORECTMesh"
        geo_type = "ORIGIN_DXDYDZ"
        xml_grid += "   <Grid Name=\"Iteration {}\"".format('{niteration:03d}')
        xml_grid += " GridType=\"Uniform\">\n"
        xml_grid += "    <Time Value=\"{}\" />\n".format('{time}')
        xml_grid += "    <Topology TopologyType=\"" + str(topo_type) + "\""
        xml_grid += " NumberOfElements=\""
        xml_grid += XMF._list_format(resolution) + " \"/>\n"
        xml_grid += "    <Geometry GeometryType=\"" + geo_type + "\">\n"
        xml_grid += "     <DataItem Dimensions=\"" + str(3) + " \""
        xml_grid += " NumberType=\"Float\" Precision=\"8\" Format=\"XML\">\n"
        xml_grid += "     " + XMF._list_format(origin) + "\n"
        xml_grid += "     </DataItem>\n"
        xml_grid += "     <DataItem Dimensions=\"" + str(3) + " \""
        xml_grid += " NumberType=\"Float\" Precision=\"8\" Format=\"XML\">\n"
        xml_grid += "     " + XMF._list_format(step) + "\n"
        xml_grid += "     </DataItem>\n"
        xml_grid += "    </Geometry>\n"
        # Append dataset parameters
        for name in dataset_names:
            xml_grid += "    <Attribute Name=\""
            xml_grid += name + "\""
            xml_grid += " AttributeType=\"Scalar\" Center=\"Node\">\n"
            xml_grid += "     <DataItem Dimensions=\""
            xml_grid += XMF._list_format(resolution) + " \""
            xml_grid += " NumberType=\"Float\" Precision=\"8\" Format=\"HDF\""
            xml_grid += " Compression=\"Raw\">\n"  #
            xml_grid += "      {filename}"
            xml_grid += ":/" + name
            xml_grid += "\n     </DataItem>\n"
            xml_grid += "    </Attribute>\n"
        xml_grid += "   </Grid>\n"
        return xml_grid
