
from hysop.deps import np, sm, copy
from hysop.tools.types import first_not_None, check_instance, to_tuple

from sympy.utilities import group
from sympy.printing.str import StrPrinter, StrReprPrinter
from sympy.printing.latex import LatexPrinter
    
# unicode subscripts for decimal numbers, signs and parenthesis
decimal_subscripts  = tuple('\u208{}'.format(i).decode('unicode-escape') for i in xrange(10))
decimal_exponents   = (u'\u2070', u'\u00b9', u'\u00b2', u'\u00b3')
decimal_exponents  += tuple('\u207{}'.format(i).decode('unicode-escape') for i in xrange(4,10))
greak = tuple('\u{:04x}'.format(i).decode('unicode-escape') for i in xrange(945, 970))
Greak = tuple('\u{:04x}'.format(i).decode('unicode-escape') for i in xrange(913, 938))
signs = (u'\u208a',u'\u208b')
parenthesis = (u'\u208d', u'\u208e')
partial = u'\u2202'
nabla = u'\u2207'
xsymbol = u'x'
freq_symbol = greak[12] # nu

def round_expr(expr, num_digits=3):
    return expr.xreplace({n : round(n, num_digits) for n in 
        expr.atoms(sm.Float).union(expr.atoms(sm.Rational)).difference(expr.atoms(sm.Integer))})
def truncate_expr(expr, maxlen=80):
    assert maxlen>=3
    parts = sstr(expr).split(' ')
    assert parts
    ss=parts.pop(0)
    while parts and (len(ss+parts[0])<maxlen):
        ss+=parts.pop(0)
    if parts:
        ss+='...'
    return ss

class CustomStrPrinter(StrPrinter):
    def _print_Derivative(self, expr):
        _partial = partial.encode('utf-8')
        syms = list(reversed(expr.variables)) 
        nvars = len(syms)
                
        if isinstance(expr.expr, (Symbol,Dummy,AppliedUndef)):
            content = self._print(expr.expr)
        else:
            content = '[{}]'.format(self._print(expr.expr))
        
        prefix = '{}{}{}/{}'.format(_partial, exponent(nvars).encode('utf-8') if nvars>1 else '', 
                content, _partial)
        for (sym, num) in group(syms, multiple=False):
            prefix += '{}{}'.format(sym, exponent(num).encode('utf-8') if num>1 else '')
        return prefix

class CustomStrReprPrinter(StrReprPrinter):
    pass

class CustomLatexPrinter(LatexPrinter):
    pass

def sstr(expr, **settings):
    p = CustomStrPrinter(settings)
    return p.doprint(expr)

def sstrrepr(expr, **settings):
    p = CustomStrReprPrinter(settings)
    return p.doprint(expr)

def latex(expr, **settings):
    p = CustomLatexPrinter(settings)
    return p.doprint(expr)

def enable_pretty_printing():
    sm.Basic.__str__ = sstr

class SymbolicBase(object):
    def __new__(cls, name, var_name=None, latex_name=None, 
            pretty_name=None, **kwds):
        if isinstance(name, unicode):
            name = name.encode('utf-8')
        if isinstance(pretty_name, unicode):
            pretty_name = pretty_name.encode('utf-8')
        if isinstance(latex_name, unicode):
            latex_name = latex_name.encode('utf-8')
        check_instance(name, str)
        check_instance(var_name, str, allow_none=True)
        check_instance(latex_name, str, allow_none=True)
        check_instance(pretty_name, str, allow_none=True)
        obj = super(SymbolicBase, cls).__new__(cls, name=name, **kwds)
        obj._name =  name
        obj._var_name    = first_not_None(var_name, name)
        obj._latex_name  = first_not_None(latex_name, name)
        obj._pretty_name = first_not_None(pretty_name, name)
        return obj
    
    @property
    def varname(self):
        return self._var_name

    def _sympystr(self, printer):
        return self._pretty_name
    def _latex(self, printer):
        return self._latex_name
    def _ccode(self, printer):
        return self._var_name
    def _pretty(self, printer):
        return self._pretty_name

    def __str__(self):
        return self._pretty_name
    def __repr__(self):
        return self._name

class Expr(sm.Expr):
    """Tag for hysop symbolic expressions."""
    pass

class UnevaluatedExpr(sm.UnevaluatedExpr):
    """Tag for hysop symbolic uneevaluated expressions."""
    pass

class UnsplittedExpr(Expr):
    """Tag for hysop symbolic unsplitted expressions."""
    pass

class Symbol(SymbolicBase, sm.Symbol):
    """Tag for hysop symbolic symbols."""
    pass

class Dummy(SymbolicBase, sm.Dummy):
    """Tag for hysop dummy symbolic variables."""
    pass

class UndefinedFunction(SymbolicBase, sm.function.UndefinedFunction):
    """
    Tag for hysop (unapplied) undefined functions.
    This is a metaclass.
    """
    pass

class AppliedUndef(sm.function.AppliedUndef):
    """Tag for hysop applied undefined functions."""

    def _latex(self, printer):
        return self._latex_name
    def _ccode(self, printer):
        return self._var_name
    def _pretty(self, printer):
        return self._pretty_name
    def _sympystr(self, printer):
        return self._pretty_name
    #def _pretty(self, printer):
        #return '{}({})'.format(self._pretty_name, 
                                #','.join(printer._print(a) for a in self.args))
    #def _sympystr(self, printer):
        #return '{}({})'.format(self._pretty_name, 
                                #','.join(printer._print(a) for a in self.args))

def subscript(i, with_sign=False):
    """
    Generate an unicode subscript of value i, signs can be enforced.
    """
    decimals = '0123456789'
    snumber = str(i)
    if with_sign:
        s0 = snumber[0]
        if s0 in decimals:
            snumber = '+'+snumber
    out = u''
    for s in snumber:
        if s in decimals:
            out += decimal_subscripts[int(s)]
        elif s=='+':
            out += signs[0]
        elif s=='-':
            out += signs[1]
        else:
            out += s
    return out

def exponent(i, with_sign=False):
    """
    Generate an unicode exponent of value i, signs can be enforced.
    """
    decimals = '0123456789'
    snumber = str(i)
    if with_sign:
        s0 = snumber[0]
        if s0 in decimals:
            snumber = '+'+snumber
    out = u''
    for s in snumber:
        if s in decimals:
            out += decimal_exponents[int(s)]
        elif s=='+':
            out += signs[0]
        elif s=='-':
            out += signs[1]
        else:
            out += s
    return out

def subscripts(ids,sep,with_sign=False,with_parenthesis=False,prefix=''):
    """
    Generate a unicode tuple subscript separated by sep,
    with or without parenthesis, prefix, and signs.
    """
    ids = to_tuple(ids)
    if with_parenthesis:
        return u'{}{}{}{}'.format(prefix,parenthesis[0],sep.join([subscript(i,with_sign) for i in ids]),parenthesis[1])
    else:
        return u'{}{}'.format(prefix,sep.join([subscript(i,with_sign) for i in ids]))

def exponents(ids,sep,with_sign=False,with_parenthesis=False,prefix=''):
    """
    Generate a unicode tuple exponent separated by sep,
    with or without parenthesis, prefix, and signs.
    """
    ids = to_tuple(ids)
    if with_parenthesis:
        return u'{}{}{}{}'.format(prefix,parenthesis[0],sep.join([exponent(i,with_sign) for i in ids]),parenthesis[1])
    else:
        return u'{}{}'.format(prefix,sep.join([exponent(i,with_sign) for i in ids]))

def tensor_symbol(prefix,shape,origin=None,mask=None,
        sep=None,with_parenthesis=False,force_sign=False):
    """
    Generate a np.ndarray of sympy.Symbol.
    Each of the symbol has given prefix and subscripts are 
    taken from specified origin if specified or else in matrix/tensor notation.
    Other parameters handles subscripts style, see the subscripts() function.
    
    It also returns all generated Symbols as a list.
    """
    origin = np.asarray(origin) if origin is not None else np.asarray([0]*len(shape))
    sep = sep if sep is not None else u','
    
    with_sign = force_sign or ((origin>0).any() and len(shape)>1)
    tensor = np.empty(shape=shape,dtype=object)
    for idx in np.ndindex(*shape):
        if (mask is None) or mask[idx]:
            ids = idx-origin
            sname = subscripts(ids,sep,with_sign=with_sign,
                    with_parenthesis=with_parenthesis,prefix=prefix)
            tensor[idx] = sm.Symbol(sname.encode('utf-8'), real=True)
        else:
            tensor[idx] = 0
    tensor_vars = tensor.ravel().tolist()
    return tensor, tensor_vars

def tensor_xreplace(tensor,vars):
    """
    Performs an xreplace and all tensor (np.ndarray) elements.
    """
    T = tensor.copy()
    for idx in np.ndindex(*tensor.shape):
        symbol = tensor[idx]
        if isinstance(symbol,sm.Expr):
            if (symbol in vars.keys()):
                T[idx] = vars[symbol]
            elif (symbol.name in vars.keys()):
                T[idx] = vars[symbol.name]
    return T

def non_eval_xreplace(expr, rule):
    """
    Duplicate of sympy's xreplace but with non-evaluate statement included.
    """
    if expr in rule:
        return rule[expr]
    elif rule:
        args = []
        altered = False
        for a in expr.args:
            try:
                new_a = non_eval_xreplace(a, rule)
            except AttributeError:
                new_a = a
            if new_a != a:
                altered = True
            args.append(new_a)
        args = tuple(args)
        if altered:
            return expr.func(*args, evaluate=False)
    return expr
       
# Convert powers to mult. in polynomial expressions V
# Example: x^3 -> x*x*x
def remove_pows(expr):
    """
    Convert pows to multiplications: x^3 -> x*x*x
    """
    pows = list(expr.atoms(sm.Pow))
    repl = [sm.Mul(*[b]*e,evaluate=False) for b,e in [i.as_base_exp() for i in pows]]
    e = non_eval_xreplace(expr,dict(zip(pows,repl)))
    return e

def evalf_str(x,n,literal='',significant=True):
    """
    Call evalf on x up to n-th decimal and removes zeros 
    if significant is set.
    """
    x = x.evalf(n).__str__()
    if significant:
        i=len(x)
        while i>1 and x[i-1] == '0':
            i-=1
        if i>1:
            x = x[:i+1]
    return x+literal

def factor_split(expr,variables,constant_var=None,include_var=False,init=0,_factor=True,_handle_const=True):
    """
    Factorize and split expresssion.
    """
    expr = expr.expand()
    factors = {}
    for var in variables:
        factors[var] = init
    for arg in expr.args:
        I = arg.atoms(sm.Symbol).intersection(variables)
        if len(I)==0:
            continue
        elif len(I)==1:
            var = I.pop()
            if not include_var:
                arg = arg.xreplace({var:1})
            factors[var] += arg
        else:
            assert False, 'Expression containing two or more variables!\n{} contains {}.'.format(arg,I)
    return factors

def build_eqs_from_dicts(d0,d1):
    """
    Build equations from two dictionnaries (lhs and rhs).
    ie. rhs-lhs = 0
    For keys that are only present in one dictionary, the other operand is set to 0.
    """
    treated = []
    eqs = []
    for k in d0.keys():
        expr0 = d0[k]
        expr1 = d1[k] if k in d1.keys() else 0
        de = expr1-expr0
        if de != 0:
            eqs.append(de)
        treated.append(k)
    for k in d1.keys():
        if k not in treated:
            expr0 = 0
            expr1 = d1[k]
            de = expr1-expr0
            if de != 0:
                eqs.append(de)
    return eqs

def recurse_expression_tree(op, expr):
    """
    Recurse through sympy expression tree and apply op on each subexpression.
    """
    op(expr)
    if isinstance(expr, sm.Expr):
        for arg in expr.args:
            recurse_expression_tree(op, arg)

def get_derivative_variables(expr):
    """Derivative arguments changed for sympy 1.2"""
    assert isinstance(expr, sm.Derivative)
    from sympy.core import containers
    if isinstance(expr.args[1], containers.Tuple):
        # sympy >= 1.2 arguments are (variable, count)
        # args = [(x0,3), (x1,1)]
        _vars = tuple(v[0] for v in expr.args[1:] for _ in xrange(v[1]))
    else:
        # sympy < 1.2 arguments are repeated 
        # args=[x0, x0, x0, x1]
        _vars = tuple(expr.args[1:])
    return _vars

class SetupExprI(object):
    """Interface for setupable expressions."""
    def setup(self, work):
        raise NotImplementedError

